#!/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   21-Nov-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import sys
import glob
import getopt
import tkinter as tk
from tkinter import messagebox

if __name__ == "__main__":
	sys.path.append(os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

import log
import rexx
import tool
import tkExtra
import tkFlair
import tkDialogs

_directories = None
current      = None
images       = {}

#-------------------------------- Names ----------------------------------------
_change = [	("\\`e",  u"è"),
		("\\'e",  u"é"),
		("\\`o",  u"ò"),
		("\\\"u", u"ü"),
		("\\o",   u"ö"),
		("\\ie\\", "i.e.") ]
_patterns = [	re.compile(r"^(\d+)} +[A-Z].*$"),
		re.compile(r"^([A-Za-z\-]\d+)} +[A-Za-z\[\]].*$"),
		re.compile(r"^{([A-Za-z0-9\-:\[\]]+)}[a-z]*$"),
		re.compile(r"^{([A-Za-z0-9\-:\[\]]+)} +[A-Za-z\[\]].*$") ]
_subpatterns = [ re.compile(r"^ *(\d+\.\d+)} +[A-Za-z0-9].*$"),
		 re.compile(r"^ *(\d+\.\d+.\d+)} +[A-Za-z].*$"),
		 re.compile(r"^ *(\d+\.\d+.\d+\.\d+)} +[A-Za-z].*$"),
		 re.compile(r"^ *([A-Za-z]\d+\.\d+)} +[A-Za-z].*$"),
		 re.compile(r"^ *([A-Za-z]\d+\.\d+\.\d+)} +[A-Za-z].*$"),
		 re.compile(r"^ *([A-Za-z]\d+\.\d+\.\d+\.\d+)} +[A-Za-z].*$") ]
_patbody1   = re.compile(r"^ +\d+\. [A-Z].*\. Code: ([A-Z].*)$")
_patbody2   = re.compile(r"^ +\d+\.\d+\.\d+\.\d+} [A-Z].*\. Code: ([A-Z].*)$")
_patwhat    = re.compile(r"\bWHAT\([1-9.]+\)")
_patsdum    = re.compile(r"\bSDUM\b")
_patcomment = re.compile(r"^ \*.*$", re.MULTILINE)
_pathtmltag = re.compile(r"<([a-z]+)[ >].*?</\1>")

_MAX_VIEWS  = 20

#===============================================================================
class Manual:
	# Manual is stored as a dictionary of sections
	manual     = {}
	section    = []
	title      = []
	pattern    = []
	topSection = None
	imgfolders = [	os.path.join(tkFlair.prgDir, "images"),
			os.path.join(tkFlair.prgDir, "icons")]

	# ----------------------------------------------------------------------
	def __init__(self, path):
		if not Manual.manual:
			if isinstance(path,str):
				Manual.load(Manual.findManual(path))
			else:
				for d in path:
					filename = Manual.findManual(d)
					if filename:
						Manual.load(filename)
		self.views   = []
		self.curview = 0
		self._move2End = False

	# ----------------------------------------------------------------------
	# show manual window
	# ----------------------------------------------------------------------
	def show(self, master=None):
		global current

		if master is not None:
			self.toplevel = master
		else:
			# Start a new interpreter (not Toplevel),
			# to have the manual page responsive at all times
			self.toplevel = tk.Tk(className="flair_manual")

		self.toplevel.title("flair manual")
		try:
			image = images.get("info.gif")
			if image is None:
				image = self.loadImage("info.gif")
			self.toplevel.call('wm','iconphoto',self.toplevel._w, image)
		except tk.TclError:
			self.toplevel.iconbitmap("info")	# standard bitmap

		section = tkFlair._FONT_SECTION
		self.defaultFont = tkFlair.config.getFont(section,
					"manual.default",
					("Sans",11), self.toplevel)
		self.fixedFont   = tkFlair.config.getFont(section,
					"manual.fixed",
					("Mono",11), self.toplevel)
		self.titleFont   = tkFlair.config.getFont(section,
					"manual.title",
					("Sans",16), self.toplevel)

		# Paned Window
		self.splitter = tkExtra.HSplitter(self.toplevel, 250, True)
		self.splitter.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
		tocFrame  = self.splitter.leftFrame()
		textFrame = self.splitter.rightFrame()

		# Search
		self.backwardBtn = tk.Button(tocFrame, text="\N{LEFTWARDS ARROW}",
					font=self.defaultFont,
					padx=2, pady=0,
					command=self.backward)
		self.backwardBtn.grid(row=0, column=0, sticky=tk.NSEW)
		self.forwardBtn = tk.Button(tocFrame, text="\N{RIGHTWARDS ARROW}",
					font=self.defaultFont,
					padx=2, pady=0,
					command=self.forward)
		self.forwardBtn.grid(row=0, column=1, sticky=tk.NSEW)
		self.entry = tkExtra.LabelEntry(tocFrame,
					label="Search",
					labelcolor="DarkGray",
					background="White",
					font=self.defaultFont,
					width=7)
		self.entry.grid(row=0, column=2, columnspan=2, sticky=tk.NSEW)
		self.entry.bind("<<LabelEntryClear>>", self.searchClear)

		self.tocList = tkExtra.ExListbox(tocFrame, borderwidth=0,
				selectmode=tk.BROWSE,
				selectborderwidth=0,
				background="White",
				font=self.defaultFont,
				takefocus=True,
				exportselection=tk.FALSE)
		self.tocList.grid(row=1, column=0, columnspan=3, sticky=tk.NSEW)

		sb = tk.Scrollbar(tocFrame, orient=tk.VERTICAL, takefocus=0,
				command=self.tocList.yview)
		sb.grid(row=1, column=3, sticky=tk.NSEW)
		self.tocList.config(yscrollcommand=sb.set)
		tocFrame.grid_rowconfigure(1, weight=1)
		tocFrame.grid_columnconfigure(2, weight=1)

		# Text Viewer
		self.txt = tk.Text(textFrame, background="White",
				font=self.fixedFont, takefocus=1)
		self.txt.config(wrap=tk.NONE, cursor="top_left_arrow")
		self.txt.grid(row=0, column=0, sticky='nsew')
		sbv = tkExtra.AutoScrollbar(textFrame, takefocus=0)
		sbv.grid(row=0, column=1, sticky='ns')
		sbh = tkExtra.AutoScrollbar(textFrame, takefocus=0,
				orient=tk.HORIZONTAL)
		sbh.grid(row=1, column=0, sticky='ew')

		self.txt.config(yscrollcommand=sbv.set, xscrollcommand=sbh.set)
		sbv.config(command=self.txt.yview)
		sbh.config(command=self.txt.xview)

		textFrame.grid_rowconfigure(0, weight=1)
		textFrame.grid_columnconfigure(0, weight=1)

		# Bindings
		self.txt.bind('<Button-1>',		self.textFocus)
		self.tocList.bind('<ButtonRelease-1>',	self.searchClick)
		self.tocList.bind('<Return>',		self.searchReturn)
		self.tocList.bind('<KP_Enter>',		self.searchReturn)
		self.toplevel.bind('<Control-Key-f>',	self.findDialog)
		self.toplevel.bind('<Control-Key-g>',	self.findNext)
		self.entry.bind('<Return>',		self.entryReturn)
		self.entry.bind('<KP_Enter>',		self.entryReturn)
		self.txt.bind('<Up>',	lambda e, s=self.txt:
						s.yview(tk.SCROLL, -1, tk.UNITS))
		self.txt.bind('<Down>',	lambda e, s=self.txt:
						s.yview(tk.SCROLL,  1, tk.UNITS))
		self.toplevel.bind("<F11>",		self.toggleTree)
		self.toplevel.bind("<F12>",		self.toggleHeight)
		self.toplevel.bind("<Escape>",		self.close)
		self.toplevel.bind('<Control-Key-w>',	self.close)
		self.toplevel.bind('<Control-Key-q>',	self.close)
		self.toplevel.protocol("WM_DELETE_WINDOW", self.close)

		self.tocList.bind('<Button-3>',		self.popupMenu)
		self.tocList.bind('<Control-Key-space>',self.popupMenu)
		self.tocList.bind('<<ListboxSelect>>',	self.searchAfter)
		self.txt.bind('<Key-slash>',		self.findDialog)
		self.txt.bind('<Button-3>',		self.popupMenu)
		self.txt.bind('<Control-Key-space>',	self.popupMenu)
		self.txt.bind('<Prior>',		self.prior)
		self.txt.bind('<Next>',			self.next)
		self.txt.bind('<Control-Prior>',	self.pageUp)
		self.txt.bind('<Control-Next>',		self.pageDown)

		# Fill TOC
		first = True
		for title in Manual.title:
			if not first and title[0:3] == "---":
				self.tocList.insert(tk.END,"")
			self.tocList.insert(tk.END, title)
			if title[0]!=" ":
				if title[0]=="-":
					self.tocList.itemconfig(tk.END,foreground="DarkRed")
				else:
					self.tocList.itemconfig(tk.END,foreground="DarkBlue")
			first = False

		# Initial values
		self.current_section = None
		self.searchStr       = None
		current              = self
		self.showSection(Manual.section[0])

		# Show window
		self.toplevel.deiconify()
		self.toplevel.focus_set()
		self.tocList.focus_set()
		self._oldHeight = 700
		if not self.toplevel.winfo_ismapped():
			self.toplevel.wait_visibility()
		self.toplevel.geometry("1024x800")
		self.splitter_pos = self.splitter.split
		self.inafter = False

	# ----------------------------------------------------------------------
	def close(self, event=None):
		global current
		if isinstance(self.toplevel, tk.Tk):
			# Destroy local icons...
			done()
		self.toplevel.destroy()
		current = None

	# ----------------------------------------------------------------------
	def textFocus(self, event):
		self.txt.focus_set()

	# ----------------------------------------------------------------------
	def toggleTree(self, event):
		if self.splitter.split > self.splitter.min:
			self.splitter.toggle()
			self.splitter.rightFrame().focus_set()
			self.toplevel.focus_set()
		else:
			self.splitter.toggle()

	# ----------------------------------------------------------------------
	def toggleHeight(self, event):
		self._oldHeight = tkExtra.toggleHeight(self.toplevel,
					self._oldHeight)

	# ----------------------------------------------------------------------
	def searchClear(self, event=None):
		self.search("")

	# ----------------------------------------------------------------------
	def backward(self, event=None):
		if self.curview>=len(self.views):
			self.curview -= 1
		if self.curview>0:
			self.curview -= 1
			section, searchstr = self.views[self.curview]
			if searchstr != self.entry.get():
				self.entry.delete(0,tk.END)
				self.entry.insert(0, searchstr)
				self.search(searchstr, False)
			self.showSection(section)

	# ----------------------------------------------------------------------
	def forward(self, event=None):
		if self.curview < len(self.views)-1:
			self.curview += 1
			section, searchstr = self.views[self.curview]
			if searchstr != self.entry.get():
				self.entry.delete(0,tk.END)
				self.entry.insert(0, searchstr)
				self.search(searchstr, False)
			self.showSection(section)

	# ----------------------------------------------------------------------
	# Create the popup menu
	# ----------------------------------------------------------------------
	def popupMenu(self, event):
		menu=tk.Menu(self.toplevel, tearoff=0)
		menu.add_command(label='Backward', underline=0,
				command=self.backward)
		menu.add_command(label='Forward', underline=0,
				command=self.forward)
		menu.add_command(label='Clear', underline=0,
				command=self.searchClear)
		menu.add_separator()
		menu.add_command(label='Find', underline=0,
				command=self.findDialog)
		menu.add_command(label='Find Next', underline=5,
				command=self.findNext)
		menu.add_separator()
		menu.add_command(label='Close',
				command=self.close)
		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	# ----------------------------------------------------------------------
	# Return hit on entry box
	# ----------------------------------------------------------------------
	def entryReturn(self, event=None):
		self.search(self.entry.get())

	# ----------------------------------------------------------------------
	# Search for an expression in help pages
	# ----------------------------------------------------------------------
	def search(self, expr, select=True):
		self.searchStr = expr.upper().split()
		self.current_section = None
		self.tocList.delete(0,tk.END)
		for section,title in zip(Manual.section,Manual.title):
			txt = Manual.manual[section].upper()
			found = True
			for w in self.searchStr:
				if txt.find(w) == -1:
					found = False
					break
			if found:
				self.tocList.insert(tk.END,title)
				if title[0]!=" ":
					if title[0]=="-":
						self.tocList.itemconfig(tk.END,foreground="DarkRed")
					else:
						self.tocList.itemconfig(tk.END,foreground="DarkBlue")
		if select and self.tocList.size() > 0:
			self.tocList.selection_clear(0,tk.END)
			self.tocList.selection_set(0)
			self.tocList.activate(0)
			self.tocList.focus_set()
			self.searchReturn()

	# ----------------------------------------------------------------------
	# button click
	# ----------------------------------------------------------------------
	def searchClick(self, event):
		self.tocList.focus_set()
		idx = self.tocList.nearest(event.y)
		if idx<0: return
		title = self.tocList.get(idx)
		if not title:
			idx  += 1
			title = self.tocList.get(idx)
		item    = Manual.title.index(title)
		section = Manual.section[item]
		self.showSection(section)

	# ----------------------------------------------------------------------
	def searchReturn(self, event=None):
		if self.tocList.size()==0: return
		sel = self.tocList.index(tk.ACTIVE)
		if sel<0: return
		self.tocList.selection_clear(0,tk.END)
		self.tocList.selection_set(sel)
		title = self.tocList.get(sel)
		if not title:
			self.tocList.selection_clear(sel)
			sel = int(sel) + 1
			self.tocList.selection_set(sel)
			title = self.tocList.get(sel)
		item    = Manual.title.index(title)
		section = Manual.section[item]
		self.showSection(section)
		if self._move2End:
			self.txt.see(tk.END)
			self._move2End = False

	# ----------------------------------------------------------------------
	def searchAfter(self, event=None):
		if self.inafter: return
		self.inafter = True
		self.toplevel.after(250, self.searchReturn)

	# ----------------------------------------------------------------------
	def prior(self, event=None):
		top,_ = self.txt.yview()
		if top < 0.001:
			self.pageUp()
			return "break"

	# ----------------------------------------------------------------------
	def next(self, event=None):
		top, bot = self.txt.yview()
		if bot > 1.0 - (bot-top)/5.0:
			self.pageDown()
			return "break"

	# ----------------------------------------------------------------------
	def pageUp(self, event=None):
		active = self.tocList.index(tk.ACTIVE)
		if active==0: return
		try: active = int(active)-1
		except: pass
		self.tocList.activate(active)
		self._move2End = True
		self.searchReturn()

	# ----------------------------------------------------------------------
	def pageDown(self, event=None):
		active = self.tocList.index(tk.ACTIVE)
		try: active = int(active)+1
		except: pass
		self.tocList.activate(active)
		self.searchReturn()

	# ----------------------------------------------------------------------
	def loadImage(self, imageName):
		# Search the following directories
		# 1. current manual location
		# 2. image folder of flair
		# 3. icons folder of flair
		# Try to load from icons directory
		if imageName.find('/') != -1:
			filename = imageName
		else:
			for folder in Manual.imgfolders:
				filename = os.path.join(folder, imageName)
				if os.path.isfile(filename):
					break
			else:
				log.warning(f"Image '{filename}' not found")
				return None
		try:
			image = tk.PhotoImage(master=self.toplevel, file=filename)
		except tk.TclError:
			log.warning(f"Cannot open '{filename}'")
			return None
		images[imageName] = image
		return image

	# ----------------------------------------------------------------------
	# show section
	# ----------------------------------------------------------------------
	def showSection(self, section=None, err=True):
		self.inafter = False
		if section is None:
			section = Manual.topSection

		try:
			idx = Manual.section.index(section)
		except:
			# Try to find the best match
			idx = 0
			su = section[:8].upper()
			for s in Manual.section:
				if su == s[:8].upper():
					section = s
					break
				idx += 1

		# Search item in list to highlight
		try: title=Manual.title[idx]
		except IndexError:
			if err:
				try:
					messagebox.showwarning("Section not found",
							f"Section {section} not found",
							parent=self.toplevel)
				except tk.TclError:
					pass
			return False
		for i in range(self.tocList.size()):
			item = self.tocList.get(i)
			if item == title or item.strip() == section:
				self.tocList.selection_clear(0,tk.END)
				self.tocList.selection_set(i)
				self.tocList.activate(i)
				self.tocList.see(i)
				break

		if self.current_section == section: return True
		self.current_section = section

		# Add to view history
		if self.curview == 0:
			self.backwardBtn.config(state=tk.DISABLED)
		else:
			self.backwardBtn.config(state=tk.NORMAL)
		if self.curview >= len(self.views):
			self.curview += 1
			self.views.append((section, self.entry.get()))
			if len(self.views) > _MAX_VIEWS:
				self.views.pop(0)
				self.curview -= 1
		else:
			if self.views[self.curview][0] != section:
				# delete the rest and append the new view
				self.curview += 1
				del self.views[self.curview:]
				self.views.append((section, self.entry.get()))
		if self.curview >= len(self.views)-1:
			self.forwardBtn.config(state=tk.DISABLED)
		else:
			self.forwardBtn.config(state=tk.NORMAL)

		txt = self.txt
		txt.config(state=tk.NORMAL)
		txt.delete(0.0, tk.END)
		text = Manual.manual.get(section,"")
		txt.insert(0.0, text)

		# Highlight items
		txt.tag_delete("title")
		txt.tag_add("title", "1.0", "1.end")
		txt.tag_configure("title", foreground="DarkRed", font=self.titleFont)
		txt.update_idletasks()

		# WHAT's & SDUM
		txt.tag_delete("what")
		for m in _patwhat.finditer(text):
			txt.tag_add("what",
					f"1.0 + {m.start()} chars",
					f"1.0 + {m.end()} chars")
		for m in _patsdum.finditer(text):
			txt.tag_add("what",
					f"1.0 + {m.start()} chars",
					f"1.0 + {m.end()} chars")
		txt.tag_configure("what", foreground="DarkMagenta")

		# Hyper links
		txt.tag_delete("card")
		tag = "card"
		for i in range(len(Manual.section)-1,-1,-1):
			s = Manual.section[i]
			if s == section: continue
			pat1, pat2 = Manual.pattern[i]
			for m in pat1.finditer(text):
				txt.tag_add(tag,
						f"1.0 + {m.start()} chars",
						f"1.0 + {m.end()} chars")

			if pat2 is not None:
				for m in pat2.finditer(text):
					txt.tag_add(tag,
							f"1.0 + {m.start()} chars",
							f"1.0 + {m.end()} chars")
		txt.tag_remove(tag, "1.0", "2.0")

		# Highlight current section
		try:
			i = Manual.section.index(section)
			tag = "me"
			pat1, pat2 = Manual.pattern[i]
			for m in pat1.finditer(text):
				txt.tag_add(tag,
						f"1.0 + {m.start()} chars",
						f"1.0 + {m.end()} chars")

			if pat2 is not None:
				for m in pat2.finditer(text):
					txt.tag_add(tag,
						f"1.0 + {m.start()} chars",
						f"1.0 + {m.end()} chars")
			txt.tag_configure("me", foreground="DarkBlue")
		except ValueError:
			pass
			#if section == Manual.section[i]:
			#	tag = "me"
			#else:
		txt.tag_configure("card",
				foreground = "Blue",
				underline  = 1)
		txt.tag_bind("card", "<Button-1>", self.hyperlink)

		# Highlight search string
		if self.searchStr is not None:
			txt.tag_delete("find")
			firstIndex = None
			for target in self.searchStr:
				index = "1.0"
				while True:
					index = txt.search(target, index,
							nocase=1, stopindex=tk.END)
					if not index: break
					endex = '%s + %d chars' % (index, len(target))
					txt.tag_add("find", index, endex)
					txt.update_idletasks()
					if firstIndex is None:
						firstIndex = index
					index = endex
			if firstIndex is not None:
				txt.see(firstIndex)
			txt.tag_configure("find",
					background = '#800000',
					foreground = '#FFFFFF')

		# Find&insert all image tags
		index = "0.0"
		while True:
			index = txt.search("{img:", index)
			if not index: break
			indexEnd = txt.search("}", index)
			imageName = txt.get(index, indexEnd)[5:]
			txt.delete(index, "%s + 1 char"%(indexEnd))
			image = images.get(imageName)
			if image is None:
				image = self.loadImage(imageName)

			if image is not None:
				try:
					txt.image_create(index, image=image)
				except tk.TclError:
					log.error(f"displaying image '{imageName}'")

		# html hyper links
		index = "0.0"
		while True:
			idx = index
			index  = txt.search("http://", idx, stopindex=tk.END)
			index2 = txt.search("https://", idx, stopindex=tk.END)
			if not index: index = index2
			if not index: break
			if index2 and txt.compare(index2, '<', index): index = index2
			line = index.split(".")[0]
			indexEnd = txt.search(" ", index, stopindex=line+".end")
			if not indexEnd: indexEnd = line+".end"
			txt.tag_add("url", index, indexEnd)
			index = indexEnd
		txt.tag_configure("url",
				foreground = "Blue",
				underline  = 1)
		txt.tag_bind("url", "<Button-1>", self.urllink)

		# No more editing is allowed
		txt.config(state=tk.DISABLED)
		return True

	# --------------------------------------------------------------------
	# Write HTML
	# --------------------------------------------------------------------
	def writeHTML(self, section, outdir, opt):
		if section is None:
			section = Manual.topSection

		#---------------------------------------------------------------
		# Substitute invalid characters in the filename
		#---------------------------------------------------------------
		def filename(name):
			return re.sub("[\\[\\]:]","_",name)

		try:
			idx = Manual.section.index(section)
		except:
			# Try to find the best match
			idx = 0
			su = section[:8].upper()
			for s in Manual.section:
				if su == s[:8].upper():
					section = s
					break
				idx += 1

		# Search item in list to highlight
		try:
			title=Manual.title[idx].strip()
		except IndexError:
			log.error(f"Section {section} not found\n")
			return False

		# Add to view history
		label = "&#9664; Previous"
		if idx==0:
			backward = f"<font color='gray'>{label}</font>"
		else:
			backward = f"<a href='{filename(Manual.section[idx-1])}.html'>{label}</a>"

		label = "Next &#9654;"
		if idx>=len(Manual.section)-1:
			foreward = f"<font color='gray'>{label}</font>"
		else:
			foreward = f"<a href='{filename(Manual.section[idx+1])}.html'>{label}</a>"

		top = f"<a href='{filename(Manual.section[1])}.html'>&#9651; Index</a>"

		text = Manual.manual.get(section,"")
		lf = text.find("\n")
		if lf>0:
			header = f"<h2>{text[:lf].strip()}</h2>\n"
			text = text[lf+1:]
		else:
			header = ""

		# Special characters
		text = re.sub(r"&",'&amp;', text)
		text = re.sub(r"<",'&lt;',  text)
		text = re.sub(r">",'&gt;',  text)

		# WHAT's & SDUM
		text = _patwhat.sub(r'<div id="flair_what">\g<0></div>', text)
		text = _patsdum.sub(r'<div id="flair_what">\g<0></div>', text)
		text = _patcomment.sub(r'<div id="flair_comment">\g<0></div>', text)

		text = re.sub(r"\bDefault\b",r'<div id="flair_default">\g<0></div>', text)
		text = re.sub(r"\bHint:",   r'<div id="flair_warning">\g<0></div>', text)
		text = re.sub(r"\bAdvice:", r'<div id="flair_warning">\g<0></div>', text)
		text = re.sub(r"\bNote:",   r'<div id="flair_default">\g<0></div>', text)
		text = re.sub(r"\bNotes:",  r'<div id="flair_default">\g<0></div>', text)
		text = re.sub(r"\bExample:",r'<div id="flair_default">\g<0></div>', text)
		text = re.sub(r"\bExample \d:",r'<div id="flair_default">\g<0></div>', text)
		text = re.sub(r"\bWARNING:",r'<div id="flair_warning">\g<0></div>', text)

		text = re.sub("1st", "1<sup>st</sup>", text)
		text = re.sub("2nd", "2<sup>nd</sup>", text)
		text = re.sub("3rd", "3<sup>rd</sup>", text)
		text = re.sub("4th", "4<sup>th</sup>", text)
		text = re.sub(r"{img:(.*?)}",r'<img class="flair" src="\1" alt="\1"/>', text)
		text = f"<pre>{text}</pre>"

		# Hyper links
		for i in range(len(Manual.section)-1,-1,-1):
			s = Manual.section[i]
			pat1, pat2 = Manual.pattern[i]
			#if s == section: continue

			# skip html syntax
			iter0 = _pathtmltag.finditer(text)

			# loop both iterators
			iter1 = pat1.finditer(text)
			if pat2:
				iter2 = pat2.finditer(text)
			else:
				iter2 = None

			# Loop over all matches in the string
			try:    m0 = next(iter0)
			except: m0 = None
			try:    m1 = next(iter1)
			except: m1 = None
			try:    m2 = next(iter2)
			except: m2 = None

			out = ""
			last = 0
			while m1 is not None or m2 is not None:
				if (m1 and m2 and m1.start()<m2.start()) or m2 is None:
					if m0 is None or m1.start() < m0.start():
						out += "%s<a href='%s.html'>%s</a>" % \
							(text[last:m1.start()],
							 filename(m1.group(1)),
							 m1.group(0))
						last = m1.end()
						try:    m1 = next(iter1)
						except: m1 = None
					else:
						out += text[last:m0.end()]
						last = m0.end()
						while m1 and m1.start()<m0.end():
							try:    m1 = iter1.next()
							except: m1 = None
						try:    m0 = next(iter0)
						except: m0 = None
				else:
					if m0 is None or m2.start() < m0.start():
						out += "%s<a href='%s.html'>%s</a>" % \
							(text[last:m2.start()],
							 filename(m2.group(1)),
							 m2.group(0))
						last = m2.end()
						try:    m2 = next(iter2)
						except: m2 = None
					else:
						out += text[last:m0.end()]
						last = m0.end()
						while m2 and m2.start()<m0.end():
							try:    m2 = iter2.next()
							except: m2 = None
						try:    m0 = next(iter0)
						except: m0 = None
			text = out + text[last:]

		# replace all https?://url.... to <a href="http://...">http://...</a>
		text = re.sub(r'(https?://[A-Z-a-z0-9.&-=/_]*)', r'<a href="\1">\1</a>', text)

		# write html
		log.say(f"Writing: {filename(section)}.html")
		#try:
		data = opt["template"]
		#except:
		#	data = default_template

		# Subsitute targets
		data = data.replace("{title}",    title)
		data = data.replace("{previous}", backward)
		data = data.replace("{index}",    top)
		data = data.replace("{next}",     foreward)
		data = data.replace("{header}",   header)

		if "{toc}" in data or idx==1:
			# create toc if needed
			prev = 0
			toc = ''
			prefix = ""
			for s,title in zip(Manual.section, Manual.title):
				# Count front spaces
				for sp,ch in enumerate(title):
					if ch!=' ': break
				sp /= 3
				while sp>prev:
					prev += 1
					prefix = "&nbsp;"*(prev*3)
				while sp<prev:
					#toc += '</ul>\n'
					prev -= 1
					prefix = "&nbsp;"*(prev*3)
				if section==s:
					toc += f"<li id='toc_current'><a href='{filename(s)}.html'>"\
						f"{prefix}&blacktriangleright;{title.strip()}</a></li>\n"
				else:
					toc += f"<li><a href='{filename(s)}.html'>"\
						f"{prefix}{title.strip()}</a></li>\n"
			data = data.replace("{toc}", toc)

			if idx==1:
				text += '\n<h2>Complete Index</h2>\n'
				text += '<ul style="list-style:none">\n'
				text += toc
				text += '</ul>\n'

		data = data.replace("{page}", text)

		f = open(os.path.join(outdir,filename(section)+".html"), "w")
		f.write(data)
		f.close()

		return True

	# --------------------------------------------------------------------
	# FindDialog
	# --------------------------------------------------------------------
	def findDialog(self, event=None):
		txt = self.entry.get()
		fd = tkDialogs.FindReplaceDialog(self.toplevel, replace=False)
		fd.show(self.find, None, None, txt)
		try:
			self.txt.tag_delete("find")
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	# Find target(s)
	# ----------------------------------------------------------------------
	def find(self, findStr=None, matchCase=None):
		if findStr is not None:
			self.searchStr = [findStr]
		else:
			if self.searchStr is None: return

		minrow = 100000
		mincol = 100000
		minlen = 0

		self.txt.tag_delete("find")
		for target in self.searchStr:
			index = str(self.txt.search(target,
					"insert + 1 chars",
					nocase = matchCase))
			if not index: continue
			row,col = map(int,index.split('.'))
			if row<minrow or (row==minrow and col<mincol):
				minrow = row
				mincol = col
				minlen = len(target)

		if minrow < 100000:
			index = "%d.%d" % (minrow,mincol)
			self.txt.mark_set('insert', index)
			self.txt.tag_add("find", 'insert', "insert + %d chars"%(minlen))
			self.txt.see(index)
			self.txt.tag_configure("find",
					background = '#800000',
					foreground = '#FFFFFF')

	# ----------------------------------------------------------------------
	def findNext(self, event=None):
		self.find()

	# ----------------------------------------------------------------------
	# hyper link
	# ----------------------------------------------------------------------
	def hyperlink(self, event):
		row,col = map(int, self.txt.index("@%d,%d"%(event.x,event.y)).split("."))

		# Find line being clicked
		lines = Manual.manual[self.current_section].splitlines()
		line = lines[row-1]
		del lines

		# Find word
		if line[col]=='}': col -= 1
		begin = col
		while begin>=0:
			ch = line[begin]
			if ch.isalnum() or ch in "-.:[]":
				begin -= 1
			else:
				break
		begin += 1

		end = col
		while end<len(line):
			ch = line[end]
			if ch.isalnum() or ch in "-.:[]":
				end += 1
			else:
				break

		section = line[begin:end]
		if section[-1]==".": section = section[:-1]

		try:
			item = Manual.section.index(section)
		except:
			# Try to find the best match
			item = 0
			su = section[:8].upper()
			for s in Manual.section:
				if rexx.abbrev(su, s[:8].upper(), 5):
					section = s
					break
				item += 1

		self.tocList.selection_clear(0,tk.END)
		self.tocList.selection_set(item)
		self.tocList.activate(item)
		self.searchStr = None
		self.showSection(section)

	# ----------------------------------------------------------------------
	def urllink(self, event=None):
		idx = self.txt.index("@%d,%d"%(event.x,event.y))
		line = idx.split(".")[0]
		index  = self.txt.search("http://", idx, backwards=True, stopindex=line+".0")
		index2 = self.txt.search("https://", idx, backwards=True, stopindex=line+".0")
		if not index: index = index2
		if not index: return
		if index2 and self.txt.compare(index2, ">", index):
			index = index2
		indexEnd = self.txt.search(" ", index, stopindex=line+".end")
		if not indexEnd: indexEnd = line+".end"
		tkFlair.openurl( self.txt.get(index, indexEnd).strip() )

	# ----------------------------------------------------------------------
	# load manual and split it into sections
	# ----------------------------------------------------------------------
	@staticmethod
	def load(filename):
		try:
			fmanual = open(filename, mode="r", encoding="utf-8")
		except IOError as err:
			messagebox.showerror("Error", err)
			return
		Manual.imgfolders.insert(0, os.path.dirname(os.path.abspath(filename)))
		fn,_ = os.path.splitext(os.path.basename(filename))
		index   = True
		txt     = ""
		section = fn
		if Manual.topSection is None:
			Manual.topSection = section
		firstline = "----- %s -----" % (fn)
		inGeo     = False
		depth     = 1
		for line in fmanual:
			if section=="":
				if index:
					section   = "INDEX-%s"%(fn)
					firstline = "Index %s"%(fn)
					index     = False
					txt       = ""
					continue

				line = line.strip()
				if line=="": continue
				for pat in _patterns:
					m = pat.match(line)
					if m is not None:
						section = m.group(1)
						if line[0]=="{":
							line = line.replace("{","")
							line = line.replace("}","")
							firstline = "      "+line
						else:
							firstline = line
						txt = line+"\n"
						inGeo = (section=="8")
						break

			elif line.startswith("1******************"):
				Manual._addSection(section, firstline, txt)
				section = ""

			elif line == "1\n":
				txt += "\n"
			else:
				# Check for a sub section
				for pat in _subpatterns:
					m = pat.match(line)
					if m is not None:
						if inGeo and _patbody2.match(line):
							depth = m.group(1).count(".")
							break
						Manual._addSection(section, firstline, txt)
						section   = m.group(1)
						depth     = section.count(".")
						firstline = ("   "*depth)+line.strip()
						txt = ""
						break

				if inGeo:
					m = _patbody1.match(line)
					if m is None:
						m = _patbody2.match(line)
					if m:
						Manual._addSection(section, firstline, txt)
						section   = m.group(1).replace(","," ")
						section   = section.replace("or","").split()
						firstline = ("   "*depth)+m.group(1)
						txt       = ""

					elif line==" Region data\n":
						Manual._addSection(section, firstline, txt)
						section   = "REGION"
						firstline = line.rstrip()
						txt       = ""

					elif line==" Region Volumes\n":
						Manual._addSection(section, firstline, txt)
						section   = line.rstrip()
						firstline = section
						txt       = ""

					elif line==" LATTICE card\n":
						Manual._addSection(section, firstline, txt)
						section   = "LATTICE"
						firstline = line.rstrip()
						txt       = ""

					elif line==" GEOEND card\n":
						Manual._addSection(section, firstline, txt)
						section   = "GEOEND card"
						firstline = "  "+section
						txt       = ""

					elif line=="1                             Voxel Geometry\n":
						Manual._addSection(section, firstline, txt)
						section   = "Voxel"
						firstline = " Voxel Geometry"
						txt       = ""
						line      = firstline+"\n"

				# Convert characters
				for src, dst in _change:
					line = line.replace(src,dst)
				txt += line
		if section!="": Manual._addSection(section, firstline, txt)
		fmanual.close()

	# ----------------------------------------------------------------------
	@staticmethod
	def _addSection(section, firstline, txt):
		if isinstance(section, list):
			# Special case for bodies
			for s in section:
				Manual.manual[s] = txt
				Manual.section.append(s)
				Manual.title.append("         "+s)
				pat1 = re.compile(r"\b(%s)\b"%(s[:8]))
				Manual.pattern.append((pat1, None))
			#section = section[0]
		else:
			section = section.strip()
			if not Manual.manual:
				Manual.manual["ABOUT"] = txt
			Manual.manual[section] = txt

			Manual.section.append(section)
			Manual.title.append(firstline)
			ch = section[0]
			target = section.replace(".",r"\.")
			target = target.replace("[",r"\[")
			target = target.replace("]",r"\]")
			if 'A'<=ch<='Z':
				pat1 = re.compile(r"\b(%s)\b"%(section[:8]))
				f = firstline.split()[0]
				if f != section[:8]:
					pat2 = re.compile(r"\b(%s)%s\b"%(f[:8],f[8:]))
				else:
					pat2 = None
			elif ch in ":[":
				pat1 = re.compile(f"({target})")
				pat2 = None
			else:
				pat1 = re.compile(f"\\b({target})}}")
				pat2 = None
			Manual.pattern.append((pat1, pat2))

	# ----------------------------------------------------------------------
	# Find latest manual file in the FLUKA directory
	# ----------------------------------------------------------------------
	@staticmethod
	def findManual(path):
		if not path:
			return None
		elif os.path.isdir(path):
			path = os.path.abspath(path)
			maxt   = 0
			manual_file = None
			for f in glob.glob(f"{path}/*.manual"):
				try:
					t = os.path.getmtime(os.path.join(path,f))
					if t > maxt: # get the newer only
						manual_file = f
						maxt = t
				except:
					pass
			return manual_file
		elif os.path.isfile(path):
			return os.path.abspath(path)

#-------------------------------------------------------------------------------
def flukaDocumentDir():
	flukadir = tkFlair.config.get("Fluka","directory")
	if not flukadir:
		path = tool.findExe("fluka")
		if path: flukadir = os.path.dirname(path)
		if flukadir == "" or flukadir is None:
			flukadir="."
	return os.path.join(flukadir,"doc")

#-------------------------------------------------------------------------------
# Show keyword
#-------------------------------------------------------------------------------
def show(keyword=None):
	global _directories, current

	if current is not None:
		current.searchClear()
		#current.showSection(keyword)
		current.toplevel.deiconify()
		current.toplevel.lift()
		current.toplevel.focus_set()
		manual = current
	else:
		if _directories is None:
			_directories = (".", flukaDocumentDir())
		manual = Manual(_directories)
		try:
			manual.show()
		except IndexError:
			messagebox.showwarning(
					"Manual error",
					"Error opening manual")

	if isinstance(keyword,list):
		for k in keyword:
			if manual.showSection(k,False): break
	else:
		manual.showSection(keyword)

#-------------------------------------------------------------------------------
def destroy():
	if current is not None:
		current.close()

#-------------------------------------------------------------------------------
# Initialize Manual by settings the default search path
#-------------------------------------------------------------------------------
def init(dirs):
	global _directories
	_directories = dirs

#-------------------------------------------------------------------------------
def done():
	global images
	for i in images.values():
		del i
	images = {}

#-------------------------------------------------------------------------------
def writeHTML(path, outdir, opt):
	manual = Manual(path)
	for s in manual.section:
		manual.writeHTML(s, outdir, opt)

#-------------------------------------------------------------------------------
def usage():
	print()
	print("Usage: Manual.py [options]")
	print()
	print("Options:")
	print("\t-?\t--help\tShow this help page")
	print("\t-h\t--html\tCreate html pages from manual")
	print("\t-m\t--manual #\tSpecify manual to load or path to find manual to load")
	print("\t-o\t--output #\tSpecify output directory to write html files (default .)")
	print("\t-t\t--template #\tHTML template for output files")
	print("\t\t--css #\t\tspecify css file")
	sys.exit(0)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	tkFlair.openIni()

	html       = False
	manualpath = None
	htmlopt    = {}
	outdir     = "."

	if len(sys.argv)>1:
		try:
			optlist, args = getopt.getopt(sys.argv[1:],
					"?hm:t:o:",
					["help",
					 "html", "manual=", "output=",
					 "template="])
		except getopt.GetoptError:
			usage()

		for opt, val in optlist:
			if opt in ("-?", "--help"):
				usage()
			elif opt in ("-h", "--html"):
				html = True
			elif opt in ("-m", "--manual"):
				manualpath = val
			elif opt in ("-o", "--output"):
				outdir = val
			elif opt in ("-t", "--template"):
				try:
					with open(val,"r") as f:
						htmlopt["template"] = f.read()
				except:
					print("Error: opening template file",val)
			else:
				print("Error: unknown argument",opt,val,file=sys.stderr)
				usage()

	if html:
		if manualpath is None:
			print("Error: manual location -m not specified")
			usage()
		if "template" not in htmlopt:
			print("Error: html template not provided")
			usage()
		writeHTML(manualpath, outdir, htmlopt)
	else:
		if manualpath is None:
			manualpath = (".", flukaDocumentDir())
		window = Manual(manualpath)
		root = tk.Tk(className="flair_manual")
		root.geometry("1024x800")
		window.show(root)
		root.mainloop()
		done()
