# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Dec-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math

import csg
import bmath
#import Global

import log
from deck import Card
from fluka import Fluka, FlukaCard

#-------------------------------------------------------------------------------
# Convert FLUKA natural material to isotopic composition
#-------------------------------------------------------------------------------
_fluka2mcnp = {
	 1:  1001,
	 2:  2004,
	 3:((3007,0.925), (3006,0.075)),
	 4:  4009,
	 5:((5011,0.801), (5010,0.199)),
	 6:((6012,0.989), (6013,0.011)),
	 7:((7014,0.996), (7015,0.004)),
	 8:  8016,
	 9:  9019,
	11: 11023,
	13: 13027,
	15: 15031,
	25: 25055,
	27: 27059,
	33: 33075,
	35:((35079,0.507), (35081,0.493)),
	36:((36084,0.570), (36078,0.003), (36080,0.023),
	    (36082,0.116), (36083,0.115), (36086,0.173)),
	38: 38090,
	41: 41093,
	43: 43099,
	53: 53127,
	55: 55133,
	56: 56138,
	62: ((62149,0.2883), (62150,0.1537), (62152,0.558)),
	73: 73181,
	74: 74186,
	75:((75187,0.626), (75185,0.374)),
	79: 79197,
	82:((82208,0.524), (82207,0.221), (82206,0.241), (82204, 0.014)),
	83: 83209,
	90: 90232,
	92:((92235,0.00726), (92238,0.99274)),
	93: 93237,
}

PAT_CELL     = re.compile(r"(.*?)#(\d+)(.*)", re.DOTALL | re.IGNORECASE)
PAT_RMCELL   = re.compile(r"#\d+", re.DOTALL | re.IGNORECASE)
PAT_CELLIMP  = re.compile(r"IMP_(.)=", re.IGNORECASE)
PAT_TRANS    = re.compile(r"^(\*?)TR(\d+)(\*?) (.*)", re.DOTALL | re.IGNORECASE)
PAT_MATERIAL = re.compile(r"^(M\d+) (.*)", re.DOTALL | re.IGNORECASE)
PAT_IMP      = re.compile(r"^IMP:(.) (.*)", re.DOTALL | re.IGNORECASE)
PAT_SURFTRCL = re.compile(r"^(S\d+)(C\d+)$")

PAT_nR       = re.compile(r"^(.*?\s)(\S+)\s*\b(\d+)R\b(.*)", re.DOTALL | re.IGNORECASE)
PAT_nI       = re.compile(r"^(.*?\s)(\S+)\s*\b(\d+)I\s*(\S*)\s*(.*)", re.DOTALL | re.IGNORECASE)
PAT_xM       = re.compile(r"^(.*?\s)(\S+)\s*\b(\d+)M\b(.*)", re.DOTALL | re.IGNORECASE)
PAT_J        = re.compile(r"^(.*?)\bJ\b(.*)", re.DOTALL | re.IGNORECASE)
PAT_nJ       = re.compile(r"^(.*?)\b(\d+)J\b(.*)", re.DOTALL | re.IGNORECASE)

EPS   = 1.0e-15

#-------------------------------------------------------------------------------
# Convert a vector to string
#-------------------------------------------------------------------------------
def vec(v): return f"{v[0]:.15g} {v[1]:.15g} {v[2]:.15g}"

#-------------------------------------------------------------------------------
# Convert matrix to string
# optionally specify position and vector
#-------------------------------------------------------------------------------
def transform(M, P=None, Z=None):
	if M is not None:
		if P is not None:
			P = M * P
		if Z is not None:
			Z = M.multNoTranslation(Z)
	elif P is None and Z is None:
		return None	 # M = P = Z = None

	if P is not None and Z is not None:
		return f"{vec(P)} 6j {vec(Z)}"
	elif P is not None:
		return vec(P)
	elif Z is not None:
		return f"9j {vec(Z)}"
	else:
		return	f"{M[0][3]:15g} {M[1][3]:15g} {M[2][3]:15g}" \
			f"{M[0][0]:15g} {M[1][0]:15g} {M[2][0]:15g}" \
			f"{M[0][1]:15g} {M[1][1]:15g} {M[2][1]:15g}" \
			f"{M[0][2]:15g} {M[1][2]:15g} {M[2][2]:15g}"

#===============================================================================
# Mcnp Importer/Exporter class
#===============================================================================
class Mcnp:
	SECTION_CELL = 0
	SECTION_SURF = 1
	SECTION_MAIN = 2
	SECTION_END  = 3

	#-----------------------------------------------------------------------
	# Mcnp constructor
	#-----------------------------------------------------------------------
	def __init__(self):
		self.title      = "-Untitled-"
		self.input      = None
		self.macroBodies= True
		self.init()

	#-----------------------------------------------------------------------
	def init(self):
		self.surfaces   = []	# List of Mcnp surfaces
		self.bodies     = {}	# Dictionary of FLUKA bodies
		self.transform  = []
		self.cells      = []
		self.regions    = []
		self.materials  = []
		self.importance = []
		self._transformCache = {}

	#-----------------------------------------------------------------------
	def setMacroBodies(self, flag=True):
		self.macroBodies = flag

	#-----------------------------------------------------------------------
	# Import FLUKA input
	#-----------------------------------------------------------------------
	def fromFluka(self, inp):
		self.input = inp
		self.init()
		try:
			card = inp.cards["TITLE"][0]
			self.title = card.text()
		except:
			pass

		rotdefi = self.input.cardProperties()
		self.input.preprocess()
		self._importSurfaces()
		self._importCells()
		self._importMaterials()
		#self._importImportance()

	#-----------------------------------------------------------------------
	# Read and convert MCNP file to FLUKA
	#-----------------------------------------------------------------------
	def toFluka(self, filename, inp):
		self.input = inp
		self.read(filename)

	#-----------------------------------------------------------------------
	# Import Surfaces
	#-----------------------------------------------------------------------
	def _importSurfaces(self):
		# Loop all cards
		for card in self.input.cardlist:
			if not card.isGeo(): continue
			if len(card.tag) != 3: continue
			if card.ignore(): continue

			# Add a new body [##, in, out, card]
			nsurf = len(self.surfaces)
			body = [0, "", "", card]
			self.bodies[card.name()] = body

			if card.tag == "RPP":
				self._importRPP(card, body)
			elif card.tag == "BOX":
				self._importBOX(card, body)
			elif card.tag == "SPH":
				self._addSphere(body,'-',card.bodyP(),card.bodyR(),
						card[Card.MATRIX], str(card))
			elif card.tag == "RCC":
				self._importRCC(card, body)
			elif card.tag == "REC":
				self._importREC(card, body)
			elif card.tag == "TRC":
				self._importTRC(card, body)
			elif card.tag == "ELL":
				self._importELL(card, body)
			elif card.tag == "ARB":
				self._importARB(card, body)
			elif card.tag == "WED":
				self._importWED(card, body)
			elif card.tag in ("XYP", "XZP", "YZP", "PLA"):
				self._addPlane(body,'-',card.bodyP(),card.bodyN(),
						card[Card.MATRIX], str(card))
			elif card.tag in ("XCC", "YCC", "ZCC"):
				self._importInfCylinder(card, body)
			elif card.tag in ("XEC", "YEC", "ZEC"):
				self._importInfEllipticalCylinder(card, body)
			elif card.tag == "QUA":
				self._importQUA(card, body)
			elif card.tag == "END":
				pass	# ignore END card
			else:
				raise Exception(f"MCNP: unknown FLUKA body type {card.tag}")

			body[0] = len(self.surfaces) - nsurf

	#-----------------------------------------------------------------------
	def _importRPP(self, card, body):
		M = card[Card.MATRIX]
		if self.macroBodies:
			self._addSurf(body,'-',"RPP %.15g %.15g %.15g %.15g %.15g %.15g"
				%(card.numWhat(1), card.numWhat(2), card.numWhat(3),
				  card.numWhat(4), card.numWhat(5), card.numWhat(6)),
				str(card), transform(M))
		else:
			P  = card.bodyP()
			Po = card.bodyPo()
			self._addPlane(body,'+',P, bmath.Vector.X, M, str(card))
			self._addPlane(body,'-',Po,bmath.Vector.X, M)
			self._addPlane(body,'+',P, bmath.Vector.Y, M)
			self._addPlane(body,'-',Po,bmath.Vector.Y, M)
			self._addPlane(body,'+',P, bmath.Vector.Z, M)
			self._addPlane(body,'-',Po,bmath.Vector.Z, M)

	#-----------------------------------------------------------------------
	def _importBOX(self, card, body):
		P = card.bodyP()
		X = card.bodyX()
		Y = card.bodyY()
		Z = card.bodyZ()
		M = card[Card.MATRIX]

		if self.macroBodies:
			self._addSurf(body,'-',f"BOX {vec(P)} {vec(X)} {vec(Y)} {vec(Z)}",
					str(card), transform(M))
		else:
			# Correct the order
			if (X.cross(Y)).dot(Z)<0:
				X, Y = Y, X		# inverse X with Y

			# outer corner
			Po = P + X + Y + Z
			X.norm()
			Y.norm()
			Z.norm()

			# first 3 planes will be
			self._addPlane(body,'+',P,X,M, str(card))
			self._addPlane(body,'+',P,Y,M)
			self._addPlane(body,'+',P,Z,M)

			# last 3 planes will be
			self._addPlane(body,'-',Po,X,M)
			self._addPlane(body,'-',Po,Y,M)
			self._addPlane(body,'-',Po,Z,M)

	#-----------------------------------------------------------------------
	def _importRCC(self, card, body):
		P = card.bodyP()
		Z = card.bodyZ()
		r = card.bodyR()
		M = card[Card.MATRIX]
		if self.macroBodies:
			self._addSurf(body,'-',f"RCC {vec(P)} {vec(Z)} {r:.15g}",
					str(card), transform(M))
		else:
			Z.norm()
			self._addSurf(body,'-',f"CZ {r:.15g}"%(),
					str(card),transform(M,P,Z))
			self._addPlane(body,'+',P,Z,M)
			self._addPlane(body,'-',card.bodyPo(),Z,M)

	#-----------------------------------------------------------------------
	def _importREC(self, card, body):
		P = card.bodyP()
		X = card.bodyX()
		Y = card.bodyY()
		Z = card.bodyZ()
		M = card[Card.MATRIX]

		if self.macroBodies:
			self._addSurf(body,'-',f"REC {vec(P)} {vec(Z)} {vec(X)} {vec(Y)}",
					str(card), transform(M))
		else:
			Ptop = P + Z

			radiusX = X.length()
			radiusY = Y.length()

			# correct system
			if (X.cross(Y)).dot(Z) < 0.0:
				X = -X	# from left-hand to right-hand system

			# normalize the coordinates
			Z = X.cross(Y)
			Y = Z.cross(X)

			X.norm()
			Y.norm()
			Z.norm()

			a = 1.0/(radiusX**2)
			b = 1.0/(radiusY**2)

			self._addSurf(body,'-',f"SQ {a:.15g} {b:.15g} 0 0 0 0 -1 0 0 0",
					str(card),"%s %s %s 3j"%(vec(P),vec(X),vec(Y)))
			self._addPlane(body,'+',P,Z,M)
			self._addPlane(body,'-',card.bodyPo(),Z,M)

	#-----------------------------------------------------------------------
	def _importTRC(self, card, body):
		P = card.bodyP()
		Z = card.bodyZ()
		M = card[Card.MATRIX]
		baseRadius = card.numWhat(7)
		apexRadius = card.numWhat(8)

		if self.macroBodies:
			self._addSurf(body,'-',
				f"TRC {vec(P)} {vec(Z)} {baseRadius:.15g} {apexRadius:.15g}",
				str(card), transform(M))
		else:
			Zlen = Z.length()
			Z.norm()
			s = (apexRadius-baseRadius)/Zlen
			if abs(s)<1e-10:
				# Normal cylinder
				self._addSurf(body,'-',f"CZ {baseRadius:.15g}",
						str(card), transform(M,P,Z))
				self._addPlane(body,'+',P,Z,M)
				self._addPlane(body,'-',card.bodyPo(),Z,M)
			else:
				self._addSurf(body,'-',f"KZ {-baseRadius/s:.15g} {s**2:.15g}",
						str(card), transform(M,P,Z))
				self._addPlane(body,'+',P,Z,M)
				self._addPlane(body,'-',card.bodyPo(),Z,M)

	#-----------------------------------------------------------------------
	def _importELL(self, card, body):
		P1 = card.bodyP1()
		P2 = card.bodyP2()
		M = card[Card.MATRIX]
		C  = 0.5 * (P1 + P2)
		a  = card.numWhat(7) / 2.0
		c  = (P1-C).length()
		b  = math.sqrt(a**2 - c**2)

		# Create axis system
		Z = (P1-P2)
		Z.norm()

		self._addSurf(body,'-',
				f"SQ {1.0/b**2:.15g} {1.0/b**2:.15g} {1.0/a**2:.15g} 0 0 0 -1 0 0 0",
				str(card), transform(M,C,Z))

	#-----------------------------------------------------------------------
	def _importQUA(self, card, body):
		self._addSurf(body,'-',
			"GQ %.15g %.15g %.15g %.15g %.15g %.15g %.15g %.15g %.15g %.15g"
			% (card.numWhat(1), card.numWhat(2), card.numWhat(3),
			   # Warning on the order xy yz zx
			   card.numWhat(4), card.numWhat(6), card.numWhat(5),
			   card.numWhat(7), card.numWhat(8), card.numWhat(9),
			   card.numWhat(10)),
			str(card), transform(M))

	#-----------------------------------------------------------------------
	def _importARB(self, card, body):
		M = card[Card.MATRIX]
		if self.macroBodies:
			self._addSurf(body,'-',
				"ARB %s %s %s %s %s %s %s %s %.15g %.15g %.15g %.15g %.15g %.15g"
				% (card.bodyPn(1), card.bodyPn(2), card.bodyPn(3),
				   card.bodyPn(4), card.bodyPn(5), card.bodyPn(6),
				   card.bodyPn(7), card.bodyPn(8),
				   card.intWhat(25), card.intWhat(26),
				   card.intWhat(27), card.intWhat(28),
				   card.intWhat(29), card.intWhat(30)),
				 str(card), transform(M))
		else:
			V = [None]*8
			for i in range(8):
				V[i] = card.bodyPn(i+1)
			#F = [None]*6
			s = str(card)
			for i in range(6):
				face = card.intWhat(i+25)
				if face == 0: continue
				v1,face = divmod(face, 1000)
				v2,face = divmod(face,  100)
				v3,v4   = divmod(face,   10)

				# use the first three to define the plane
				E1 = V[v2-1] - V[v1-1]
				E2 = V[v3-1] - V[v1-1]
				N = E1.cross(E2)
				N.norm()
				self._addPlane(body,'-',V[v1-1],N,M,s)
				s = None

	#-----------------------------------------------------------------------
	def _importWED(self, card, body):
		P = card.bodyP()
		X = card.bodyX()
		Y = card.bodyY()
		Z = card.bodyZ()
		M = card[Card.MATRIX]

		if self.macroBodies:
			self._addSurf(body,'-',f"WED {vec(P)} {vec(X)} {vec(Y)} {vec(Z)}",
					str(card), transform(M))
		else:
			# Correct the order
			if (X.cross(Y)).dot(Z)<0:
				X, Y = Y, X		# inverse X with Y

			XC = P + X
			YC = P + Y
			ZC = P + Z

			# normal of the plane
			N = (YC-XC).cross(ZC-XC)
			N.norm()

			# first 3 planes will be
			self._addPlane(body,'+',P,X,M,str(card))
			self._addPlane(body,'+',P,Y,M)
			self._addPlane(body,'+',P,Z,M)

			# cutting plane
			self._addPlane(body,'-',XC,N,M)

	#-----------------------------------------------------------------------
	def _importInfCylinder(self, card, body):
		u = card.numWhat(1)
		v = card.numWhat(2)
		r = card.numWhat(3)
		M = card[Card.MATRIX]
		axis = card.tag[0]

		if u==0.0 and v==0.0:
			self._addSurf(body,'-',f"C{axis} {r:.15g}",
					str(card), transform(M))
		else:
			if axis=="Y": u,v = v,u		# Reverse for Y
			self._addSurf(body,'-',f"C/{axis} {u:.15g} {v:.15g} {r:.15g}",
						str(card), transform(M))

	#-----------------------------------------------------------------------
	def _importInfEllipticalCylinder(self, card, body):
		P     = card.bodyP()
		a,b,c = card.bodyR()
		M     = card[Card.MATRIX]
		#axis  = card.tag[0]
		if a!=0.0: a = 1.0/a**2
		if b!=0.0: b = 1.0/b**2
		if c!=0.0: c = 1.0/c**2

		# FIXME it should not be correct!!!
		self._addSurf(body,'-', f"SQ {a:.15g} {b:.15g }{c:.15g }0 0 0 -1 {vec(P)}",
					str(card), transform(M))

	#-----------------------------------------------------------------------
	# add a plane and convert it if possible to best match
	#-----------------------------------------------------------------------
	def _addPlane(self, body, op, pos, normal, matrix, comment=""):
		normal.normalize()
		if matrix is not None:
			pos = matrix * pos
			normal = matrix.multNoTranslation(normal)

		dire = normal.direction()
		if dire == "X":
			self._addSurf(body, op, f"PX {pos.x:.15g}", comment)
		elif dire == "Y":
			self._addSurf(body, op, f"PY {pos.y:.15g}", comment)
		elif dire == "Z":
			self._addSurf(body, op, f"PZ {pos.z:.15g}", comment)
		elif dire == "-X":
			op = "-" if op=="+" else "+"
			self._addSurf(body, op, f"PX {pos.x:.15g}", comment)
		elif dire == "-Y":
			op = "-" if op=="+" else "+"
			self._addSurf(body, op, f"PY {pos.y:.15g}", comment)
		elif dire == "-Z":
			op = "-" if op=="+" else "+"
			self._addSurf(body, op, f"PZ {pos.z:.15g}", comment)
		else:
			self._addSurf(body, op, f"P {vec(normal)} {pos.dot(normal):.15g}",
					comment)

	#-----------------------------------------------------------------------
	# add sphere with the appropriate card
	#-----------------------------------------------------------------------
	def _addSphere(self, body, op, pos, r, matrix, comment=""):
		if matrix is not None:
			pos = matrix * pos
		x,y,z = pos
		small = bmath._accuracy

		if abs(x)<small and abs(y)<small and abs(z)<small:
			self._addSurf(body,'-',f"SO {r:.15g}",comment)
		elif abs(y)<small and abs(z)<small:
			self._addSurf(body,'-',f"SX {x:.15g} {r:.15g}",comment)
		elif abs(x)<small and abs(z)<small:
			self._addSurf(body,'-',f"SY {y:.15g} {r:.15g}",comment)
		elif abs(x)<small and abs(y)<small:
			self._addSurf(body,'-',f"SZ {z:.15g} {r:.15g}",comment)
		else:
			self._addSurf(body,'-',f"S {x:.15g} {y:.15g} {z:.15g} {r:.15g}",comment)

	#-----------------------------------------------------------------------
	# Add a surface to MCNP
	#-----------------------------------------------------------------------
	def _addSurf(self, body, op, surface, comment="", tr=None):
		nop = "-" if op=="+" else "+"
		n = len(self.surfaces)+1

		inside  = body[1]
		outside = body[2]
		if inside == "":
			inside  = "%s%d" % ( op, n)
			outside = "%s%d" % (nop, n)
		else:
			inside  += " %s%d" % ( op, n)
			outside += ":%s%d" % (nop, n)

		body[1] = inside
		body[2] = outside

		if tr is not None:
			try:
				tn = self._transformCache[tr]
			except KeyError:
				tn = len(self.transform)+1
				self._transformCache[tr] = tn
				self.transform.append(f"TR{tn} {tr}")
			surface = f"{n:5d} {tn} {surface}"
		else:
			surface = f"{n:5d} {surface}"
		self.surfaces.append((surface, comment))

	#-----------------------------------------------------------------------
	# Import Cells
	#-----------------------------------------------------------------------
	def _importCells(self):
		# Loop all region
		self.input.addCardProperty("REGION","BIASING","@importance",3,4,5,6)

		region = None
		expr   = ""
		for card in self.input["REGION"]:
			if card.ignore(): continue
			if card.name() == "&":
				# append to previous expression
				expr += card.expr()
				continue

			if region is None:
				# Remember starting region
				region = card
				expr   = region.expr()
				continue

			self.__importCell(region, expr)

			# Remember starting region
			region = card
			expr   = region.expr()

		# Last one
		if region: self.__importCell(region, expr)

		# Create universe
		# ... XXX assume is composed by only one body
		#self.cells.append(("%d 0 %s"%(len(self.cells)+1,body[2])
		#			,"Universe"))

	#-----------------------------------------------------------------------
	def __importCell(self, region, expr):
		# Find material
		material = region[Card.MATERIAL]
		if material is None:
			log.error(f"Region {region.name()} is not assigned material")
			return True

		density = material.numWhat(3)
		if density <= 0.0:
			cell_expr = "%d 0"%(region["@n"])	# create a void volume
		else:
			n = material["@n"]
			if n<0:
				log.warning(f"Predefined material {material.name()} "
					"is substituted by BLCKHOLE")
				n=1
			cell_expr = "%d %d %g"%(region["@n"], n, -density)

			# Add material to the list
			n -= 1
			if n >= len(self.materials):
				self.materials.extend([None] *
					(n-len(self.materials)+1))
			self.materials[n] = material

		# remove spaces
		region_expr = "".join(expr.split())
		if not region_expr: return True

		if region_expr.find("#")>=0 or region_expr.find("(")>=0:
			# Expand parenthesis
			log.warning(f"Region '{region.name()}' contains parentheses. Expanded!")
			try:
				exp = csg.tokenize(region_expr)
				csg.exp2rpn(exp)
				csg.rpnorm(exp)
				region_expr = csg.rpn2exp(exp)
			except csg.CSGException:
				log.error(f"Region '{region.name()}' contains "
					"too many parentheses. Unable to expanded!")
				region_expr = ""
		else:
			region_expr = region_expr.replace("+", " + ")
			region_expr = region_expr.replace("-", " - ")
			region_expr = region_expr.replace("|", " | ")
			region_expr = region_expr.split()

		op = "+"
		first = True
		# construct the MCNP region line
		for item in region_expr:
			if item in ("+", "-"):
				op = item
			elif item == "|":
				op = "+"
				if first:
					first = False
				else:
					cell_expr += ":"
			else:
				first = False
				body = self.bodies[item]
				n, inside, outside, bcard = body
				if op=="+":
					cell_expr += f" {inside}"
				else:
					if n>1:
						cell_expr += f" ({outside})"
					else:
						cell_expr += f" {outside}"
		self.cells.append((cell_expr, str(region)))

		if region[Card.MATERIAL].intWhat(4)==1:
			self.importance.append(0.0)
		else:
			biasing = region["@biasingN"]
			if biasing is None:
				biasing = region["@biasingA"]
			if biasing is not None:
				imp = biasing.numWhat(3)
			else:
				imp = 1.0
			self.importance.append(imp)
		return False

	#-----------------------------------------------------------------------
	# Import Materials that are in use
	# FIXME converts everything to atom fractions
	#       maybe it be nice to keep mass if needed
	#-----------------------------------------------------------------------
	def _importMaterials(self):
		for mat in self.materials:
			if mat is None: continue
			ZAID = {}

			def add(z, f, am):
				nonlocal ZAID
				try:
					ZAID[z][0] += f
				except (KeyError, IndexError):
					ZAID[z] = [f,am]

			for za,[f,am] in self.input.materialZAID(mat,False,-1).items():
				z,a = divmod(za,1000)
				if a==0:
					zaid = _fluka2mcnp.get(z)
				else:
					zaid = None
				if zaid is not None:
					if isinstance(zaid, int):
						add(zaid,f,am)
					else:
						for zz,ff in zaid:
							add(zz,ff*f,am)
				else:
					add(za,f,am)
			mat.zaid = list(ZAID.items())

	#-----------------------------------------------------------------------
	# Read an mcnp file and convert it to FLUKA
	#-----------------------------------------------------------------------
	def read(self, filename):
		# Initialize internal variables for reading
		self._density = {}			# material density combination
		self._trcl    = {}			# list of inlined TRCL
		self._section = Mcnp.SECTION_CELL	# section of file during reading

		fin = open(filename,"r")

		if self.input is None:
			self.input = Fluka()

		cardcomment = ""# comment of the card
		comment = ""	# comments at the end of the card which are meant for
				# the next one
				# first populate the last comment and if the card
				# continuous append it to the comment
		cmdline = ""

		def appendComment(comment, line):
			if comment:
				return comment + "\n" + line
			else:
				return line

		# Read title as first line
		self.__parseCommand(fin.readline().rstrip())

		# Read the remaining lines
		cont = False
		while True:
			line = fin.readline()
			if not line: break
			line = line.rstrip()

			# Check for comments
			if line and line[0] in ("c", "C"):
				comment = appendComment(comment, line[1:])
				continue
			dollar = line.find("$")
			if dollar>=0:
				comment = appendComment(comment, line[dollar+1:])
				line = line[:dollar].rstrip()

			# Empty line
			if not line:
				# and no comment then change section
				if dollar<0:
					self.__parseCommand(cmdline, cardcomment)
					self.__parseCommand(None)	# Advance section
				cardcomment = comment
				comment = ""
				cmdline = ""

			elif cont or line[:5]=="     " or line[0]=="\t":
				if line[-1]=="&":
					line = line[:-1].rstrip()
					cont = True
				else:
					cont = False
				# Continuation
				cmdline += line
				if comment:
					cardcomment = appendComment(cardcomment, comment)
					comment = ""

			else:
				if line[-1]=="&":
					line = line[:-1].rstrip()
					cont = True
				else:
					cont = False
				# Execute command
				if cmdline != "":
					self.__parseCommand(cmdline, cardcomment)
				cardcomment = comment
				comment = ""
				cmdline = line

		# execute last command
		self.__parseCommand(cmdline, comment)

		# Write region properties
		self.__regionProperties()

		del self._density
		del self._trcl

	#-----------------------------------------------------------------------
	def __regionProperties(self):
		# Create assignmat and optionally biasing
		for region in self.regions:
			name = region.name()
			# FIXME unique material if multiple densities are supplied!
			mat = region["@mat"]
			# biasing
			impn = region.get("@IMP:N",1.0)
			impp = region.get("@IMP:P",1.0)
			impe = region.get("@IMP:E",1.0)

			if mat == "0":
				if impn==0.0 and impp in (0.0, 1.0) and impe in (0.0, 1.0):
					mat = "BLCKHOLE"
				else:
					mat = "VACUUM"
			else:
				mat = "M"+mat

			self.input.add(FlukaCard("ASSIGNMA", ["", mat, name]))

			if impn not in (0.0, 1.0):
				self.input.add(FlukaCard("BIASING", ["", "3.0", "", impn, name]))
			if impe not in (0.0, 1.0):
				self.input.add(FlukaCard("BIASING", ["", "2.0", "", impe, name]))
			if impp not in (0.0, 1.0):
				self.input.add(FlukaCard("BIASING", ["", "2.0", "", impp, name]))

	#-----------------------------------------------------------------------
	# parse Mcnp command line
	#-----------------------------------------------------------------------
	def __parseCommand(self, command, comment=""):
		if command:
			command = self.__parseFeatures(command)

		if command is None:
			# Advance section
			self._section += 1
			if self._section == Mcnp.SECTION_END:
				self.input.add(FlukaCard("END"))
				self.__parseCells()
				self.input.add(FlukaCard("END"))
				self.input.add(FlukaCard("GEOEND"))

				# Dump inline trcl transformation
				for name,(pos,Rx,Ry,Rz,M) in self._trcl.items():
					self.__addTransformation(name, pos, Rx, Ry, Rz, M)

		# Cell parsing
		elif self._section == Mcnp.SECTION_CELL:
			self.input.add(FlukaCard("TITLE", [],comment,command))
			self._section += 1
			self.input.add(FlukaCard("GEOBEGIN",
				["COMBNAME"],"",command))

		elif self._section == Mcnp.SECTION_SURF:
			# add cells to be parsed latter
			self.cells.append((command.upper(), comment))

		elif self._section == Mcnp.SECTION_MAIN:
			# parse surfaces
			self.__parseSurface(command.upper(), comment)

		else:
			# Generic commands
			ucommand = command.upper()
			m = PAT_MATERIAL.match(ucommand)
			if m:
				self.__parseMaterial(m.group(1), m.group(2), comment)
				return

			m = PAT_TRANS.match(ucommand)
			if m:
				deg = m.group(1)=="*" or m.group(3)=="*"
				name = "rot%d"%(int(m.group(2)))
				try:
					tokens = list(map(float,m.group(4).split()))
				except:
					log.error(f"TR{m.group(2)} parsing transformation '{ucommand}'")
				pos, Rx,Ry,Rz, M = self.__parseTransformation(tokens, deg)
				self.__addTransformation(name, pos, Rx, Ry, Rz, M, comment)
				return

			m = PAT_IMP.match(ucommand)
			if m:
				var = f"@IMP:{m.group(1)}"
				for region,imp in zip(self.regions, map(float, m.group(2).split())):
					region[var] = imp

		#if comment: log.say("C",comment)
		#log.say(self._section, command)

	#-----------------------------------------------------------------------
	# nR: repeat the proceeding entry n times
	# nI: insert n linear interpolate entries
	# xM: multiply proceeding entry by x
	# nJ: replace by the default value (or jump) n entries (J alone is accepted)
	#-----------------------------------------------------------------------
	def __parseFeatures(self, command):
		# Parse nR
		while True:
			m = PAT_nR.match(command)
			if m is None: break
			command = m.group(1) \
				  + " ".join([m.group(2)]*(int(m.group(3))+1)) \
				  + m.group(4)

		# Parse nI
		while True:
			m = PAT_nI.match(command)
			if m is None: break
			try:
				command = m.group(1) + m.group(2)
				a = float(m.group(2))
				n = int(m.group(3))
				b = float(m.group(4))
				d = (b-a)/(n+1)
				x = a
				for _ in range(n):
					x += d
					command += " %g"%(x)
				command += " " + m.group(4) + m.group(5)
			except ValueError:
				break

		# Parse xM
		while True:
			m = PAT_xM.match(command)
			if m is None: break
			try:
				command = m.group(1) + m.group(2) + " " + \
					  str(float(m.group(2))*float(m.group(3))) + \
					  m.group(4)
			except ValueError:
				break

		# Parse single J
		while True:
			m = PAT_J.match(command)
			if m is None: break
			command = m.group(1) + "0" + m.group(2)

		# Parse nJ
		while True:
			m = PAT_nJ.match(command)
			if m is None: break
			command = m.group(1) + " ".join(["0"]*int(m.group(2))) + m.group(3)

		return command

	#-----------------------------------------------------------------------
	# Parse Surfaces to Bodies
	#
	# TODO
	#   * If a surface is preceded by an asterisk * then an reflective
	#     surface is created
	#
	#   * Cones K/XYZ the last variable, single +/-1 is not taken into account
	#
	#   * Surface cards of type X, Y, and Z may be used to describe surfaces
	#     by coordinate points rather than by equation coefficients as in the previous
	#     section. The surfaces described by these cards must be symmetric about the
	#     x-, y- or z-axis, respectively, and if the surface consists of more than one
	#     sheet, the specified coordinate points must all be on the same sheet.
	#     Each of the coordinate pairs defines a geometrical point on the surface.
	#
	#     On the Y card, for example, the entries may be
	#
	#          j  Y  y1 r1  y2 r2
	#
	#     where ri = sqrt(xi^2 + zi^2) and yi is the coordinate of point i.
	#
	#   - If one coordinate pair is used, a plane (PX, PY, or PZ) is defined.
	#
	#   - If two coordinate pairs are used, a linear surface
	#       (PX, PY, PZ, CX, CY, CZ, KX, KY or KZ) is defined.
	#
	#   - If three coordinate pairs are used, a quadratic surface
	#       (PX, PY, PZ, SO, SX, SY, SZ, CX, CY, CZ, KX, KY, KZ or SQ) is defined.
	#
	#   - When a cone is specified by two points, a cone of only one sheet is
	#     generated.
	#-----------------------------------------------------------------------
	def __parseSurface(self, command, comment):
		#if Global.verbose: log.say(">S>",command)
		token   = command.split()
		try:
			sid = token.pop(0)	# split into 2 instructions
			sid = int(sid)		# to catch & display a possible exception
		except ValueError:
			log.error(f"Invalid surface id {sid}")
			return False
		name = f"S{sid}"
		surface = token.pop(0)
		try:
			tr = int(surface)
			surface = token.pop(0)
		except ValueError:
			tr = 0
		arg = [float(x) for x in token]

		if surface=="P":
			# Plane
			tag = "PLA"
			if len(arg)==4:
				a,b,c,d = arg
				t = d/(a*a + b*b + c*c)
				what = [name, a, b, c, a*t, b*t, c*t]
			elif len(arg)==9:
				A = bmath.Vector(arg[0], arg[1], arg[2])
				B = bmath.Vector(arg[3], arg[4], arg[5])
				C = bmath.Vector(arg[6], arg[7], arg[8])
				n = (B-A) ^ (C-A)
				n.normalize()
				d = n*A
				# From MCNP4B manual 3-18
				if abs(d)<1e-10:				# D=0
					if abs(n[2])<1e-10 and abs(n[1])<1e-10: # D=C=B=0
						n = bmath.Vector.X
					elif abs(n[2])<1e-10:			# D=C=0
						n[1] = abs(n[1])
					else:					# D=0
						n[2] = abs(n[2])
				elif d < 0.0:
					n = -n
				what = [name, n[0], n[1], n[2], A[0], A[1], A[2]]
			else:
				log.error(f"Plane {sid} has invalid number of arguments")
				what = [name] + arg
		elif surface=="PX":
			# Plane normal to X-axis
			tag  = "YZP"
			what = [name, arg[0]]
		elif surface=="PY":
			# Plane normal to Y-axis
			tag  = "XZP"
			what = [name, arg[0]]
		elif surface=="PZ":
			# Plane normal to Z-axis
			tag  = "XYP"
			what = [name, arg[0]]

		#elif surface=="X":
		#	pass	# Not implemented yet see notes

		#elif surface=="Y":
		#	pass	# Not implemented yet see notes

		#elif surface=="Z":
		#	pass	# Not implemented yet see notes

		elif surface=="SO":
			# Sphere at Origin
			tag  = "SPH"
			what = [name, 0., 0., 0., arg[0]]
		elif surface=="S":
			# Sphere general
			tag  = "SPH"
			what = [name, arg[0], arg[1], arg[2], arg[3]]
		elif surface=="SX":
			# Sphere centered on X-axis
			tag  = "SPH"
			what = [name, arg[0], 0., 0., arg[1]]
		elif surface=="SY":
			# Sphere centered on Y-axis
			tag  = "SPH"
			what = [name, 0., arg[0], 0., arg[1]]
		elif surface=="SZ":
			# Sphere centered on Z-axis
			tag  = "SPH"
			what = [name, 0., 0., arg[0], arg[1]]

		elif surface=="C/X":
			# Cylinder parallel to X-axis
			tag  = "XCC"
			what = [name, arg[0], arg[1], arg[2]]
		elif surface=="C/Y":
			# Cylinder parallel to Y-axis
			tag  = "YCC"
			what = [name, arg[1], arg[0], arg[2]]
		elif surface=="C/Z":
			# Cylinder parallel to Z-axis
			tag  = "ZCC"
			what = [name, arg[0], arg[1], arg[2]]

		elif surface=="CX":
			# Cylinder on X-axis
			tag  = "XCC"
			what = [name, 0., 0., arg[0]]
		elif surface=="CY":
			# Cylinder on Y-axis
			tag  = "YCC"
			what = [name, 0., 0., arg[0]]
		elif surface=="CZ":
			# Cylinder on Z-axis
			tag  = "ZCC"
			what = [name, 0., 0., arg[0]]

		elif surface=="K/X":
			# Cone parallel to X-axis
			# FIXME the last variable, single is not taken into account
			tag = "QUA"
			x,y,z,t2  = arg[:4]
			try:
				#single = arg[4]
				log.warning("Single side cone is not implemented yet!")
				log.warningCont(command)
			except IndexError:
				pass
				#single = None
			what = [name,	-t2, 1.0, 1.0,			# Cxx Cyy Czz
					0.0, 0.0, 0.0,			# Cxy Cxz Cyz
					2.0*t2*x, -2.0*y, -2.0*z,	# Cx  Cy  Cz
					y*y + z*z - t2*x*x]

		elif surface=="K/Y":
			# Cone parallel to Y-axis
			# FIXME the last variable, single is not taken into account
			tag = "QUA"
			x,y,z,t2  = arg[:4]
			try:
				#single = arg[4]
				log.warning("single side cone is not implemented yet!")
				log.warningCont(command)
			except IndexError:
				pass
				#single = None
			what = [name,	1.0, -t2, 1.0,			# Cxx Cyy Czz
					0.0, 0.0, 0.0,			# Cxy Cxz Cyz
					-2.0*x, 2.0*t2*y, -2.0*z,	# Cx  Cy  Cz
					x*x + z*z - t2*y*y]

		elif surface=="K/Z":
			# Cone parallel to Z-axis
			# FIXME the last variable, single is not taken into account
			tag = "QUA"
			x,y,z,t2  = arg[:4]
			try:
				#single = arg[4]
				log.warning("single side cone is not implemented yet!")
				log.warningCont(command)
			except IndexError:
				pass
				#single = None
			what = [name,	1.0, 1.0, -t2,			# Cxx Cyy Czz
					0.0, 0.0, 0.0,			# Cxy Cxz Cyz
					-2.0*x, -2.0*y, 2.0*t2*z,	# Cx  Cy  Cz
					x*x + y*y - t2*z*z]

		elif surface=="KX":
			# Cone on X-axis
			tag  = "QUA"
			x  = arg[0]
			t2 = arg[1]
			try:
				#single = arg[2]
				log.warning("single side cone is not implemented yet!")
				log.warningCont(command)
			except IndexError:
				pass
				#single = None
			what = [name,	-t2, 1.0, 1.0,		# Cxx Cyy Czz
					0.0, 0.0, 0.0,		# Cxy Cxz Cyz
					2.0*t2*x, 0.0, 0.0,	# Cx  Cy  Cz
					-t2*x*x]

		elif surface=="KY":
			# Cone on Y-axis
			tag = "QUA"
			y  = arg[0]
			t2 = arg[1]
			try:
				single = arg[2]
				log.warning("single side cone is not implemented yet!")
				log.warningCont(command)
			except IndexError:
				single = None
			what = [name,	1.0, -t2, 1.0,		# Cxx Cyy Czz
					0.0, 0.0, 0.0,		# Cxy Cxz Cyz
					0.0, 2.0*t2*y, 0.0,	# Cx  Cy  Cz
					-t2*y*y]

		elif surface=="KZ":
			# Cone on Z-axis
			tag = "QUA"
			tag = "QUA"
			z  = arg[0]
			t2 = arg[1]
			try:
				single = arg[2]
				log.warning("single side cone is not implemented yet!")
				log.warningCont(command)
			except IndexError:
				single = None
			what = [name,	1.0, 1.0, -t2,		# Cxx Cyy Czz
					0.0, 0.0, 0.0,		# Cxy Cxz Cyz
					0.0, 0.0, 2.0*t2*z,	# Cx  Cy  Cz
					-t2*z*z]

		elif surface=="SQ":
			# Ellipsoid, Hyperboloid, Paraboloid axes parallel to X,Y,Z-axis
			tag  = "QUA"
			a,b,c, d,e,f, g, x,y,z = arg
			what = [name, a, b, c, 0, 0, 0,
					2*(d-a*x), 2*(e-b*y), 2*(f-c*z),
					g + a*x**2 + b*y**2 + c*z**2 - 2.0*(d*x + e*y + f*z)]

		elif surface=="GQ":
			# Cylinder, Cone, Ellipsoid, Hyperboloid, Paraboloid generic
			tag  = "QUA"
			what = [name, arg[0], arg[1], arg[2],
			              arg[3], arg[5], arg[4],	# Warning on the order xy yz zx
			              arg[6], arg[7], arg[8],
			              arg[9]]

		elif surface in ("TX","TY","TZ"):
			tag  = f"TR{surface[-1]}"
			what = [name] + arg

		elif surface in ("BOX", "REC", "WED"):
			tag  = surface
			what = [name, arg[0], arg[1], arg[2],
			              arg[3], arg[4], arg[5],
			              arg[6], arg[7], arg[8],
			              arg[9], arg[10], arg[11]]

		elif surface=="RPP":
			tag  = surface
			what = [name, arg[0], arg[1], arg[2],
			              arg[3], arg[4], arg[5]]

		elif surface=="SPH":
			tag  = surface
			what = [name, arg[0], arg[1], arg[2], arg[3]]

		elif surface=="RCC":
			tag  = surface
			what = [name, arg[0], arg[1], arg[2],
			              arg[3], arg[4], arg[5],
			              arg[6]]

		elif surface=="TRC":
			tag  = surface
			what = [name, arg[0], arg[1], arg[2],
			              arg[3], arg[4], arg[5],
			              arg[6], arg[7]]

		else:
			raise Exception(command)

		if comment:
			comment += f"\n{command}"
		else:
			comment = command
		card = FlukaCard(tag, what, comment)

		if tr != 0: self.input.add(FlukaCard("$start_transform",[f"rot{tr}"]))
		self.input.add(card)
		if tr != 0: self.input.add(FlukaCard("$end_transform"))

		return True

	#-----------------------------------------------------------------------
	# Parse MCNP cells and convert them to FLUKA regions
	#-----------------------------------------------------------------------
	def __parseCells(self):
		self.regions = []
		cells   = {}
		trsurf  = {}		# transformed surfaces for "TRCL" cells

		# Parse cells
		for tokens,comment in self.cells:
			#oexpr = tokens		# original expr
			log.say(">C>", tokens)
			tokens = tokens.replace("("," ( ")
			tokens = re.sub(r"#\s+", "#", tokens)
			tokens = re.sub(r"\s*=\s*", "= ", tokens)
			tokens = tokens.replace("#", " #")
			tokens = tokens.replace("# +(", " -( ")
			tokens = tokens.replace("# (", " -( ")
			tokens = tokens.replace(")", " ) ")
			tokens = tokens.replace(":", " : ")
			tokens = tokens.split()
			#log.say(">T>", tokens)

			idx  = tokens.pop(0)
			name = f"C{idx}"
			region = FlukaCard("REGION", [name], comment)
			self.input.add(region)
			cells[idx] = region
			self.regions.append(region)

			material = tokens.pop(0)
			expr  = []
			if material == "LIKE":
				like = tokens.pop(0)
				if tokens.pop(0) != "BUT":
					log.error(f"CELL {name} 'BUT' was expected after LIKE")
					continue
				region["@ref"] = ref = cells[like]
				material = ref["@mat"]
				density  = ref["@density"]

				expr = ref.expr().split()	# copy expression
			elif material != "0":
				density = tokens.pop(0)
			else:
				density = 0

			region["@mat"]     = material
			region["@density"] = density
			mat = f"M{material}"
			try:
				# FIXME make an array of densities per mat
				if self._density[mat]!=density:
					log.warning(f"CELL {name} Material {material} found with "
						"multiple densities."
						" Not supported for the moment.")
			except KeyError:
				self._density[mat] = density

			# Parse normal expression
			if not expr:
				while tokens:
					token = tokens.pop(0)
					if token==":":
						expr.append("\n|")
					elif token in ("(", "-(", ")"):
						if token == "(": token = "+("
						expr.append(token)
					elif token[0] == "#":
						expr.append(token)
					elif token[0] in "+-" and token[1:].isnumeric() or \
					     token.isnumeric():
						num = int(token)
						if num<0:
							op = "+"
							num = -num
						else:
							op = "-"
						expr.append(f"{op}S{num}")
					else:
						tokens.insert(0,token)
						break

			# Parse remaining tokens
			while tokens:
				token = tokens.pop(0)
				# -- Importance ---
				if token == "IMP":
					colon = tokens.pop(0)	# :
					equal = tokens.pop(0)	# particle=
					try:
						value = float(tokens.pop(0))	# value
					except ValueError:
						value = None
					if colon != ":" or equal[-1] != "=" or value is None:
						log.error(f"CELL {name} bad importance IMP{colon}{equal}")
					else:
						region["@IMP:"+equal[:-1]] = value

				# -- Translation ---
				elif token in ("TRCL=", "*TRCL="):
					deg = token[0]=="*"
					token = tokens.pop(0)
					if token == "(":
						transform = []
						while True:
							token = tokens.pop(0)
							if token==")": break
							try:
								transform.append(float(token))
							except ValueError:
								log.error(f"CELL {name} bad value '{token}' for TRCL")
						rot = f"trcl{idx}"
						self._trcl[rot] = self.__parseTransformation(
									transform, deg)
					else:
						# Use existing transformation
						rot = f"rot{token}"

					for i,t in enumerate(expr):
						# Surfaces should be as: [+\-]S\d+ (C\d+)?
						if len(t)>2 and t[0] in "+-" and t[1] == "S":
							m = PAT_SURFTRCL.match(t[1:])
							if m:	# append cell name
								surface = m.group(1) + name
							else:
								surface = t[1:] + name
							trsurf[surface] = rot	# assign rotation
							expr[i] = t[0] + surface

				# --- Material override ---
				elif token == "MAT=":
					region["@mat"] = tokens.pop(0)

				# FIXME: tokens to be implemented
				# VOL, PWT, EXT, FCL, WWN, DXC, NONU, PD, TMP, U, LAT, FILL
				else:
					value = tokens.pop(0)
					log.error(f"CELL {name} Unsupported token {token}{value} found")

			# Set the expression
			region.setExpr(" ".join(expr))

		# --------------------------------
		# Correct cell references # as -()
		# --------------------------------
		for idx,card in cells.items():
			expr = card.expr()
			if "#" not in expr: continue
			while True:
				m = PAT_CELL.match(expr)
				if m is None: break
				try:
					ref = cells[m.group(2)]
					# Expand only first level by removing
					refexpr = re.sub(PAT_RMCELL, "", ref.expr()).strip()
					expr = "%s\n-(%s)%s"%(m.group(1),refexpr,m.group(3))
				except KeyError:
					log.error(f"Cell {idx} refers to non existing cell #{m.group(2)}")
					expr = f"{m.group(1)}\n{m.group(3)}"
			card.setText(expr)

		# Create clones of surfaces if needed
		for name, rot in trsurf.items():
			m = PAT_SURFTRCL.match(name)
			surf = m.group(1)
			#cell = m.group(2)
			# Find body and clone it
			for card in self.input.cardlist:
				if len(card.tag) != 3 or card.name() != surf: continue
				pos = card.pos()
				clone = card.clone()
				clone.setName(name)
				self.input.add(FlukaCard("$start_transform",[rot]),pos+1)
				self.input.add(clone,pos+2)
				self.input.add(FlukaCard("$end_transform"),pos+3)
				self.input.renumber(pos)
				break
		return self.regions

	#-----------------------------------------------------------------------
	# Convert transformation list to translation and euler rotation
	#-----------------------------------------------------------------------
	def __parseTransformation(self, tokens, deg):
		if deg:
			for i in range(3,min(len(tokens),12)):
				x = math.cos(math.radians(tokens[i]))
				if abs(x)<EPS: x = 0.0
				tokens[i] = x

		pos = bmath.Vector(tokens[0], tokens[ 1], tokens[ 2])
		M = 1
		if len(tokens) > 5:
			X = bmath.Vector(tokens[3], tokens[ 4], tokens[ 5])
			X.normalize()
			if len(tokens) > 8:
				Y = bmath.Vector(tokens[6], tokens[ 7], tokens[ 8])
				Y.normalize()
				if len(tokens) > 11:
					Z = bmath.Vector(tokens[9], tokens[10], tokens[11])
					Z.normalize()
					if len(tokens)==13:
						M = int(tokens[12])
				else:
					Z = X ^ Y
			else:
				Y = X.orthogonal()
				Z = X ^ Y

			# Get Euler rotation as Rotx(Rx) * Roty(Ry) * Rotz(Rz)
			#+-----------------------+-----------------------+-----------------+
			#|          x.x          |           x.y         |       x.z       |
			#|     cos(z)*cos(y)     |      sin(z)*cos(y)    |     -sin(y)     |
			#+-----------------------+-----------------------+-----------------+
			#|          y.x          |           y.y         |       y.z       |
			#|    -sin(z)*cos(x)     |      cos(z)*cos(x)    |  cos(y)*sin(x)  |
			#| +cos(z)*sin(y)*sin(x) | +sin(z)*sin(y)*sin(x) |                 |
			#+-----------------------+-----------------------+-----------------+
			#|         z.x           |           z.y         |       z.z       |
			#|     sin(z)*sin(x)     |     -cos(z)*sin(x)    |  cos(y)*cos(x)  |
			#| +cos(z)*sin(y)*cos(x) | +sin(z)*sin(y)*cos(x) |                 |
			#+-----------------------+-----------------------+-----------------+
			#Rx =  math.degrees(math.atan2(Y.z, Z.z))
			#Ry = -math.degrees(math.asin(X.z))
			#Rz =  math.degrees(math.atan2(X.y, X.x))

			# Use the inverse=transpose matrix
			Rz = math.atan2(Y.x, X.x)

			# -sin(y) can have two solutions Ry or pi-Ry
			# Test against x.x or x.y depending whether cos(z) or sin(z) is zero
			cosz = math.cos(Rz)
			if abs(cosz)>0.001:
				denom = X.x / cosz
			else:
				denom = Y.x / math.sin(Rz)
			Ry = math.degrees(math.atan2(-Z.x, denom))
			Rx = math.degrees(math.atan2(Z.y, Z.z))
			Rz = math.degrees(Rz)
		else:
			Rx = Ry = Rz = 0.0
		return pos, Rx, Ry, Rz, M

	#-----------------------------------------------------------------------
	# Add a transformation
	#-----------------------------------------------------------------------
	def __addTransformation(self, name, pos, Rx, Ry, Rz, M, comment=""):
		if M!=1 and (abs(pos.x)>EPS or abs(pos.y)>EPS or abs(pos.z)>EPS):
			card = FlukaCard("ROT-DEFI", [name, "", "", "", pos.x,pos.y,pos.z], comment)
			self.input.addCard(card)

		if abs(Rz) > EPS:
			self.input.add(FlukaCard("ROT-DEFI",
					[name, "", "", bmath.format(Rz,10)], comment))
			comment=""

		if abs(Ry) > EPS:
			self.input.add(FlukaCard("ROT-DEFI",
					[name, 200, "", bmath.format(Ry,10)], comment))
			comment=""

		if abs(Rx) > EPS:
			self.input.add(FlukaCard("ROT-DEFI",
					[name, 100, "", bmath.format(Rx,10)], comment))
			comment=""

		if M==1 and (abs(pos.x)>EPS or abs(pos.y)>EPS or abs(pos.z)>EPS):
			self.input.add(FlukaCard("ROT-DEFI",
					[name, "", "", "", pos.x,pos.y,pos.z], comment))

	#-----------------------------------------------------------------------
	# FIXME densities I need a dictionary of material/density from the regions
	#-----------------------------------------------------------------------
	def __parseMaterial(self, name, compound, comment=""): # , regions to get densities....
		def za(x): return divmod(int(x.split(".")[0]), 1000)

		elems = {}
		tokens = compound.split()

		try:
			density = float(self._density[name])
		except:
			log.warning(f"Material {name} is not used. Setting density to 1.0")
			density = -1.0

		if density>0.0:
			log.warning(f"Material {name} is defined with atom density, please correct it")
		else:
			density = -density

		if len(tokens)/2 == 1:
			# isotope
			z,a = za(tokens[0])
			elems[z*1000+a] = name
			if a==0: a=""
			self.input.addCard(FlukaCard("MATERIAL",
						[name, z, "", density, "", "", a],comment))
		else:
			# compound
			self.input.addCard(FlukaCard("MATERIAL",
						[name, "", "", density, "", "", ""],comment))
			what = [name]
			for i in range(0,len(tokens),2):
				try:
					z,a = za(tokens[i])
				except:
					break
				frac = float(tokens[i+1])

				try:
					n = elems[z*1000+a]
				except:
					# Add a new isotope
					n = f"M{z*1000+a}"
					elems[z*1000+a] = n
					if a==0: a=""
					self.input.add(FlukaCard("MATERIAL",
							[n, z, "", "1.0", "", "", a]))

				what.append(frac)
				what.append(n)
			# make what multiples of 6+1
			while (len(what)-1)%6 != 0: what.append("")
			self.input.add(FlukaCard("COMPOUND", what))

	#-----------------------------------------------------------------------
	# Write mcnp file
	#-----------------------------------------------------------------------
	def write(self, filename):
		with open(filename,"w") as fout:
			fout.write(f"{self.title}\n")
			self._writeCells(fout)
			self._writeSurfaces(fout)
			self._writeTransformations(fout)
			self._writeMaterials(fout)
			self._writeBiasing(fout)
		return False

	#-----------------------------------------------------------------------
	# Write cells
	#-----------------------------------------------------------------------
	def _writeCells(self, fout):
		fout.write("C ---------- Cells ----------\n")
		for cell, comment in self.cells:
			self._writeLine(fout, cell, comment)
		fout.write("\n")

	#-----------------------------------------------------------------------
	# Write surfaces
	#-----------------------------------------------------------------------
	def _writeSurfaces(self, fout):
		fout.write("C ---------- Surfaces ----------\n")
		for surface, comment in self.surfaces:
			self._writeLine(fout, surface, comment)
		fout.write("\n")

	#-----------------------------------------------------------------------
	# Write transformations
	#-----------------------------------------------------------------------
	def _writeTransformations(self, fout):
		if not self.transform: return
		fout.write("C ---------- Transformations ----------\n")
		for trans in self.transform:
			self._writeLine(fout, trans)

	#-----------------------------------------------------------------------
	# Write materials
	#-----------------------------------------------------------------------
	def _writeMaterials(self, fout):
		fout.write("C ---------- Materials ----------\n")
		for mat in self.materials:
			if mat is None: continue
			self._writeLine(fout, "M%d %s" % (mat["@n"],
				" ".join(["%d %g"%(z,f) for z,[f,am] in mat.zaid])),
				str(mat))
			if mat.name() in ("WATER","H2O"):
				self._writeLine(fout, f"MT{mat['@n']} lwtr.01")

	#-----------------------------------------------------------------------
	# Write Biasing
	#-----------------------------------------------------------------------
	def _writeBiasing(self, fout):
		fout.write("C ---------- Biasing ----------\n")
		line   = []
		repeat = 0
		prev   = -1.0
		for imp in self.importance:
			if imp == prev:
				repeat += 1
			else:
				if prev<0:
					pass
				else:
					line.append(f"{prev:.15g}")
					if repeat == 1:
						line.append(line[-1])
					elif repeat > 1:
						line.append(f"{repeat}r")
				prev = imp
				repeat = 0
		line.append(f"{prev:.15g}")
		if repeat == 1:
			line.append(line[-1])
		else:
			line.append(f"{repeat}r")

		self._writeLine(fout, f"IMP:n {' '.join(line)}")

	#-----------------------------------------------------------------------
	# Write an mcnp line, with comments
	#-----------------------------------------------------------------------
	def _writeLine(self, fout, s, comment=""):
		# write first the comments
		if comment != "":
			self._writeMultiLine(fout, comment.expandtabs(), "C --- ")

		# parse the line not to exceed 72 chars
		line = None
		for a in s.split():
			if line is None:
				line = "%-5s"%(a)
			elif len(line)+len(a) > 72:
				fout.write(f"{line}\n")
				line = f"      {a}"
			else:
				line += f" {a}"
		fout.write(f"{line}\n")

	#-----------------------------------------------------------------------
	def _writeMultiLine(self, fout, lines, prefix=""):
		if lines:
			for line in lines.split("\n"):
				fout.write(f"{prefix}{line}\n")

#===============================================================================
if __name__ == "__main__":
	import os
	import sys
	import inputs
	from element import Element

	if len(sys.argv) <= 3:
		print("syntax: python Mcnp.py <input.inp|mcnp|m> <output.inp|mcnp|m> [macrobodies]")
		sys.exit(0)

	inputs.init()
	Element.init("db/isotopes.ini")

	_,ext = os.path.splitext(sys.argv[1])
	if ext == ".inp":
		fluka = Fluka()
		fluka.verbose = True
		fluka.load(sys.argv[1])
		mcnp  = Mcnp()
		try:
			mcnp.macroBodies = bool(int(sys.argv[3]))
		except:
			mcnp.macroBodies = False
		mcnp.fromFluka(fluka)
		mcnp.write(sys.argv[2])

	elif ext in (".m", ".mcnp"):
		mcnp.read(sys.argv[1])
		mcnp.input.save(sys.argv[2])
