# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import io
import os
import pickle
import tkinter as tk

import undo
import Ribbon
import Manual
import Project
import FlairRibbon

import icons
import tkFlair
import tkExtra
import tkDialogs
import bFileDialog

import Plot
#import Global
import GPPlot
import MPPlot
import UsrPlot
import BasePlot
import Usr2Plot
import UsrbinPlot
import GeometryPlot
import UserdumpPlot
import ResnucleiPlot

#from flair import PLOT_ENGINES
PLOT_ENGINES = ["Gnuplot", "Matplotlib"]

#===============================================================================
# Action Group
#===============================================================================
class _ActionGroup(FlairRibbon.FlairMenuGroup):
	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0)

		for n in PLOT_ENGINES:
			menu.add_radiobutton(
					label=n,
					image=icons.get("plot"), compound=tk.LEFT,
					variable=self.page._plotEngine,
					value=n)
		return menu

#===============================================================================
# Plot List Frame
#===============================================================================
class PlotListPage(FlairRibbon.FlairPage):
	"""Create plots of geometry or from the detectors defined in the input."""

	_name_ = "Plot"
	_icon_ = "plot"

	#----------------------------------------------------------------------
	def init(self):
		self._plotEngine = tk.StringVar()
		self._plotEngine.set(PLOT_ENGINES[0])
		self.undoredo  = undo.UndoRedo()
		self.selected  = None

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Plot List ===========
		group = Ribbon.LabelGroup(self.ribbon, "Plot List")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row = 0,0
		menulist = [	("Geometry", "GEOMETRY",  self.addGeometry),
				("USR-1D",   "USR-1D",    self.addUsr1D),
				("USR-2D",   "USR-2D",    self.addUsr2D),
				("USRBIN",   "USRBIN",    self.addUsrbin),
				("USERDUMP", "USERDUMP",  self.addUserdump),
				("RESNUCLE", "RESNUCLE",  self.addResnuclei)]

		b = Ribbon.MenuButton(group.frame, menulist,
				image=icons.get("add32"),
				text="Add \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				compound=tk.TOP,
				command=self.add)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add new plot [Ins]")

		# ---
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame,
				text="Oz",
				image=icons.get("wizard32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.wizard)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Automatically create plots")

		# ---
		col,row = 2,0
		b = Ribbon.LabelButton(group.frame,
				text="Move Up",
				image=icons.get("up"),
				command=self.moveUp)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move selected plots up [Ctrl-Up]")

		# ---
		col,row = 2,1
		b = Ribbon.LabelButton(group.frame,
				text="Move Down",
				image=icons.get("down"),
				command=self.moveDown)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move selected plots down [Ctrl-Down]")

		# ---
		col,row = 2,2
		b = Ribbon.LabelButton(group.frame,
				text="Rename",
				image=icons.get("rename"),
				command=self.rename)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Rename plot [F2]")

		# ---
		col,row = 3,0
		b = Ribbon.LabelButton(group.frame,
				text="Delete",
				image=icons.get("x"),
				command=self.delete)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Delete selected plots [Del]")

		# ---
		col,row = 3,2
		b = Ribbon.LabelButton(group.frame,
				text="Clone",
				image=icons.get("clone"),
				command=self.clone)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Clone selected plots [Ctrl-D]")

		# ========== Plot ===========
		group = _ActionGroup(self.ribbon, self, "Action")
		#group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()
		group.frame.grid_columnconfigure(2, weight=1)

		# ---
		col,row = 0,0
		menulist = []
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("save"),
				command=self.save)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Save plot")

		col,row = 1,0
		menulist = []
		for fmt in sorted(tkFlair._PLOT_FORMAT):
			cmd = lambda s=self,f=fmt: s.save(f)
			menulist.append((fmt, "save", cmd))
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Save",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Save plot as")

		# ---
		col,row = 0,1
		b = Ribbon.LabelButton(group.frame,
				text="Print",
				image=icons.get("print"),
				command=self.hardcopy)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Print selected plots")

		# ---
		col,row = 0,2
		b = Ribbon.LabelButton(group.frame,
				text="Notes",
				image=icons.get("document"),
				command=self.toNotes)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Insert to notes")

		# ---
		col,row = 2,0
		b = Ribbon.LabelButton(group.frame,
				text="Clean",
				image=icons.get("clean32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.clean)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove all generated files")

		# ---
		col,row = 3,0
		b = Ribbon.LabelButton(group.frame,
				text="Plot",
				image=icons.get("plot32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.execute)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Do the plot. Generate files when needed")

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(scrolly=True)

		self.splitter = tkExtra.HSplitter(self.frame,
					tkFlair.config.getInt(tkFlair._PAGE_SECTION,
						self.name+".split", 240),
						True)
		self.splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# -----------
		self.listbox = tkExtra.ImageListbox(self.splitter.leftFrame(),
					width=16)
		self.listbox.pack(fill=tk.BOTH, expand=tk.YES, side=tk.LEFT)
		sb = tk.Scrollbar(self.splitter.leftFrame(),
				orient=tk.VERTICAL,
				command=self.listbox.yview)
		sb.pack(fill=tk.Y, side=tk.RIGHT)
		self.listbox.config(yscrollcommand=sb.set)

		self.listbox.bind('<ButtonRelease-1>',	self.release)
		self.listbox.bind('<Button-3>',		self.popup)
		self.listbox.bind('<<Rename>>',		self.rename)
		self.listbox.bind('<Return>',		self.rename)
		self.listbox.bind('<Control-Return>',	self.execute)
		self.listbox.bind('<KP_Enter>',		self.rename)
		self.listbox.bind('<Delete>',		self.delete)
		self.listbox.bind('<<MoveUp>>',		self.moveUp)
		self.listbox.bind('<<MoveDown>>',	self.moveDown)
		self.listbox.bind('<<ListboxSelect>>',	self.select)
		self.listbox.bind('<<Copy>>',		self.copy)
		self.listbox.bind('<<Cut>>',		self.cut)
		self.listbox.bind('<<Paste>>',		self.paste)

		self.listbox.focus_set()
		self.lastrelease = (-1, 0.0)

		# -----------
		self.base = BasePlot.CommonPlotFrame(self.splitter.rightFrame(), self.flair)
		self.base.pack(fill=tk.BOTH, expand=tk.YES)

		# ----- Create Plot Frames ----
		self.frames = {}
		for C in [GeometryPlot.GeometryPlotFrame,
			  UsrPlot.Usr1DPlotFrame,
			  Usr2Plot.Usr2DPlotFrame,
			  UsrbinPlot.UsrbinPlotFrame,
			  UserdumpPlot.UserdumpPlotFrame,
			  ResnucleiPlot.ResnucleiPlotFrame]:
			frame = C(self.base.common, self.flair)
			self.frames[frame.type()] = frame
		self.displayed = None

	# ----------------------------------------------------------------------
	def help(self):
		Manual.show("F5")

	# ----------------------------------------------------------------------
	def canUndo(self):	return self.undoredo.canUndo()
	def canRedo(self):	return self.undoredo.canRedo()
	def resetUndo(self):	self.undoredo.reset()

	# ----------------------------------------------------------------------
	def undo(self, event=None):
		if self.canUndo():
			self.undoredo.undo()
			self.selected = None
			self.refresh()

	# ----------------------------------------------------------------------
	def redo(self, event=None):
		if self.canRedo():
			self.undoredo.redo()
			self.selected = None
			self.refresh()

	# ----------------------------------------------------------------------
	def delPlotUndo(self, idx):
		undoinfo = (f"delete plot {self.project.plots[idx].name}",
				self.addPlotUndo, idx, self.project.plots[idx].clone())
		del self.project.plots[idx]
		return undoinfo

	# ----------------------------------------------------------------------
	def addPlotUndo(self, idx, plot):
		undoinfo = (f"add plot {plot.name}", self.delPlotUndo, idx)
		self.project.plots.insert(idx, plot)
		return undoinfo

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		# Operate only on the listbox
		if self.page.focus_get() is not self.listbox:
			super().copy(event)
		else:
			self.get()
			plots = []
			for i in self.selected:
				plot = self.project.plots[i]
				plots.append((plot.name, plot.type, plot))

			bio = io.BytesIO()
			pickler = pickle.Pickler(bio)
			pickler.dump(plots)
			self.flair.clipboard_clear()
			self.flair.clipboardAppend(bio, tkFlair._PLOT_MIME)
			return "break"

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		# Operate only on the listbox
		if self.page.focus_get() is not self.listbox: return
		self.copy()
		self.delete()
		return "break"

	# ----------------------------------------------------------------------
	def paste(self, event=None):
		# Operate only on the listbox
		if self.page.focus_get() is not self.listbox:
			super().paste(event)
		else:
			self.get()
			self.current = None
			if tkFlair._PLOT_MIME in self.flair.clipboardTargets():
				unpickler = self.flair.clipboardUnpickler(tkFlair._PLOT_MIME)
				#try:
				plots = unpickler.load()
				#except:
				#	self.flair.notify("Cannot paste",
				#			"Invalid content in clipboard",
				#			tkFlair.NOTIFY_WARNING)
				#	return "break"
			else:
				self.flair.notify("Cannot paste",
						"Invalid content in clipboard",
						tkFlair.NOTIFY_WARNING)
				return "break"

			active = self.listbox.index(tk.ACTIVE)
			if active < 0 or active >= self.listbox.size()-1:
				active = -1
			self.listbox.selection_clear(0,tk.END)

			new = []
			for n,t,v in plots:
				plot = Project.PlotInfo(self.project, n)
				plot.type = t
				plot.update(v)
				if active < 0:
					new.append(len(self.project.plots))
					self.project.plots.append(plot)
				else:
					active += 1
					new.append(active)
					self.project.plots.insert(active, plot)

			self.refresh()
			self.listbox.selection_clear(0,tk.END)
			for s in new: self.listbox.selection_set(s)
			self.listbox.activate(new[0])
			self.listbox.see(new[0])
			self.select()
			return "break"

	# ----------------------------------------------------------------------
	# Delete plot
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		if self.page.focus_get() is not self.listbox: return
		self.get()	# get all information before pushing into undo

		undoinfo = []

		self.listbox.focus_set()
		lst = list(map(int, self.listbox.curselection()))
		if not lst: return
		lst.reverse()
		act = lst[0]
		for i in lst:
			undoinfo.append(self.delPlotUndo(i))
			self.listbox.delete(i)
		if act >= self.listbox.size(): act = self.listbox.size()-1
		if act >= 0:
			self.listbox.activate(act)
			self.listbox.selection_set(act)
		del self.selected[:]
		self.undoredo.add(undoinfo)
		self.select()
		self.modified()
		self.flair.refreshGeometry()
		return "break"

	# ----------------------------------------------------------------------
	# Clone plots
	# ----------------------------------------------------------------------
	def clone(self, event=None):
		self.listbox.focus_set()
		sel = list(map(int,self.listbox.curselection()))
		if not sel: return
		self.listbox.selection_clear(0,tk.END)
		insertPoint = sel[-1]+1
		undoinfo = []
		for i in reversed(sel):
			plot = self.project.plots[i].clone()
			self.project.findUniquePlotName(plot)
			undoinfo.append(self.addPlotUndo(insertPoint, plot))
			self.listbox.insert(insertPoint, icons.get(plot.type.upper()), plot.name)
			self.listbox.selection_set(insertPoint)
		self.modified()
		self.undoredo.add(undoinfo)
		self.listbox.activate(i+1)
		self.flair.refreshGeometry()
		self.select()

	# ----------------------------------------------------------------------
	# Activate and set focus
	# ----------------------------------------------------------------------
	def activate(self):
		#self.refresh()
		super().activate()

	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		engine = self.project["plotengine"]
		if engine: self._plotEngine.set(engine)
		if self.flair.geometryPage:
			self.flair.geometryPage.saveState()
		saveSelected = list(map(int, self.listbox.curselection()))
		self.listbox.delete(0,tk.END)
		for plot in self.project.plots:
			self.listbox.insert(tk.END,icons.get(plot.type.upper()), plot.name)
		self.listbox.selection_clear(0,tk.END)
		for s in saveSelected:
			self.listbox.selection_set(s)
		if saveSelected:
			self.listbox.activate(saveSelected[0])
		else:
			self.listbox.selection_set(0)
			self.listbox.activate(0)
		self.select()

	# ----------------------------------------------------------------------
	# Reset before loading new projects
	# ----------------------------------------------------------------------
	def reset(self):
		if self.page is None: return
		self.selected = None
		self.listbox.delete(0,tk.END)

	# ----------------------------------------------------------------------
	# Get information from the plots
	# ----------------------------------------------------------------------
	def get(self):
		if not self.selected: return
		if self.displayed is None: return
		self.project["plotengine"] = self._plotEngine.get()
		for i in self.selected:
			try:
				plot = self.project.plots[i]
			except IndexError:
				break

			# Remember for undo
			oldplot = plot.clone()

			# Get information
			try:
				self.base.get(plot)
				if plot.type == self.displayed.type():
					self.displayed.get(plot)
			except tk.TclError:
				continue

			#if Global.verbose:
			#	say("----old---")
			#	oldplot.write(sys.stdout)
			#	say("\n---new---")
			#	plot.write(sys.stdout)

			if plot.hash() != oldplot.hash():
				undoinfo = (plot.setUndo, oldplot)
				self.undoredo.add(undoinfo)
				self.flair.modified()

	# ----------------------------------------------------------------------
	# Selection changed in the listbox
	# ----------------------------------------------------------------------
	def select(self, event=None):
		self.get()

		# As a precaution remove everything
		for child in self.base.common.winfo_children():
			child.pack_forget()

		self.selected = list(map(int, self.listbox.curselection()))
		if not self.selected: return

		# Find common type
		plot = self.project.plots[self.selected[0]]
		t = plot.type
		for i in self.selected:
			if t != self.project.plots[i].type:
				t = None
				break

		# Display the new frame
		if t:
			self.displayed = self.frames[plot.type]
			self.displayed.pack(expand=tk.YES, fill=tk.BOTH)

		# Fill values with plot type
		self.displayed.setState(tk.NORMAL)
		self.base.setState(tk.NORMAL)

		# Fill in reverse order so maybe the set to modify the plot like in usrbin
		self.displayed.set(plot)
		self.base.set(plot)

		if len(self.selected)>1:
			self.base.setState(tk.DISABLED)
			self.displayed.setState(tk.DISABLED)
		else:
			self.base.setState(tk.NORMAL)
			self.displayed.setState(tk.NORMAL)

	# ----------------------------------------------------------------------
	# Add a new plot
	# ----------------------------------------------------------------------
	def add(self, typ=None):
		plot = Project.PlotInfo(self.project, "")
		self.project.plots.append(plot)
		plot["title"] = f"Plot #{len(self.project.plots)}"
		if self.project.name() == "":
			plot.name   = f"plot{len(self.project.plots):02d}"
		else:
			fn = os.path.splitext(self.project.name())[0]
			plot.name   = f"{fn}_plot{len(self.project.plots):02d}"
		plot["format"] = Plot.DEFAULT_FORMAT
		if typ:
			plot.type = typ
		else:
			plot.type = Plot.PLOT_TYPES[0]

		# Detector specific
		if plot.type == "USRBIN":
			plot["cbtics"]   = 1
			plot["cblog"]    = 1
			plot["cbcolors"] = 30

		elif plot.type == "USR-1D":
			plot["ndetectors"] = 1
			plot["name.0"] = "#Detector 1"
			plot["det.0"]  = 0

		elif plot.type == "USR-2D":
			plot["ztics"]  = 1
			plot["cbtics"] = 1

		elif plot.type == "RESNUCLE":
			plot["cbtics"] = 1

		elif plot.type == "USERDUMP":
			plot["type"] = "Histogram"

		self.listbox.insert(tk.END,icons.get(plot.type.upper()), plot.name)
		self.modified()
		self.listbox.selection_clear(0,tk.END)
		self.listbox.selection_set(tk.END)
		self.listbox.activate(tk.END)
		self.listbox.see(tk.END)
		self.listbox.focus_set()
		self.select()
		self.flair.refreshGeometry()
		self.lastrelease = (self.listbox.index(tk.END), 0)

	# ----------------------------------------------------------------------
	def addGeometry(self):	self.add("Geometry")
	def addUsr1D(self):	self.add("USR-1D")
	def addUsr2D(self):	self.add("USR-2D")
	def addUsrbin(self):	self.add("USRBIN")
	def addUserdump(self):	self.add("USERDUMP")
	def addResnuclei(self):	self.add("RESNUCLE")

	# ----------------------------------------------------------------------
	# Swap two items in the list
	# ----------------------------------------------------------------------
	def swap(self, a, b):
		lb = self.listbox
		if a>b:
			a, b = b, a

		ima, at = lb.elicit(a);	al = self.project.plots[a]
		imb, bt = lb.elicit(b);	bl = self.project.plots[b]

		lb.delete(b)
		lb.delete(a)

		lb.insert(a, imb, bt);	self.project.plots[a] = bl
		lb.insert(b, ima, at);	self.project.plots[b] = al

	# ----------------------------------------------------------------------
	# Move up select items by one
	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		self.listbox.focus_set()
		self.get()
		self.selected = None
		lb  = self.listbox
		for i in map(int,lb.curselection()):
			if i==0: continue
			prev = i-1
			if not lb.selection_includes(prev):
				act = lb.index(tk.ACTIVE)
				self.swap(prev,i)
				lb.selection_set(prev)
				if act == i: lb.activate(prev)
		self.selected = list(map(int, self.listbox.curselection()))
		return "break"

	# ----------------------------------------------------------------------
	# Move down select items by one
	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		self.listbox.focus_set()
		self.get()
		self.selected = None
		lb  = self.listbox
		sz  = lb.size()-1
		lst = list(map(int,lb.curselection()))
		lst.reverse()
		for i in lst:
			if i >= sz: continue
			next_ = i+1
			if not lb.selection_includes(next_):
				act = lb.index(tk.ACTIVE)
				self.swap(i,next_)
				lb.selection_set(next_)
				if act == i: lb.activate(next_)
		self.selected = list(map(int, self.listbox.curselection()))
		return "break"

	# ----------------------------------------------------------------------
	def wizard(self):
		"""scan input file for possible plots"""
		if self.project.name() == "":
			pattern = "%s"
		else:
			pattern = "%s_%%s" % (os.path.splitext(self.project.name())[0])

		# Add one geometry plot if doesn't exist
		for p in self.project.plots:
			if p.type == Plot.PLOT_TYPES[0]: break
		else:
			plot = Project.PlotInfo(self.project, pattern % ("geometry"))
			plot["title"]  = f"Geometry {self.project.title}"
			plot.type      = Plot.PLOT_TYPES[0]
			plot["format"] = ".eps"
			self.project.plots.append(plot)

		# For default Run find all usrinfos
		run = self.project.runs[0]
		for usr in run.usrinfo:
			name = os.path.splitext(usr.name)[0]

			plot = Project.PlotInfo(self.project, f"{name}_plot")
			self.project.plots.append(plot)
			if usr.type=="usrbin":
				plot.type = "USRBIN"
				plot["datafile"] = usr.name
				plot["geo"]      = BasePlot._GEO_AUTO
				plot["cbtics"]   = 1
				plot["cblog"]    = 1
				plot["cbcolors"] = 30

			elif usr.type=="resnucle":
				plot.type = "RESNUCLE"
				plot["datafile"] = usr.name

			else:
				plot.type = "USR-1D"
				plot["xlog"] = 1
				plot["ylog"] = 1
				plot["ndetectors"] = 1
				tablis = f"{name}_tab.lis"
				plot["name.0"] = "Detector 1"
				plot["file.0"] = tablis
				plot["det.0"]  = "0"
				plot["x.0"]    = "2"
				plot["y.0"]    = "0"
				plot["lw.0"]   = "2"
				plot["lc.0"]   = "red"
				plot["with.0"] = "histerror"
			plot["title"]  = f"{plot.type} {name.replace('_',' ')}"
			plot["format"] = ".eps"

		self.refresh()
		self.modified()

	# ----------------------------------------------------------------------
	def release(self, event=None):
		listbox = event.widget
		act = listbox.nearest(event.y)
		if act: self.listbox.activate(act)
		sel = self.listbox.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])

		try: active = listbox.index(tk.ACTIVE)
		except: return
		if active != sel: return

		# Character position under cursor
		if self.lastrelease[0] == active:
			if event.time <= 500:
				self.lastrelease = (active, event.time)
				return
		else:
			self.lastrelease = (active, event.time)
			return

		index = self.listbox.index(f"@{event.x},{event.y}")
		if index is None:
			self.lastrelease = (-1, event.time)
			return

		self.lastrelease = (active, event.time)

		if int(index.split(".")[1]) > 1:
			self.rename(listbox)
		else:
			self.changeTypePopup()

	# ----------------------------------------------------------------------
	def rename(self, event=None):
		active = self.listbox.index(tk.ACTIVE)
		if active is None: return
		try:
			plot = self.project.plots[active]
		except IndexError:
			return

		# Edit the filename
		filetypes = []
		for ext in sorted(tkFlair._PLOT_FORMAT):
			desc = bFileDialog.DESCRIPTION.get(ext[1:],f"File type {ext}")
			filetypes.append((desc, "*"+ext))
		filetypes.append(("All","*"))

		# Edit the filename
		edit = tkExtra.InPlaceFile(self.listbox, value=plot.name+plot("format"),
						title="Save plot file as",
						showext=False,
						filetypes=filetypes)
		if not edit.value:	# None or ""
			# Break not to insert the <Return> in the Text and modify the text
			return "break"

		fn = self.project.relpath(edit.value)
		plot.name, ext = os.path.splitext(fn)
		self.modified()
		if ext: plot["format"] = ext
		self.listbox.delete(active)
		self.listbox.insert(active,icons.get(plot.type.upper()), plot.name)
		self.listbox.selection_set(active)
		self.listbox.activate(active)
		self.listbox.see(active)
		# Break not to insert the <Return> in the Text and modify the text
		return "break"

	# ----------------------------------------------------------------------
	def popup(self, event):
		menu=tk.Menu(self.listbox, tearoff=0)

		submenu = tk.Menu(menu)
		menu.add_command(label="Rename", image=icons.get("rename"), compound=tk.LEFT,
				command=self.rename)

		menu.add_cascade(label="Add", image=icons.get("add"), compound=tk.LEFT,
				underline=0, menu=submenu)
		for t in Plot.PLOT_TYPES:
			icon = icons.get(t.upper())
			submenu.add_command(label=t,
				compound=tk.LEFT, image=icon,
				command=lambda s=self,t=t:s.add(t))

		submenu = tk.Menu(menu)
		menu.add_cascade(label="Change to", image=icons.get("change"), compound=tk.LEFT,
				underline=7, menu=submenu)

		self._changeTypeMenu(submenu)

		menu.add_command(label="Delete", image=icons.get("x"), compound=tk.LEFT,
				underline=0, command=self.delete)
		menu.add_command(label="Clone", image=icons.get("clone"), compound=tk.LEFT,
				underline=0, command=self.clone)

		menu.tk_popup(event.x_root, event.y_root)

	# ----------------------------------------------------------------------
	def changeTypePopup(self):
		menu=tk.Menu(self.listbox, tearoff=0)
		self._changeTypeMenu(menu)
		bbox = self.listbox.dlineinfo(tk.ACTIVE)
		menu.tk_popup(self.listbox.winfo_rootx(),
			      self.listbox.winfo_rooty() + bbox[1] + bbox[3])

	# ----------------------------------------------------------------------
	def _changeTypeMenu(self, menu):
		for t in Plot.PLOT_TYPES:
			icon = icons.get(t.upper())
			menu.add_command(label=t,
				compound=tk.LEFT, image=icon,
				command=lambda s=self,t=t:s.changePlotType(t))

	# ----------------------------------------------------------------------
	# Change the Plot type
	# ----------------------------------------------------------------------
	def changePlotType(self, newtype):
		active = self.listbox.index(tk.ACTIVE)
		if active is None: return
		plot   = self.project.plots[active]
		if plot.type != newtype:
			self.modified()
			plot.type = newtype

			self.listbox.delete(active)
			self.listbox.insert(active,icons.get(plot.type.upper()), plot.name)
			self.listbox.selection_set(active)
			self.listbox.activate(active)
			self.listbox.see(active)

			self.select()
			self.flair.refreshGeometry()

	# ----------------------------------------------------------------------
	# generate the plot
	# ----------------------------------------------------------------------
	def execute(self, show=True, save=False, export=False, clean=False,
			printer=None, fmt=None):
		if not self.selected: return

		self.get()
		if self.flair.startPlotEngine(self._plotEngine.get()) is None: return

		doall    = False
		saved    = []
		printed  = []
		indexes  = {}	# display indexes updated
		for i, p in enumerate(self.selected):
			plot = self.project.plots[p]

			# Ask to wait before next plot
			if show and not save  and not export and not clean and \
			   i>0  and not doall and int(plot("index")) in indexes:
				try:
					ans = tkDialogs.Dialog(self.page.winfo_toplevel(),
							{"title":"Multiple Plot",
							"text":"Continue with the next\nOr do all plots",
							"bitmap": "questhead",
							"default": 0,
							"strings": ("Next","All","Cancel")})
				except tk.TclError:
					ans = 1
				indexes.clear()
				if ans.num==1:
					doall = True
				elif ans.num in (2,-1):
					break

			self.base.get(plot)
			if plot.type == self.displayed.type():
				self.displayed.get(plot)

			self.busy()
			#try
			log = self.flair.newLog(plot.type, plot.name)

			# FIXME to be changed with one class!!!
			if self._plotEngine.get() == PLOT_ENGINES[0]:
				plotter = GPPlot.GPPlot(
						self.flair.plotEngine,
						self.project,
						plot,
						log)
			elif self._plotEngine.get() == PLOT_ENGINES[1]:
				plotter = MPPlot.MPPlot(
						self.flair.plotEngine,
						self.project,
						plot,
						log)
			if clean or plot.verify() is not None:
				plotter.cleanup()

			err = None
			if show:
				self.setStatus(f"Processing plot: {plot.name}",
						"Tomato", update=True)
				try:
					err = plotter.show()
				except (OSError, IOError) as error:
					self.flair.notify("OS Error",
						str(error),
						tkFlair.NOTIFY_ERROR)
					err = error
				self.displayed.afterPlot()
				self.setStatus("Plot completed")
			self.notBusy()
			#except
			#...
			if err:
				if isinstance(err,tuple):
					err,inp,pid = err
				else:
					inp = pid = None
				self.flair.notify("Plot Error",
					f"{err}\nCheck Output window for additional messages.",
					tkFlair.NOTIFY_ERROR,
					"Output", self.flair.showOutput, showOnOutput=False)
				if inp is not None:
					tkFlair.OutputFilesDialog(self.flair,
							os.path.basename(inp),
							pid, 1, self.project)
					if Project.cleanup:
						self.project.deleteFiles(inp, pid)
				break

			# Save file if needed
			if save:
				if fmt is None:
					saved.append(plotter.save(plot("format")))
				else:
					plot["format"] = fmt
					saved.append(plotter.save(fmt))

			# Print if needed
			if printer:
				plotter.hardcopy(printer)
				printed.append(plot.name)

			# Update which display indexes have been used up to now
			indexes[int(plot("index"))] = True

		if clean:
			self.setStatus("Plot files cleaned", "DarkGreen")

		if save:
			self.setStatus(f"Saved: {' '.join(saved)}", "DarkBlue")

		if printer:
			self.setStatus(f"Printed: {' '.join(printed)}", "DarkBlue")

	# ----------------------------------------------------------------------
	def export(self):		self.execute(export=True)
	def save(self,fmt=None):	self.execute(save=True, fmt=fmt)
	def clean(self):		self.execute(show=False, clean=True)

	# ----------------------------------------------------------------------
	# Select plot by name
	# ----------------------------------------------------------------------
	def showPlot(self, name):
		for i,plot in enumerate(self.project.plots):
			if plot.name == name:
				self.listbox.selection_clear(0,tk.END)
				self.listbox.selection_set(i)
				self.listbox.activate(i)
				self.listbox.see(i)
				self.listbox.focus_set()
				self.select()
				break

	# ----------------------------------------------------------------------
	def hardcopy(self):
		printer = tkFlair.printer(self.page)
		if printer is None: return
		self.execute(printer=printer)
		printer.close()

	# ----------------------------------------------------------------------
	def showNotesPage(self):
		self.flair.notifyHide()
		self.flair.tabs.changePage("Flair")

	# ----------------------------------------------------------------------
	# Insert to project notes a link to the image
	# ----------------------------------------------------------------------
	def toNotes(self):
		added = 0

		self.save(".png")

		for p in self.selected:
			plot = self.project.plots[p]
			link = f"{{img:{plot.name}.png}}"
			if link not in self.project.notes:
				self.project.notes += f"\n\n{link}"
				added += 1

		if added:
			self.flair.modified()
			self.flair.refresh("image")
			self.setStatus(f"{added} plots added to notes")
			self.flair.notify("Plots to Notes",
				f"{added} plots added to notes")
		else:
			self.setStatus("Replaced existing plots in Notes")
			self.flair.notify("Plots to Notes",
				"Replaced existing plots to notes",
				tkFlair.NOTIFY_NORMAL,
				"Notes",
				self.showNotesPage)
