# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import tkinter as tk
import tkinter.messagebox as messagebox

import icons
import Ribbon
import RunTab
import DataTab
import Project
import RunList
import tkFlair
import tkExtra
import FilesTab
import Evaluation
import bFileDialog
import FlairRibbon

from fluka import Fluka
from moira import Moira

#===============================================================================
# Run Frame
#===============================================================================
class RunPage(FlairRibbon.FlairPage):
	"""Execute a simulation, inspect the output files and merge the generated data."""

	_name_ = "Run"
	_icon_ = "run"

	#----------------------------------------------------------------------
	def init(self):
		self.selected  = []
		self.refreshOn = ("save","run","exe","files","data")

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Tools ===========
		group = Ribbon.LabelGroup(self.ribbon, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		kw = {
			"compound" : tk.TOP,
			"variable" : self.tabVar,
			"command"  : self.changeTab,
			"padx"     : 5
		}

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Runs",
				image=icons.get("run32"),
				anchor=tk.CENTER,
				value=0,
				**kw)
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		tkExtra.Balloon.set(b, "Run setting page")

		b = Ribbon.LabelRadiobutton(group.frame,
				text="Files",
				image=icons.get("file-manager32"),
				anchor=tk.CENTER,
				value=1,
				**kw)
		tkExtra.Balloon.set(b, "Output files inspecting page")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)

		b = Ribbon.LabelRadiobutton(group.frame,
				text="Data",
				image=icons.get("data32"),
				anchor=tk.CENTER,
				value=2,
				**kw)
		tkExtra.Balloon.set(b, "Merge output data page")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)

		split = tk.PanedWindow(self.frame,
					orient=tk.HORIZONTAL,
					showhandle=0,
					sashrelief=tk.RAISED,
					sashwidth=4,
					opaqueresize=1)
		split.pack(expand=tk.YES, fill=tk.BOTH)

		# ======= Runs ========
		self.runList = RunList.RunListbox(split, True, height=5, stretch="middle")
		split.add(self.runList, minsize=50) #, padx=3, pady=3)

		# Bind events
		self.runList.bind('<<ListboxSelect>>',	self.select)
		self.runList.bindList("<Insert>",	self.add)
		self.runList.bindList("<Delete>",	self.delete)
		self.runList.bindList("<<MoveUp>>",	self.moveUp)
		self.runList.bindList("<<MoveDown>>",	self.moveDown)

		# ======== Tabs =======
		tabFrame = tk.Frame(split, padx=0, pady=0)
		split.add(tabFrame)

		# --------- Run tab ---------
		self.run = RunTab.RunTab(tabFrame, self)
		self.addTab(self.run)
		# --------- Files tab ---------
		self.addTab(FilesTab.FilesTab(tabFrame, self))
		# --------- Data tab ---------
		self.addTab(DataTab.DataTab(tabFrame, self))

		self.runList.set(self.flair)

	# ----------------------------------------------------------------------
	def select(self, event=None):
		self.activeTab.activate()
		self.setValid()

	# ----------------------------------------------------------------------
	def get(self):
		self.run.get()

	# ----------------------------------------------------------------------
	# Activate and set focus
	# ----------------------------------------------------------------------
	def activate(self):
		super().activate()

	# ----------------------------------------------------------------------
	# Update run progress information
	# and Enable or Disable the fields
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		if self.project is None: return

		# Refresh run listbox
		self.run.fillQueueList()	# If list is changed from ConfigDialog
		self.run.setQueueList()

		# Run list
		self.runList.updateSelection()
		self.runList.fill()
		self.select()

	# ----------------------------------------------------------------------
	def refreshStatus(self):
		super().refreshStatus()
		if self.activeTab: self.activeTab.refreshStatus()

	# ----------------------------------------------------------------------
	def reset(self):
		self.selected = []

	# ----------------------------------------------------------------------
	def beforeDock(self):
		self.reset()

	# ----------------------------------------------------------------------
	def clean(self):
		self.activeTab.clean()

	# ----------------------------------------------------------------------
	# Move runs to a different position
	# expect a list of runs (idx, run)
	# ----------------------------------------------------------------------
	def _moveRunsToPosition(self, runs, pos):
		self.runList.updateSelection()

		# Delete selected runs from the list
		runs.reverse()
		for run in runs:
			self.project.runs.remove(run)

		# Insert them at pos
		for run in runs:
			append = (pos >= len(self.project.runs))
			if append:
				self.project.runs.append(run)
			else:
				self.project.runs.insert(pos, run)

		self.runList.fill()
		# Activate the first selected item
		self.runList.activate(self.runList.curselection()[0])
		self.runList.see(tk.ACTIVE)
		self.select()
		self.flair.refresh("run")
		self.modified()

	# ----------------------------------------------------------------------
	# Move up select items by one
	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		self.get()
		del self.selected[:]
		runs = self._selectedParentsList()
		if not runs: return

		# Find position to insert them
		pos = self.project.getRunIndex(runs[0]) - 1
		if pos <= 0: return

		# find a parent position
		while pos>0:
			if not self.project.runs[pos]("parent"): break
			pos -= 1

		if pos <= 0: return
		self._moveRunsToPosition(runs, pos)
		return "break"

	# ----------------------------------------------------------------------
	# Move down select items by one
	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		self.get()
		del self.selected[:]
		runs = self._selectedParentsList()
		if not runs: return

		# Find position to insert them
		pos = self.project.getRunIndex(runs[-1]) + 1
		sz  = len(self.project.runs) - 1
		if pos > sz: return

		# find a parent position
		while pos <= sz:
			if not self.project.runs[pos]("parent"):
				# move to the end of the family
				if self.project.runs[pos].family:
					# skip child's
					pos += 1
					while pos <= sz:
						if not self.project.runs[pos]("parent"): break
						pos += 1
				else:
					pos += 1
				break
			pos += 1
		pos -= len(runs)
		self._moveRunsToPosition(runs, pos)
		return "break"

	# ----------------------------------------------------------------------
	# Add new run
	# ----------------------------------------------------------------------
	def add(self, event=None):
		self.get()
		name = self.project.name()
		filetypes = [
				("Fluka Input files","*.inp"),
				("Fluka %s*.inp files"%(name),"%s*.inp"%(name)),
				("Moira Input files","*.moira"),
				("Moira %s*.moira files"%(name),"%s*.moira"%(name)),
				("All","*")]

		if self.project.mode() == Moira.MODE:
			# Re-order and move moira on top
			filetypes.insert(0, filetypes.pop(3))
			filetypes.insert(0, filetypes.pop(3))

		fn = bFileDialog.asksaveasfilename(master=self.page,
					title="Create new run-input file as",
					initialdir=self.project.dir(),
					filetypes=filetypes)
		if fn:
			fn,ext = os.path.splitext(self.project.relpath(fn))
			# Check if run is there
			newfn = self.project.findUniqueRunName(fn)
			if newfn != fn:
				self.flair.notify("Run already exists",
					"Run '%s' already exists. Renamed to '%s'"%(fn,newfn),
					tkFlair.NOTIFY_WARNING)
				fn = newfn

			run = Project.RunInfo(self.project, fn)
			run["exe"] = self.project.compile.name		# Default executable
			if ext == ".inp":
				run["mode"] = Fluka.MODE
			elif ext == ".moira":
				run["mode"] = Moira.MODE

			self.project.addRun(run)
			del self.selected[:]
			self.runList.fill()
			self.flair.refresh("run")
			self.runList.selection_clear(0,tk.END)
			self.runList.selection_set(tk.END)
			self.runList.activate(tk.END)
			self.runList.see(tk.END)
			self.select()
			self.run.selectDefaultDefines()
			self.modified()

	# ----------------------------------------------------------------------
	# Delete runs
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		# protect against accidental delete
		#if self.flair.event is not None:
		#	if not self.focus_get() in self.runList.listboxes(): return

		sel = self.runList.curselection()
		if not sel: return "break"

		if tkFlair.askyesno("Remove Runs",
				   "Are you sure you want to remove the selected runs?",
				   parent=self.ribbon):
			del self.selected[:]
			sel = list(self.runList.curselection())
			first = int(sel[0])
			sel.reverse()
			for r in sel:
				if int(r) == 0: continue
				run = self.runList.getRun(r)
				if run is not None:
					self.project.delRun(run)

			if first > self.runList.size():
				first = tk.END

			self.runList.fill()
			self.runList.selection_clear(0,tk.END)
			self.runList.selection_set(first)
			self.runList.activate(first)
			self.select()
			self.flair.refresh("run")
			self.modified()
		return "break"

	# ----------------------------------------------------------------------
	# Clone runs
	# ----------------------------------------------------------------------
	def clone(self, event=None):
		self.get()
		# protect against accidental delete
		#if self.flair.event is not None:
		#	if not self.focus_get() in self.runList.listboxes(): return

		sel = self.runList.curselection()
		if not sel: return "break"
		oldSize = self.runList.size()
		for r in sel:
			run = self.runList.getRun(r)

			if run is None or run("parent"): continue # ignore child nodes

			child = run.clone()
			if run.isdefault():
				if self.project.name():
					child.name = self.project.name()+"-1"
				else:
					child.name = "untitled-1"
				child.defaultDefines()

			child.name = self.project.findUniqueRunName(child.name)
			self.project.addRun(child)

		self.runList.fill()
		self.flair.refresh("run")
		self.runList.selection_clear(0,tk.END)
		self.runList.selection_set(oldSize,tk.END)
		self.runList.activate(tk.END)
		self.runList.see(tk.END)
		#self.selected = list(map(int,self.runList.curselection()))
		self.select()
		self.modified()
		return "break"

	# ----------------------------------------------------------------------
	# Loop runs
	# ----------------------------------------------------------------------
	def loop(self):
		runs = []
		for r in self.runList.curselection():
			run = self.runList.getRun(r)
			if run is None or run("parent"): continue # ignore child nodes
			runs.append(run)

		if not runs: return

		LoopRunsDialog(self.page, self.flair, runs)
		self.flair.refresh("run")
		self.runList.fill()

	# ----------------------------------------------------------------------
	# Rename selected runs
	# ----------------------------------------------------------------------
	def rename(self, event=None):
		self.get()
		sel = list(self.runList.curselection())
		if not sel: return
		sel.reverse()
		for r in sel:
			# Skip default and childs
			if int(r)==0: continue
			run = self.runList.getRun(r)
			if run is None or run("parent"): continue # ignore child nodes

			newname = bFileDialog.asksaveasfilename(master=self.page,
					title="Rename run '%s' to"%(run.name),
					initialfile=run.name,
					filetypes=[("Fluka Input files","*.inp"),
						("All","*")])
			if not newname: continue
			fn, ext = os.path.splitext(self.project.relpath(newname))
			self.runList.renameRun(run, fn)
			self.runList.fill()
			self.flair.refresh("run")
			self.modified()

	# ----------------------------------------------------------------------
	# Return a list of the selected parents only
	# ----------------------------------------------------------------------
	def _selectedParentsList(self):
		runs = []
		for r in self.runList.curselection():
			if int(r)==0: continue		# ignore default run
			run = self.runList.getRun(r)
			if run is None or run("parent"): continue # ignore child nodes
			runs.append(run)
			if run.family:
				# Append all its family
				for childname in run.family:
					child = self.project.getRunByName(childname, run.name)
					runs.append(child)
		return runs

	# ----------------------------------------------------------------------
	def help(self):
		if str(self.page.focus_get()).startswith(str(self.runList)):
			super().help()
		else:
			self.activeTab.help()

#===============================================================================
# Create clone of Runs by looping on one variable
#===============================================================================
class LoopRunsDialog(tk.Toplevel):
	def __init__(self, master, flair, runs):
		super().__init__(master)
		self.transient()
		self.title("Loop runs")

		self.flair = flair
		self.runs  = runs

		self.incparent = tk.BooleanVar(self)

		l = tk.Label(self, text="Create clone of the selected runs by\n"
					"looping the following variable over\n"
					"the defined range",
				justify=tk.LEFT,
				anchor=tk.NW,
				pady=2)
		l.pack(side=tk.TOP)

		frame = tk.Frame(self)
		frame.pack(fill=tk.BOTH)

		# ---
		row = 0
		tk.Label(frame, text="Variable:").grid(
			row=row, column=0, sticky=tk.E)

		self.variable = tkExtra.Combobox(frame, width=16)
		self.variable.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.variable,"Select the variable to loop")

		# Update defines
		lst = []
		for card in self.flair.project.input.cardsSorted("#define",2).values():
			if card._indent != 0: continue
			lst.append(card.name())
		self.variable.fill(lst)

		# ---
		row += 1
		tk.Label(frame, text="Start value:").grid(
			row=row, column=0, sticky=tk.E)
		self.start = tk.Entry(frame, background="White", justify=tk.LEFT)
		self.start.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.start, "Enter starting value or expression")

		# ---
		row += 1
		tk.Label(frame, text="End value:").grid(
			row=row, column=0, sticky=tk.E)
		self.stop = tk.Entry(frame, background="White", justify=tk.LEFT)
		self.stop.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.stop, "Enter ending value or expression")

		# ---
		row += 1
		tk.Label(frame, text="Step:").grid(
			row=row, column=0, sticky=tk.E)
		self.step = tk.Entry(frame, background="White", justify=tk.LEFT)
		self.step.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.step, "Step value or expression")

		# ---
		row += 1
		b = tk.Checkbutton(frame, text="Use Parent as 1st iteration",
				variable=self.incparent,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Check to use the parent run as the first iteration of the loop.\n"
			"Uncheck to create a new run (suffix -1) for the first iteration")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		for run in runs:
			if run.isdefault():
				b.config(state=tk.DISABLED)
				break
		frame.grid_columnconfigure(1, weight=1)

		# ----
		frame = tk.Frame(self)
		tk.Button(frame, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel).pack(side=tk.RIGHT)
		tk.Button(frame, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok).pack(side=tk.RIGHT)
		self.bind("<Return>",  self.ok)
		self.bind("<KP_Enter>",self.ok)
		self.bind("<Escape>",  self.cancel)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		self.deiconify()
		# Get the focus
		self.focus_set()
		self.grab_set()
		self.wait_window()

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		var = self.variable.get()
		if var=="":
			messagebox.showerror("No variable selected",
				"Please select a valid variable first",
				parent=self)
			return

		inp = self.flair.project.input
		try:
			start = float(eval(self.start.get(),
					Evaluation._globalDict,
					inp.localDict))
		except:
			messagebox.showerror("Error start value",
				"Please enter a valid starting value",
				parent=self)
			return

		try:
			stop = float(eval(self.stop.get(),
					Evaluation._globalDict,
					inp.localDict))
		except:
			messagebox.showerror("Error stop value",
				"Please enter a valid stopping value",
				parent=self)
			return

		step = self.step.get()
		if step == "": step="1"
		try:
			step = float(eval(step,
					Evaluation._globalDict,
					inp.localDict))
		except:
			messagebox.showerror("Error step value",
				"Please enter a valid step value",
				parent=self)
			return

		if step==0.0: step = 1.0

		if start<stop and step<0.0:
			messagebox.showerror("Wrong step value",
				"Start<Stop and step<0. Please change sign of step",
				parent=self)
			return
		elif start>stop and step>0.0:
			messagebox.showerror("Wrong step value",
				"Start>Stop and step>0. Please change sign of step",
				parent=self)
			return

		# Find run entry positions
		self.findPositions()

		# Create runs
		val   = start
		first = True
		if step > 0.0:
			while val <= stop:
				self.createRuns(var, val, first)
				val += step
				first = False

		else:
			while val >= stop:
				self.createRuns(var, val, first)
				val += step
				first = False

		self.flair.modified()
		self.destroy()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self.destroy()

	# ----------------------------------------------------------------------
	def findPositions(self):
		project = self.flair.project
		sz  = len(project.runs) - 1
		self.pos = []
		for run in self.runs:
			pos = project.getRunIndex(run) + 1
			while pos<sz and project.runs[pos]("parent"):
				pos += 1
			self.pos.append(pos)

	# ----------------------------------------------------------------------
	def createRuns(self, var, val, first):
		try:
			f = float(val)
			if float(int(f)) == f:
				val = str(int(f))
			else:
				val = str(f)
		except:
			val = str(val)
		project = self.flair.project
		for i,run in enumerate(self.runs):
			if first and self.incparent.get():
				clone = run
			else:
				clone = run.clone()
				if run.isdefault():
					clone.name = run.inputBaseName()
					clone.defaultDefines()
				clone.name = project.findUniqueRunName(clone.name+"-1")

				if clone("title"):
					title = clone("title")
				else:
					title = project.title
				clone["title"] = f"{title} [{var}={val}]"

			# Check defines and append the var,value
			for j,nv in enumerate(clone.defineList):
				if nv[0]==var:
					del clone.defineList[j]
					break
			clone.defineList.append((var,val))

			if not (first and self.incparent.get()):
				project.addRun(clone, self.pos[i])

				# increase all positions higher than us by 1
				for k in range(i,len(self.pos)):
					self.pos[k] += 1
