# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   8-Jan-2019

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math

import log
import csg
import bmath
from Evaluation import fwhm

from deck    import Input, Card
from fluka   import Fluka
from moira   import Moira, MoiraCard
from element import Element

from .convert import Convert, register

#===============================================================================
# Conversion class from Fluka to Moira
#===============================================================================
class Fluka2Moira(Convert):
	src = Fluka.MODE
	dst = Moira.MODE
	INVMATRIX = "%s{-1}"	# pattern for inverse matrix lattices = []
	BEAMPOS_ORIENTATION = ("", "NEGATIVE")
	#BEAMPOS_VOLSOURCES  = ("FLOOD", "CART-VOL", "CYLI-VOL", "SPHE-VOL")

	#-----------------------------------------------------------------------
	def __init__(self, inp, defines=None):
		super().__init__(inp, defines)
		self.input.preprocess(defines)

		self._rotdefi = self.input.getTransformation()
		self.input.bodyProperties(self._rotdefi)
		self.input.regionProperties()#self._rotdefi)
		self._materials = {}		# keep track of defined materials
		self._rotid     = {}
		self.lattices   = []
		self.output     = None

	#-----------------------------------------------------------------------
	def addCard(self, tag, what=None, comment=""):
		card = MoiraCard(tag, what, comment)
		self.add(card)
		return card

	#-----------------------------------------------------------------------
	def material(self, name):
		try:
			return self._materials[name]
		except KeyError:
			try:
				g4mat = Moira.materialFromFluka(name)
				self._materials[name] = g4mat	# remember for next time
				return g4mat
			except KeyError:
				log.error(f"Material {name} was not converted to moira")
		return "G4_Galactic"

	#-----------------------------------------------------------------------
	# Convert materials from fluka to moira
	#-----------------------------------------------------------------------
	def convertMaterials(self):
		# 1st pass register all fluka defined materials
		for mat in self.input["MATERIAL"]:
			if mat.ignore(): continue
			# Check if material is an element existing in G4
			#if mat.what(2)!="" or mat.what(6)!="" or \
			#   not Moira.materialDefaultInFluka(mat.name()):
			self._materials[mat.name()] = mat.name()

		for mat in self.input["MATERIAL"]:
			if mat.ignore(): continue
			mat[Card.USED] = True
			name = mat.name()
			Z    = mat.intWhat(1)
			Am   = mat.intWhat(2)
			rho  = mat.numWhat(3)
			if mat.what(5): self.notImplemented(mat, 5)
			A    = mat.intWhat(6)
			if A != 0:
				self.notImplemented(mat, 6)
				#self.addCard("ISOT", [name, Z, Am, A])
				#self.addCard("MAT", [name, name,
			if Am:
				self.addCard("MATE", [name,Z,Am,rho], mat.comment())
			else:
				# Create mixture by finding the compound
				tag = None
				what = [name, rho, 0 ]	# 3rd element is the # of components
				for card in self.input["COMPOUND"]:
					card[Card.USED] = True
					if card.ignore(): continue
					if card.name() != name: continue
					# normalize fractions if needed
					total = abs(sum([card.numWhat(x) for x in range(1,card.nwhats(),2)]))
					for i in range(1,card.nwhats(),2):
						frac = card.numWhat(i)
						elem = card.what(i+1)
						if frac==0 or not elem: break
						if tag is None:
							if frac>0:
								tag = "MIXT_BY_NATOMS"
							elif elem[0]=="-":
								tag = "MIXT_BY_VOLUME"
							else:
								tag = "MIXT"
						if tag != "MIXT_BY_NATOMS":
							frac = bmath.format(abs(frac)/total)
						if elem[0]=="-": elem = elem[1:]
						what.append(self.material(elem))
						what.append(frac)
				if tag is not None:
					what[2] = (len(what)-3)//2
					self.addCard(tag, what, mat.comment())
				elif Z != 0:
					# Find A from database and create material
					element = Element.byZ(Z)
					self.addCard("MATE",
							[name,Z,element.Amass(),rho],
							mat.comment())

	#-----------------------------------------------------------------------
	def convertTransformations(self):
		# find unique names
		self._rotid = {}
		for name, rot in self._rotdefi.items():
			if not isinstance(rot,int): continue
			if rot in self._rotid: continue
			self._rotid[rot] = name
			rotm = self._rotdefi[rot]
			#rx,ry,rz = [bmath.format(math.degrees(x))
			#		for x in rotm.getEulerRotation()]
			#self.addCard("ROTM", [name, rx, ry, rz])
			what = [name]
			for i in range(3):
				what.extend([bmath.format(x) for x in rotm[i][:3]])
			self.addCard("ROTM", what)
			#print()
			#print(name)
			#print("M=",rotm)
			#print("Rxyz=",rx,ry,rz, "T=",rotm[0][3],rotm[1][3],rotm[2][3])

			# and the inverse matrix
			roti = rotm.clone()
			roti.inv()
			#rx,ry,rz = [bmath.format(math.degrees(x),16)
			#		for x in roti.getEulerRotation()]
			#self.addCard("ROTM", [Fluka2Moira.INVMATRIX%(name), rx, ry, rz])
			what = [Fluka2Moira.INVMATRIX%(name)]
			for i in range(3):
				what.extend([bmath.format(x) for x in roti[i][:3]])
			self.addCard("ROTM", what)
			#print("I=",roti)
			#print("Rxyz=",rx,ry,rz, "T=",roti[0][3],roti[1][3],roti[2][3])
			#print()

	#-----------------------------------------------------------------------
	#def findBody(self, name):
	#	for t in CardInfo.bodies[self.input.MODE]:
	#		for x in self.input[t]:
	#			if x.notIgnore() and x.name()==name:
	#				return x

	#-----------------------------------------------------------------------
	# --- Find world and void ---
	# Find outer blackbody
	# FIXME very very primitive
	# it will search for:
	#  1. regions with BLCKHOLE as material
	#  2. name: BLKBODY
	#  3. in expression to find +BODY -[...]
	#     where BODY is referred only in this expression
	#  4. typically the first one or the last one
	# it should be the biggest region with
	# BLACKHOLE material and typically a simple expression +BLK -VOID
	# where both BLK and VOID are simple objects RPP or BOX.
	#-----------------------------------------------------------------------
	def findBlackBody(self):
		# build dictionary with biggest dimension
		bodyMax = {}
		bodies = self.input.cardsCache(Card.BODY)
		for name,body in bodies.items():
			m = max([abs(body.numWhat(x))
						for x in range(1,body.nwhats())])
			bodyMax[name] = m, body

		# For all regions with BLCKHOLE as material
		# find the biggest body
		regionMax = 0
		blkregion = None
		for region in self.input["REGION"]:
			if region.ignore(): continue
			material = region[Card.MATERIAL]
			if material is not None and \
			   material.name() == "BLCKHOLE":
				for token in csg.tokenize(region.expr()):
					if token in ("(","+","-","|",")"): continue
					try:
						r, body = bodyMax[token]
						if r>regionMax:
							regionMax = r
							blkregion = region
					except KeyError:
						pass

		if blkregion is None:
			raise Exception("Black body search didn't work out!")

		# Possible patterns of blkregion
		# 1. +blkbody -world
		expr = "".join(blkregion.expr().split())
		m = re.match(r"^\+(\w+)-(\w+)$", expr)
		if m:
			world = bodies[m.group(2)]
		else:
			# 2. -blkbody
			m = re.match(r"^-(\w+)$", blkregion.expr())
			if m:
				world = bodies[m.group(1)]
			else:
				# 3. +blkbody -body1 -body2 ....
				m = re.match(r"^\+(\w+)\b", expr)
				if m:
					world = bodies[m.group(1)]
				else:
					raise Exception("Black body search didn't work out!")

		if world.tag not in ("RPP","SPH"):
			raise Exception("Black body seems it is not RPP or SPH")
		log.say("Finding blackbody:")
		log.say("\tBlkregion=", blkregion)
		log.say("\tWorld=", world)
		return blkregion, world

	#-----------------------------------------------------------------------
	def convertBodies(self, world):
		matrix = None
		numMat=0
		for card in self.input.cardlist:
			if card.ignore():    continue
			if not card.isGeo(): continue
			#if card is world:
			#	card[Card.USED] = True
			#elif card.type() == Card.BODY:
			if card.type() == Card.BODY:
				if card[Card.MATRIX] is not matrix:
					matrix = card[Card.MATRIX]
					if matrix is not None:
						#FIXME Find a better idea..
						what = [f"MATRIX{numMat}"]
						numMat += 1
						for row in matrix[:-1]: # skip last 0 0 0 1 row
							what.extend(row)
						self.add(MoiraCard("MATRIX", what))
					else:
						self.add(MoiraCard("MATRIX"))
				self.add(MoiraCard(card.tag, card.whats()[:], card.comment()))
				card[Card.USED] = True
			elif card.tag[0] == "#":
				self.add(card.clone())
				card[Card.USED] = True
			#elif card.tag[0] == "$":
			#	card[Card.USED] = True

	#-----------------------------------------------------------------------
	# Convert them to volumes
	#-----------------------------------------------------------------------
	def volumeCard(self, body, parent=None, mat="G4_Galactic", comment=""):
		name = body.name()
		if body.tag == "SPH":
			# Comm
			card = MoiraCard("VOLU.ORB", [name, mat, body.numWhat(4)], comment)
			x = body.numWhat(1)
			y = body.numWhat(2)
			z = body.numWhat(3)
		elif body.tag == "RPP":
			dx = (body.numWhat(2)-body.numWhat(1))/2
			dy = (body.numWhat(4)-body.numWhat(3))/2
			dz = (body.numWhat(4)-body.numWhat(5))/2
			card = MoiraCard("VOLU.BOX", [name, mat, dx, dy, dz], comment)
			x = bmath.format((body.numWhat(2)+body.numWhat(1))/2)
			y = bmath.format((body.numWhat(4)+body.numWhat(3))/2)
			z = bmath.format((body.numWhat(4)+body.numWhat(5))/2)
		else:
			raise Exception(name+" with type "+body.tag+" is not supported as volume")
		self.add(card)
		if parent:
			card = MoiraCard("PLACE", [name, 1, parent, "", x, y, z], comment)
			self.add(card)

	#-----------------------------------------------------------------------
	# Regions & Volumes
	#-----------------------------------------------------------------------
	def convertRegions(self, blkregion, world):
		# Write the basic volumes
		self.volumeCard(world, comment=world.comment())
		#self.writeline(":VIS %s OFF"%(world.name()))

		#self.volumeCard(void, world.name(), comment=void.comment())
		#self.writeline(":VIS %s OFF"%(void.name()))

		card = None
		blkregion[Card.USED] = True
		try:
			blkregion[Card.ASSIGNMAT][Card.USED] = True
		except (KeyError, TypeError):
			pass
		for region in self.input["REGION"]:
			if region.ignore() or region==blkregion: continue
			material = region[Card.MATERIAL]
			try:
				region[Card.ASSIGNMAT][Card.USED] = True
			except (KeyError, TypeError):
				pass
			expr = "\n".join([x.strip() for x in region.expr().splitlines()])
			if region.name()=="&":
				# continuation of previous card
				card.setComment(card.comment()+"\n"+region.comment())
				card.setExpr(card.expr()+"\n"+expr)
				region[Card.USED] = True
			else:
				lattice = region[Card.LATTICE]
				if lattice:
					g4mat = ""
				elif material:
					g4mat = self.material(material.name())
				else:
					g4mat = "G4_Galactic"
				# FIXME
				#field = region[Card.ASSIGNMAT].intWhat(5)
				card = MoiraCard("CELL",[region.name(),
							world.name(),
							g4mat],
							#field],
						region.comment())
				card.setExpr(expr)
				self.add(card)
				region[Card.USED] = True
				if lattice:
					name = lattice.name()
					inv = True
					if name[0] == "-":
						inv = not inv
						name = name[1:]
					rotm = self.input.getTransformation(name, self._rotdefi, inv)
					# try to get the real name used based on the id
					rid  = self._rotdefi[name]
					name = self._rotid[rid]
					if inv: name = Fluka2Moira.INVMATRIX%(name)
					dx = rotm[0][3]
					dy = rotm[1][3]
					dz = rotm[2][3]
					self.addCard("LATTICE",
							[region.name(), name,
							dx, dy, dz],
							lattice.comment())
					self.lattices.append(region.name())

	#-----------------------------------------------------------------------
	# Process all other cards (except geometry)
	#-----------------------------------------------------------------------
	def convertAllOtherCards(self):
		self._beamposShape = ""
		self._beamPos = bmath.Vector(0,0,0)
		self._beamDir = bmath.Vector(0,0,1)
		INTER2D = ["MB", "MBRB", "MBHDP", "MBW", "MCBC" ]
		INTER3D = ["FCCdet"]
		QUADINT = ["MQW", "MQXA", "MQXB", "MQXF", "MQXFv3", "MQY97mm", "MQTL", "MQY"]
		SPECIAL = ["MB", "MBW"]
		KICKINT = ["MBXF", "MBXF_v3", "MBXW"]

		# Pre-processing
		#fields = []
		for card in self.input.cardlist:
			if card.ignore(): continue
			if card.tag == "BEAMPOS":
				self._beamposShape = card.sdum()
			#elif card.tag == "LATTICE":
			elif card.tag == "USRGCALL":
				fields.append(card.sdum())
			elif card.tag == "USERDUMP":
				self.add(MoiraCard("/tally/userdump", [card.sdum(),""]))

		#fields = list(set(fields))
		#for fld in fields:
		#	if fld in (INTER2D, KICKINT, QUADINT, SPECIAL):
		#		new = MoiraCard("FIELD.B.INTER",
		#				[fld, "2DInterpolated"],
		#				card.comment())
		#		new["macroPath"] = "*pathtofieldmaps*"
		#		new["execute"]   = f"{fld}.mac"
		#		self.add(new)
		#	elif fld in INTER3D:
		#		new = MoiraCard("FIELD.B.INTER",
		#				[fld, "3DInterpolated" ],
		#				card.comment())
		#		new["macroPath"] = "*pathtofieldmaps*"
		#		new["execute"]   = f"{fld}.mac"
		#		self.add(new)
		#	else:
		#		# FIXME BNV ?????? card with empty tag?
		#		new = MoiraCard("",
		#				[ "CONST", "constant" ],
		#				card.comment())
		#		new["macroPath"] = "*pathtofieldmaps*"
		#		new["execute"]   = f"{fld}.mac"
		#		self.add(new)

		# Loop over all cards to convert
		convertFunc = { "BEAM"     : Fluka2Moira._convertBEAM,
				"BEAMPOS"  : Fluka2Moira._convertBEAMPOS,
				"BEAMAXES" : Fluka2Moira._convertBEAMAXES,
				"BIASING"  : Fluka2Moira._convertBIASING,
				"USRGCALL" : Fluka2Moira._convertUSRGCALL,
				"USRBIN"   : Fluka2Moira._convertUSRBIN,
				"RANDOMIZ" : Fluka2Moira._convertRANDOMIZ,
				"START"    : Fluka2Moira._convertSTART,
				}
		for card in self.input.cardlist:
			if isinstance(card, MoiraCard):
				self.add(card.clone())
				continue
			if card.ignore(): continue

			try:
				convertFunc[card.tag](self, card)
				continue
			except KeyError:
				pass

			if card[Card.USED] or \
			   card.tag in ("GEOBEGIN", "GEOEND", "STOP", "END"):
				pass
			else:
				clone = card.clone()
				clone.setEnable(False)
				self.add(clone)

	#-----------------------------------------------------------------------
	# Card conversion methods
	#-----------------------------------------------------------------------
	def _convertBEAM(self, card):
		particle = card.name()
		if particle in ("ISOTOPE","HEAVYION"):
			self.notImplemented(card,0)
			return
		self.add(MoiraCard("/run/initialize"))
		self.add(MoiraCard("/primary/select",["gps"],
				"Primary particle generator"))
		self.add(MoiraCard("/gps/particle",
				[Moira.particleFromFluka(particle)],
				card.comment()))

		# We should keep the formulas if any
		energy = card.numWhat(1)	# or momentum
		dE     = card.numWhat(2)
		dphi   = card.numWhat(3)
		dx     = card.numWhat(4)
		dy     = card.numWhat(5)
		shape  = card.numWhat(6)
		if energy > 0:
			# FIXME Energy TODO conversion from p to E
			self.notImplemented(card,"conversion from momentum")
		else:
			energy = -energy
		self.add(MoiraCard("/gps/energy",[energy,"GeV"]))
		# Energy/Momentum spread
		if dE < 0:
			self.add(MoiraCard("/gps/ene/type",["Gauss"]))
			self.add(MoiraCard("/gps/ene/sigma",[-dE/fwhm, "GeV"]))
		else:
			self.add(MoiraCard("/gps/ene/type",["Mono"]))
			# XXX FLUKA the Flat case?
			self.add(MoiraCard("/gps/ene/min",[energy-dE, "GeV"]))
			self.add(MoiraCard("/gps/ene/max",[energy+dE, "GeV"]))
		# Divergence
		if dphi < 0:
			self.add(MoiraCard("/gps/ang/type",["beam1d"]))
			self.add(MoiraCard("/gps/ang/sigma_r",[-dphi/fwhm, "mrad"]))
		elif dphi > 1000:
			self.add(MoiraCard("/gps/ang/type",["iso"]))
		else:
			self.add(MoiraCard("/gps/ang/type",["iso"]))
			self.add(MoiraCard("/gps/ang/maxtheta",[dphi/2, "mrad"]))
		# Beam Shape
		if self._beamposShape in Fluka2Moira.BEAMPOS_ORIENTATION:
			if shape < 0:
				self.add(MoiraCard("/gps/pos/type",["Plane"]))
				self.add(MoiraCard("/gps/pos/shape",["Annulus"]))
				self.add(MoiraCard("/gps/pos/radius",[dy,"cm"]))
				self.add(MoiraCard("/gps/pos/inner_radius",[dx,"cm"]))
			# elif shape > 0:
			else:
				self.add(MoiraCard("/gps/pos/type",["Beam"]))
				if dx < 0:
					self.add(MoiraCard("/gps/pos/sigma_x",[-dx/fwhm, "cm"]))
				else:
					self.add(MoiraCard("/gps/pos/halfx",[dx/2, "cm"]))
				if dy < 0:
					self.add(MoiraCard("/gps/pos/sigma_y",[-dy/fwhm, "cm"]))
				else:
					self.add(MoiraCard("/gps/pos/halfy",[dy/2, "cm"]))
			# else:
				# self.add(MoiraCard("/gps/pos/type",["Point"]))

	#-----------------------------------------------------------------------
	def _convertBEAMPOS(self, card):
		if card.sdum() in Fluka2Moira.BEAMPOS_ORIENTATION:
			x = card.numWhat(1)
			y = card.numWhat(2)
			z = card.numWhat(3)
			self._beamPos = bmath.Vector(x,y,z)
			self.add(MoiraCard("/gps/pos/centre",[x,y,z,"cm"],card.comment()))
			u = card.numWhat(4)
			v = card.numWhat(5)
			tsqr = math.sqrt(u*u + v*v)	# do it ala fluka beacrd.f
			if abs(u)>1 or abs(v)>1 or abs(tsqr)>1:
				u /= tsqr
				v /= tsqr
				w = 0
			else:
				w = math.sqrt(max(0, (1-tsqr)*(1+tsqr)))
				if card.sdum()=="NEGATIVE":
					w = -w
			self._beamDir = bmath.Vector(u, v, w)
			self._beamDir.normalize()

			if self.input["BEAMAXES"]:
				# skip beam direction it will be set by the beam axes
				pass
			elif u or v:
				X = self._beamDir.orthogonal()
				X.normalize()
				Y = self._beamDir ^ X
				self.add(MoiraCard("/gps/ang/rot1", [Y[0], Y[1], Y[2]],
						card.comment()))
				self.add(MoiraCard("/gps/ang/rot2", [X[0], X[1], X[2]]))
			elif card.sdum()!="":
				self.add(MoiraCard("/gps/ang/rot1",[1,  0, 0], card.comment()))
				self.add(MoiraCard("/gps/ang/rot2",[0, -1, 0]))
		elif card.sdum() == "FLOOD":
			r = card.numWhat(1)
			self.add(MoiraCard("/gps/pos/type",  ["Surface"], card.comment()))
			self.add(MoiraCard("/gps/pos/shape", ["Sphere"]))
			self.add(MoiraCard("/gps/pos/radius",[r, "cm"]))
			self.add(MoiraCard("/gps/pos/centre",
					[self.beamPos[0], self.beamPos[1], self.beamPos[2], "cm"]))
		elif card.sdum() == "CART-VOL":
			Vin  = bmath.Vector(card.numWhat(1),card.numWhat(3),card.numWhat(5))
			Vout = bmath.Vector(card.numWhat(2),card.numWhat(4),card.numWhat(6))
			half = (Vout - Vin)/2
			pos = self._beamPos + (Vout + Vin)/2
			# self.add(MoiraCard("/gps/pos/type",["Volume"]))
			self.add(MoiraCard("/gps/pos/type",["Beam"], card.comment()))
			self.add(MoiraCard("/gps/pos/halfx",[half[0], "cm"]))
			self.add(MoiraCard("/gps/pos/halfy",[half[1], "cm"]))
			self.add(MoiraCard("/gps/pos/halfz",[half[2], "cm"]))
			self.add(MoiraCard("/gps/pos/centre",[pos[0], pos[1], pos[2], "cm"]))
		elif card.sdum() == "CYLI-VOL":
			rin  = card.numWhat(1)
			rout = card.numWhat(2)
			hin  = card.numWhat(3)
			hout = card.numWhat(4)
			# XXX Or volume?
			self.add(MoiraCard("/gps/pos/type",["Surface"], card.comment()))
			# self.add(MoiraCard("/gps/pos/type",["Volume"])) # XXX Or volume?
			self.add(MoiraCard("/gps/pos/shape",["Cylinder"]))
			self.add(MoiraCard("/gps/pos/radius",[rout, "cm"]))
			self.add(MoiraCard("/gps/pos/halfz",[(hout-hin)/2, "cm"]))
			self.add(MoiraCard("/gps/pos/centre",
					[self._beamPos[0],
					 self._beamPos[1],
					 self._beamPos[2]+(hout+hin)/2,
					 "cm"]))
		elif card.sdum() == "SPHE-VOL":
			rin  = card.numWhat(1)
			rout = card.numWhat(2)
			self.add(MoiraCard("/gps/pos/type",["Volume"], card.comment()))
			self.add(MoiraCard("/gps/pos/shape",["Sphere"]))
			self.add(MoiraCard("/gps/pos/radius",[rout, "cm"]))
			self.add(MoiraCard("/gps/pos/centre",
					[self.beamPos[0], self.beamPos[1], self.beamPos[2], "cm"]))
		else:
			self.notImplemented(card,0,f"uknown SDUM={card.sdum()}")

	#-----------------------------------------------------------------------
	def _convertBEAMAXES(self, card):
		X = bmath.Vector(card.numWhat(1), card.numWhat(2), card.numWhat(3))
		if X.normalize() < 1e-10:
			X = bmath.Vector.Xo
		Z = bmath.Vector(card.numWhat(4), card.numWhat(5), card.numWhat(6))
		if Z.normalize() < 1e-10:
			Z = bmath.Vector.Zo
		Y = Z ^ X
		X = Y ^ Z
		self.add(MoiraCard("/gps/pos/rot1", [X[0], X[1], X[2]]))
		self.add(MoiraCard("/gps/pos/rot2", [Y[0], Y[1], Y[2]]))
		# FIXME to be verified!!!
		self._beamDir = bmath.Vector(
				X[0]*self._beamDir[0]+Y[0]*self._beamDir[1]+Z[0]*self._beamDir[2],
				X[1]*self._beamDir[0]+Y[1]*self._beamDir[1]+Z[1]*self._beamDir[2],
				X[2]*self._beamDir[0]+Y[2]*self._beamDir[1]+Z[2]*self._beamDir[2])
		self._beamDir.normalize()
		X = self._beamDir.orthogonal()
		X.normalize()
		Y = self._beamDir ^ X
		self.add(MoiraCard("/gps/ang/rot1", [Y[0], Y[1], Y[2]]))
		self.add(MoiraCard("/gps/ang/rot2", [X[0], X[1], X[2]]))

	#-----------------------------------------------------------------------
	def _convertBIASING(self, card):
		part  = card.what(1)
		imp   = card.what(3)
		reg   = card.what(4)
		if part == "0": part = "all"
		if part == "1": part = "hadrons"
		if part == "2": part = "em"
		if part == "3": part = "neutrons"
		self.add(MoiraCard("/bias/region",[reg,part,imp],card.comment()))

	#-----------------------------------------------------------------------
	# Magnetic Fields
	# FIXME this specific to our BMI interpretation
	#       we should make somehow a plugin
	#-----------------------------------------------------------------------
	def _convertUSRGCALL(self, card):
		reg   = card.intWhat(2) - 1
		step  = card.what(4)
		grad  = card.what(5)
		offX  = card.what(7)
		offY  = card.what(8)
		offZ  = card.what(9)
		alpha = card.what(10)
		beta  = card.what(11)
		gamma = card.what(12)
		if card.sdum() == "&": return	# FIXME
		fldtype = card.sdum()
		# if fldtype in INTER2D:
		new = MoiraCard("FIELD.REG",
				[ self.lattices[reg], fldtype, "", ""],
				card.comment())
		# new["B"] = fldtype
		# new["E"] = ""
		# new["G"] = ""
		new["stepper"] = "DormandPrince745"
		new["epsilon.emin"] = ""
		new["epsilon.emax"] = ""
		# new["minStep"] = ""
		# new["deltastep"] = ""
		# new["deltastep.Unit"] = ""
		# new["deltaint"] = ""
		# new["deltaint.Unit"] = ""
		self.add(new)
		new = MoiraCard("&FIELD.REG",
				[grad , "tesla"],
				card.comment())
		new["rot.alpha"] = alpha
		new["rot.beta"] = beta
		new["rot.gamma"] = gamma
		new["offset.Xo"] = offX
		new["offset.Yo"] = offY
		new["offset.Zo"] = offZ
		self.add(new)

	#-----------------------------------------------------------------------
	def _convertUSRBIN(self, card):
		# FIXME
		# - combine same dimension/type meshes under common mesh
		name  = card.name()
		utype = card.intWhat(1)
		#part = card.what(2)
		#unit = abs(card.intWhat(3))
		xmax  = card.numWhat(4)
		ymax  = card.numWhat(5)
		zmax  = card.numWhat(6)
		xmin  = card.numWhat(7)
		ymin  = card.numWhat(8)
		zmin  = card.numWhat(9)
		nx    = max(card.intWhat(10),1)
		ny    = max(card.intWhat(11),1)
		nz    = max(card.intWhat(12),1)
		if utype in (0,10):
			new = MoiraCard("MESH.BOX",[name],card.comment())
			x  = (xmin+xmax)/2
			y  = (ymin+ymax)/2
			z  = (zmin+zmax)/2
			new["boxSize.dx"] = (xmax-xmin)/2
			new["boxSize.dy"] = (ymax-ymin)/2
			new["boxSize.dz"] = (zmax-zmin)/2
			new["boxSize.Unit"] = "cm"
			new["nBin.Nx"] = nx
			new["nBin.Ny"] = ny
			new["nBin.Nz"] = nz
		elif utype in (1,11):
			new = MoiraCard("MESH.CYLINDER",[name],card.comment())
			# FIXME no rmin
			#r  = xmax
			x  = ymin
			y  = ymax
			z  = (zmin+zmax)/2
			new["cylinderSize.R"]    = xmax
			new["cylinderSize.dz"]   = (zmax-zmin)/2
			new["cylinderSize.Unit"] = "cm"
			new["nBin.Nr"]   = nx
			new["nBin.Nz"]   = nz
			new["nBin.Nphi"] = ny
		else:
			self.notImplemented(card,1)
			clone = card.clone()
			clone.setEnable(False)
			self.add(clone)
			return

		new["xyz.X"] = x
		new["xyz.Y"] = y
		new["xyz.Z"] = z
		new["xyz.Unit"] = "cm"
		self.add(new)

		# FIXME create groups of particles as particle-lists
		#
		#/score/quantity/energyDeposit enedep MeV
		#/score/quantity/cellFlux neutron percm2
		#/score/filter/particle . neutron
		#/score/quantity/cellFlux electron percm2
		#/score/filter/particle . e+ e-
		self.add(MoiraCard("/score/close"))

	#-----------------------------------------------------------------------
	def _convertRANDOMIZ(self, card):
		seed = f"{card.what(2)} {card.what(2)} {card.what(2)}".strip()
		self.add(MoiraCard("/random/setSeeds",[seed]))

	#-----------------------------------------------------------------------
	def _convertSTART(self, card):
		self.add(MoiraCard("/run/beamOn",[card.intWhat(1)]))

	#-----------------------------------------------------------------------
	# FIXME some automatic unit conversion system!
	#-----------------------------------------------------------------------
	def convert(self):
		self.output = Moira()
		try:
			card = self.input["TITLE"][0]
			card[Card.USED] = True
			self.addCard("/moira/title",[card.text()],card.comment())
		except IndexError:
			pass
		self.addCard("/physics_lists/select")
		self.convertMaterials()
		self.convertTransformations()
		blkregion, world = self.findBlackBody()
		self.convertBodies(world)
		self.convertRegions(blkregion, world)
		self.convertAllOtherCards()
		return self.output

#-------------------------------------------------------------------------------
# Execute only once and register the conversion program
#-------------------------------------------------------------------------------
register(Fluka.MODE, Moira.MODE, Fluka2Moira)
