# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   23-Nov-2018

import os
import re

import tool
from deck import Config

#===============================================================================
# A container class to keep the programs
# It should contain only static and class methods
#===============================================================================
class FlukaConfig(Config):
	SECTION   = "Fluka"	# Configuration section name

	DEFAULT_EXE   = "fluka"
	DEFAULT_EXE2  = "flukadpm"
	DEFAULT_LIB   = "libfluka.a"

	CYCLE         = "fluka_"	# prefix of temporary cycle directory

	directory     = ""		# Executable installation directory
	program       = DEFAULT_EXE	# Main application program
	version       = "Unknown"
	released      = ""
	user          = {}		# Description of user routines
	VERSION_PAT   = re.compile(r".*^Version: (\d.*?)$.Date: *(.*?)$",
				re.MULTILINE | re.DOTALL)
	VERSION2_PAT  = re.compile(r".*^The latest version is.*?: (\d+..*?)$.*?: (.*?)$",
				re.MULTILINE | re.DOTALL)
	CYCLEDIR_PAT  = re.compile(r"^%s(\d+)$"%(CYCLE))
	CYCLEDIR2_PAT = re.compile(r"^(.*/?)%s(\d+)$"%(CYCLE))

	_checkedVar = None			# Is directory already checked?

	# fluka programs, the name is the default program
	programs  = {   "fff"       : "",
			"rfluka"    : "",
			"detsuw"    : "",
			"usbsuw"    : "",
			"ustsuw"    : "",
			"usxsuw"    : "",
			"usbmax"    : "",
			"usrsuw"    : "",
			"usbrea"    : "",
			"gplevbin"  : "",
			"detect"    : "detsuw",
			"resnucle"  : "usrsuw",
			"usrbin"    : "usbsuw",
			"usrbdx"    : "usxsuw",
			"usrtrack"  : "ustsuw",
			"usryield"  : "usysuw",
			"usrcoll"   : "ustsuw",
		}

	#-----------------------------------------------------------------------
	# Find FLUKA version
	#-----------------------------------------------------------------------
	@staticmethod
	def findVersion(directory=None):
		if directory is None:
			d = FlukaConfig.directory
		else:
			d = directory
		try:
			with open(os.path.join(d,"Version.tag"),"r") as f:
				m = FlukaConfig.VERSION_PAT.match(f.read())
				version  = None
				released = None
				if m:
					version, released = m.groups()
				else:
					m = FlukaConfig.VERSION2_PAT.match(f.read())
					if m:
						version, released = m.groups()
				if version is not None:
					FlukaConfig.version = version
					FlukaConfig.released = released
			return version
		except IOError:
			return "Unknown"

	#-----------------------------------------------------------------------
	@staticmethod
	def setFlukaDir(path=""):
		FlukaConfig.directory = path
		if FlukaConfig.directory == "":
			path = tool.findExe("fluka")
			if path:	# bin was given set the parent as directory
				FlukaConfig.directory = os.path.dirname(path)
		else:
			# Check if the bin was provided instead of the master dir
			exefile = os.path.join(FlukaConfig.directory, "fluka")
			if os.path.isfile(exefile) and os.access(exefile, os.X_OK):
				FlukaConfig.directory = os.path.dirname(FlukaConfig.directory)

	#-----------------------------------------------------------------------
	@staticmethod
	def checkDir():
		return not os.path.isdir(FlukaConfig.directory)

	#-----------------------------------------------------------------------
	# Directories
	#-----------------------------------------------------------------------
	@staticmethod
	def binDir():		return os.path.join(FlukaConfig.directory,"bin")
	@staticmethod
	def documentDir():	return os.path.join(FlukaConfig.directory,"doc")
	@staticmethod
	def includeDir():	return os.path.join(FlukaConfig.directory,"include")

	#-----------------------------------------------------------------------
	# Return a fluka command, default if cmd=empty
	#-----------------------------------------------------------------------
	@staticmethod
	def command(prg):
		path = FlukaConfig.programs.get(prg, "")
		if FlukaConfig.checkDir(): return path
		if not path:
			# Try fluka directory
			if os.path.basename(FlukaConfig.directory) == "bin":
				cmd = os.path.join(FlukaConfig.directory, prg)
			else:
				cmd = os.path.join(FlukaConfig.directory, "bin", prg)
			if os.path.isfile(cmd):
				return cmd
			return Config.command(prg)
		elif path[0] == os.path.sep:
			return path
		elif os.path.basename(FlukaConfig.directory) == "bin":
			return os.path.join(FlukaConfig.directory, path)
		else:
			return os.path.join(FlukaConfig.directory, "bin", path)

	#-----------------------------------------------------------------------
	# Load from config
	#-----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		section  = FlukaConfig.SECTION
		FlukaConfig.setFlukaDir(config.getStr(section, "directory", ""))
		FlukaConfig.program = config.getStr(section, "program", FlukaConfig.DEFAULT_EXE)

		FlukaConfig.setCmd("rfluka",  config.getStr(section, "rfluka",  ""))
		FlukaConfig.setCmd("fff",     config.getStr(section, "fff",     ""))
		FlukaConfig.setCmd("usbsuw",  config.getStr(section, "usbsuw",  ""))
		FlukaConfig.setCmd("ustsuw",  config.getStr(section, "ustsuw",  ""))
		FlukaConfig.setCmd("usxsuw",  config.getStr(section, "usxsuw",  ""))
		FlukaConfig.setCmd("usrsuw",  config.getStr(section, "usrsuw",  ""))
		FlukaConfig.setCmd("usysuw",  config.getStr(section, "usysuw",  ""))
		FlukaConfig.setCmd("detsuw",  config.getStr(section, "detsuw",  ""))
		FlukaConfig.setCmd("usbmax",  config.getStr(section, "usbmax",  ""))
		FlukaConfig.setCmd("usbrea",  config.getStr(section, "usbrea",  ""))
		FlukaConfig.setCmd("gplevbin",config.getStr(section, "gplevbin",""))

	#-----------------------------------------------------------------------
	@staticmethod
	def setCmd(cmd, val):
		val = val.strip()
		if not val: val = cmd
		path = os.path.dirname(val)
		if not path or path==os.path.join(FlukaConfig.directory, "bin"):
			FlukaConfig.programs[cmd] = os.path.basename(val)
		else:
			FlukaConfig.programs[cmd] = val

	#-----------------------------------------------------------------------
	# Save to config
	#-----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass
