# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   11-Jun-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import configparser
from legacy import cmp, cmp_to_key

#-------------------------------------------------------------------------------
def _cmpA(x, y):
	x = x[0]
	y = y[0]
	try:
		ix = int(x)
		nx = ""
	except:
		ix = int(x[:-1])
		nx = x[-1:]
	try:
		iy = int(y)
		ny = ""
	except:
		iy = int(y[:-1])
		ny = y[-1:]

	c = cmp(ix, iy)
	if c == 0:
		return cmp(nx, ny)
	return c

#===============================================================================
# Element Class
#===============================================================================
class Element:
	elements = []		# database of elements sorted by Z-1
	footnote = {}

	# ----------------------------------------------------------------------
	def __init__(self, symbol, Z):
		self.Z          = Z
		self.symbol     = symbol
		self.name       = ""
		self.AmassStr   = ""
		self.densityStr = ""
		self.melting    = ""
		self.boiling    = ""
		self.oxidation  = ""
		self.isotopes   = []

	# ----------------------------------------------------------------------
	def inputName(self, mode=None):
		if mode == "fluka":
			ename = self.name.upper()[:8]
			if ename=="PHOSPHOR": ename = "PHOSPHO"
			return ename
		return self.name

	# ----------------------------------------------------------------------
	def Amass(self):
		return float(self.AmassStr.split()[0])

	# ----------------------------------------------------------------------
	def density(self):
		return float(self.densityStr.split()[0])

	# ----------------------------------------------------------------------
	def densityNote(self):
		d = self.densityStr.split()
		if d: return d[-1]
		return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def byZ(Z):
		try:
			return Element.elements[Z-1]
		except IndexError:
			return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def bySymbol(symbol):
		symbol = symbol.upper()
		for element in Element.elements:
			if element and element.symbol.upper() == symbol:
				return element
		return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def byName(name):
		name = name.upper()
		for element in Element.elements:
			if element and element.name.upper() == name:
				return element
		return None

	# ----------------------------------------------------------------------
	@staticmethod
	def init(filename):
		global _elements, _footnote
		c = configparser.RawConfigParser()
		c.read(filename)

		# Read Elements
		Element.elements = [None] * len(c.sections())
		for symb in c.sections():
			if symb[0].islower(): continue
			Z         = c.getint(symb, "Z")
			elem      = Element(symb, Z)
			elem.name = c.get(symb, "name")
			try:	elem.AmassStr   = c.get(symb, "Amass")
			except: pass
			try:	elem.densityStr = c.get(symb, "density")
			except: pass
			try:	elem.melting = c.get(symb, "melting")
			except: pass
			try:	elem.boiling = c.get(symb, "boiling")
			except: pass
			try:	elem.oxidation = c.get(symb, "oxidation")
			except: pass

			elem.isotopes = []
			for item,value in c.items(symb):
				if item.startswith("isotope."):
					iso = item.split(".")[1]
					if "\n" in value:
						for line in value.splitlines():
							elem.isotopes.append((iso, line[1:]))
					else:
						elem.isotopes.append((iso, value[1:]))
			elem.isotopes.sort(key=cmp_to_key(_cmpA))
			Element.elements[Z-1] = elem

		# Remove all None parts (Do I need that?)
#		try:
#			while True: Element.elements.remove(None)
#		except ValueError:
#			pass

		for (i,n) in c.items("footnotes"):
			Element.footnote[i] = n #.replace("\n"," ")

		del c
