# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   14-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import tkinter as tk
import tkinter.messagebox as messagebox

import tkTree
import tkExtra
import bFileDialog
import tkTextEditor

_DEFAULT_SPLIT = 150

#===============================================================================
# OutputViewer Node
#===============================================================================
class OVNode(tkTree.Node):
	def __init__(self, *args, **kw_args):
		super().__init__(*args, **kw_args)
		self.widget.tag_bind(self.label,  '<1>', self.select)
		self.widget.tag_bind(self.symbol, '<1>', self.selectImage)
		self.widget.tag_bind(self.label,  '<3>', self.popupMenu)

	# ----------------------------------------------------------------------
	def select(self, event=None):
		tkTree.Node.select(self)
		self.PVT_enter(event)
		self.widget.move_cursor(self)
		self.showSection()

	# ----------------------------------------------------------------------
	def selectImage(self, event=None):
		self.PVT_click(event)
		self.select(event)

	# ----------------------------------------------------------------------
	def showSection(self):
		self.widget.viewer.showSection(self.full_id())
		self.widget.viewer.splitter.rightFrame().focus_set()

	# ----------------------------------------------------------------------
	def popupMenu(self, event=None):
		self.select()
		menu=tk.Menu(self.widget, tearoff=0)
		menu.add_command(label='Reload', command=self.widget.viewer.reloadSelected)
		menu.add_command(label='Close', command=self.widget.viewer.close)
		menu.tk_popup(event.x_root, event.y_root)

#===============================================================================
# File Data Viewer
#===============================================================================
class FileData:
	def __init__(self, filename, maxsize=None):
		self.filename = filename
		self.fn, self.ext = os.path.splitext(filename)
		self.maxsize  = maxsize
		self.filedata = ""
		self.mtime    = 0
		self.ypos     = 0.0

		# Sections contains a dictionary with the range in the file
		# to display: self.sections[idname] = (from, to)
		self.sections = {}
		# Tree contains a nested list of lists to construct the
		# displayed tree
		# [ (title, idname, None|another-tree),... ]
		self.tree     = []

	# ----------------------------------------------------------------------
	# Returns True on failure
	# ----------------------------------------------------------------------
	def load(self):
		try:
			stat = os.stat(self.filename)
			size = stat.st_size
			self.mtime = stat.st_mtime
			if self.maxsize is not None:
				if size >= self.maxsize: return True
			with open(self.filename, "r",
					encoding="utf-8",
					errors="replace") as f:
				self.filedata = f.read()
		except (OSError, IOError):
			return True
		except (OverflowError, MemoryError):
			return True
		return False

	# ----------------------------------------------------------------------
	def set(self, text):
		self.filedata = text
		self.mtime    = time.time()

	# ----------------------------------------------------------------------
	# Override routine for creating the sections table
	# Sections dictionary should contain tuples with the range start:end
	# ----------------------------------------------------------------------
	def makeTOC(self):
		pass

	# ----------------------------------------------------------------------
	# Return icon to display
	# ----------------------------------------------------------------------
	def icon(self, name):
		return None

	# ----------------------------------------------------------------------
	# Override routine to highlight messages
	# @param widget the text widget to highlight
	# ----------------------------------------------------------------------
	def highlight(self, widget):
		pass

#===============================================================================
# File Viewer frames
#===============================================================================
class FileViewer(tk.Frame):
	split  = _DEFAULT_SPLIT

	def __init__(self, master, **kw):
		"""Open a viewer assuming that the file exists."""
		super().__init__(master, **kw)

		self._inFocus = False
		self.splitter  = tkExtra.HSplitter(self, FileViewer.split, True)
		self.splitter.pack(expand=tk.YES, fill=tk.BOTH)

		self.treeFrame = self.splitter.leftFrame()
		self.editorFrame = self.splitter.rightFrame()

		# Tree
		self.tree = tkTree.Tree(master=self.treeFrame,
			    root_id="/",
			    root_label="Viewer",
			    get_contents_callback=self.treeContents,
			    width=150,
			    node_class=OVNode
			    )
		self.tree.grid(row=0, column=0, sticky=tk.NSEW)
		self.tree.viewer = self

		self.treeFrame.grid_rowconfigure(0, weight=1)
		self.treeFrame.grid_columnconfigure(0, weight=1)

		# Scroll bars
		sb=tkExtra.AutoScrollbar(self.treeFrame, takefocus=0)
		sb.grid(row=0, column=1, sticky=tk.NS)
		self.tree.configure(yscrollcommand=sb.set)
		sb.configure(command=self.tree.yview)

		sb=tkExtra.AutoScrollbar(self.treeFrame,
				orient=tk.HORIZONTAL, takefocus=0)
		sb.grid(row=1, column=0, sticky=tk.EW)
		self.tree.configure(xscrollcommand=sb.set)
		sb.configure(command=self.tree.xview)

		self.tree.focus_set()

		self.editor = tkTextEditor.TextEditor(self.editorFrame)
		self.editor.txt.config(state=tk.DISABLED)
		self.editor.txt.focus_set()
		self.editor.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		self.editor.txt.bind('<Up>',   lambda e, s=self.editor.txt:
					s.yview(tk.SCROLL, -1, tk.UNITS))
		self.editor.txt.bind('<Down>', lambda e, s=self.editor.txt:
					s.yview(tk.SCROLL,  1, tk.UNITS))
		self.editor.txt.bind('<Button-1>', lambda e, s=self.editor.txt:
					s.focus_set())
		self.editor.txt.bind("<F11>",	self.toggleSplitter)
		self.tree.bind("<F11>",		self.toggleSplitter)
		self.bind("<F11>",	self.toggleSplitter)
		self.bind('<FocusIn>', self._focusIn)

		self.files = []
		self.cwd   = ""		# current working directory
		self.curid = None

	# ----------------------------------------------------------------------
	def setCwd(self, cwd):
		self.cwd = cwd

	# ----------------------------------------------------------------------
	def _append(self, curfile):
		self.files.append(curfile)

		#self.tree.root.set_label(curfile.filename)
		self.tree.root.set_label("Files")
		if self.tree.root.expanded():
			self.tree.root.collapse()
		self.tree.root.expand()

		node = self.tree.root.child_nodes[-1]
		if node is not None:
			node.select()
			if node.expandable(): node.expand()

	# ----------------------------------------------------------------------
	# Routine to be overwritten
	# @return the FileData structure
	# ----------------------------------------------------------------------
	def loadData(self, filename, maxsize):
		curfile = FileData(filename, maxsize)
		if curfile.load():
			return None
		curfile.makeTOC()
		return curfile

	# ----------------------------------------------------------------------
	# Load file, returns True on failure
	# ----------------------------------------------------------------------
	def load(self, filename, maxsize=None):
		curfile = self.loadData(filename, maxsize)
		if curfile is None: return True
		self._append(curfile)
		return False

	# ----------------------------------------------------------------------
	# Show a text(file) from memory
	# ----------------------------------------------------------------------
	def show(self, name, text):
		curfile = FileData(name)
		curfile.set(text)
		self._append(curfile)

	# ----------------------------------------------------------------------
	def hasFile(self, filename):
		return filename in [x.filename for x in self.files if x]

	# ----------------------------------------------------------------------
	def reload(self, filename):
		for i,f in enumerate(self.files):
			if f is not None and f.filename == filename:
				node = self.tree.find_full_id_expand(["/",i])
				break
		if node is not None:
			node.select()
			self.reloadSelected()

	# ----------------------------------------------------------------------
	def reloadSelected(self, event=None):
		selid = self.tree.cursor_node().full_id()
		if len(selid)==1: return
		curfile = self.files[selid[1]]
		curfile.load()
		self.showSection(selid)

	# ----------------------------------------------------------------------
	# Return selected file
	# ----------------------------------------------------------------------
	def selected(self):
		selid = self.tree.cursor_node().full_id()
		if len(selid)==1: return
		return selid[1]

	# ----------------------------------------------------------------------
	def openFile(self, event=None):
		fn = bFileDialog.askopenfilename(master=self)
		if fn:
			self.load(fn)

	# ----------------------------------------------------------------------
	def close(self, event=None):
		if len(self.files)<1:
			return True
		else:
			selid = self.tree.cursor_node().full_id()
			if len(selid)==1: return
			#if len(selid)>2:
			#	node = self.tree.find_full_id_expand(selid[:2])
			#else:
			#	node = self.tree.find_full_id_expand(selid)
			del self.files[selid[1]]
			self.curid = None

			if self.tree.root.expanded():
				self.tree.root.collapse()
			self.tree.root.expand()
			self.tree.last()
			self.tree.select()

			return False

	# ----------------------------------------------------------------------
	def _focusIn(self, event=None):
		if self._inFocus: return
		# FocusIn is generated for all sub-windows, handle only the main window
		if self is not event.widget: return
		self._inFocus = True

		# Check time stamp for all files
		changed = []
		fn      = []
		for i,f in enumerate(self.files):
			if f is None: continue
			try:
				mtime = os.path.getmtime(f.filename)
				if mtime > f.mtime:
					changed.append(i)
					fn.append(f.filename)
			except OSError:
				pass

		if not changed:
			pass
		elif messagebox.askyesno("File has changed",
				f"Files:\n{', '.join(fn)}\nhave changed. Reload?",
				parent=self):
			for i in changed:
				try:
					node = self.tree.root.child_nodes[i]
					self.tree.move_cursor(node)
				except IndexError:
					pass
				self.reloadSelected()
		else:
			for i in changed:
				f = self.files[i]
				try:
					f.mtime = os.path.getmtime(f.filename)
				except OSError:
					pass

		self._inFocus = False

	# ----------------------------------------------------------------------
	def toggleTree(self, event=None):
		if self.splitter.split > self.splitter.min:
			self.splitter.toggle()
			self.splitter.rightFrame().focus_set()
			self.focus_set()
		else:
			self.splitter.toggle()
			if self.splitter.split <= self.splitter.min:
				self.splitter.split = _DEFAULT_SPLIT
				self.splitter.place()

	# ----------------------------------------------------------------------
	# Callback routine to fill tree contents
	# ----------------------------------------------------------------------
	def treeContents(self,node):
		nid = node.full_id()
		if len(nid)==1:
			for idx,f in enumerate(self.files):
				if f is not None:
					self.tree.add_node(
							name=os.path.relpath(f.filename, self.cwd),
							id=idx,
							flag=bool(self.files[idx].sections))
			return

		treelist = None
		for nid in node.full_id():
			if nid == "/":
				treelist = None

			elif isinstance(nid, int):
				curfile  = self.files[nid]
				treelist = curfile.tree

			else:
				# scan list for target
				for name,idx,sublist in treelist:
					if nid==idx:
						treelist = sublist
						break
				else:
					treelist = None

		if treelist is None: return

		nid = node.full_id()
		for name,idx,sublist in treelist:
			icon = curfile.icon(idx)
			if sublist is not None or idx in curfile.sections:
				self.tree.add_node(name=name,
					id=idx,
					flag=(sublist is not None),
					expanded_icon=icon,
					collapsed_icon=icon)

	# ----------------------------------------------------------------------
	def showSection(self, fullid):
		self.editor.txt.config(state=tk.NORMAL)

		# Remember position
		if isinstance(self.curid, int):
			self.files[self.curid].ypos = self.editor.txt.yview()[0]

		# Delete everything
		self.editor.txt.delete(1.0, tk.END)

		# find section infos
		if len(fullid)==1:
			# Root node don't display anything
			self.curid = None
			if isinstance(self.master,tk.Toplevel):
				self.title("FileViewer")
		else:
			self.curid = fullid[1]
			curfile = self.files[self.curid]
			if curfile is None: return
			if isinstance(self.master,tk.Toplevel):
				self.master.title(f"FileViewer: {curfile.filename}")
			if len(fullid)==2:
				self.editor.txt.insert(1.0, curfile.filedata)
				self.editor.txt.yview_moveto(curfile.ypos)
				curfile.highlight(self.editor.txt)
			else:
				section = curfile.sections.get(fullid[-1])
				if section is not None:
					self.editor.txt.insert(tk.INSERT,
						curfile.filedata[section[0]:section[1]])
					curfile.highlight(self.editor.txt)
				self.curid = None	# forget it

		self.editor.txt.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def clear(self):
		self.editor.txt.config(state=tk.NORMAL)
		self.editor.txt.delete(1.0, tk.END)
		self.editor.txt.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def toggleSplitter(self, event=None):
		self.splitter.toggle()
