# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import os
import re
import Evaluation
from deck import CardInfo, Card, CSGCard

#===============================================================================
# Global dictionary for moira, containing all units constants
#===============================================================================
class MoiraGlobalDict(Evaluation.GlobalDict):
	def __init__(self):
		super().__init__()
		#self.redefineDistance_cm(10)	# in mm
		self.redefineEnergy_GeV(1000)	# in MeV
		self.redefineTime_s(1e9)	# in ns

#===============================================================================
# Moira CardInfo class
#===============================================================================
class MoiraCardInfo(CardInfo):
	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		mode = "moira"
		# Populate volumes
		CardInfo.volumes[mode] = [ x.tag for x in CardInfo.cards(mode)
					if CardInfo.GEOMETRY_VOLUME_GROUP in x.group and
					   x.tag[:4] in ("VOLU","CELL")]
		CardInfo.volumes[mode].sort()

		# Populate material definition
		CardInfo.materials[mode] = ["ELEM", "ELEM_FROM_ISOT",
					"ISOT", "MATE",
					"MIXT", "MIXT_BY_WEIGHT",
					"MIXT_BY_NATOMS", "MIXT_BY_VOLUME"]
		CardInfo.materials[mode].sort()

#===============================================================================
# MoiraCard base class
# FIXME:
#	subclass (dict) to replace the whats and properties
#	- whats / list arguments are assigned as [#] = value, starting from 0
#	- properties will be assigned as [name] = value
#	- all internal variables not to be saved will be prefixed
#         with "@" [@name] = value
#===============================================================================
class MoiraCard(Card, CSGCard):
	"""moira input card"""
	MODE       = "moira"
	COMMENT    = "//"
	COMMENTMAC = "#"

	DEF_ROTM   = "__IDENT__"	# default rotation that will be generated always

	cardInfo   = MoiraCardInfo	# Pointer to Card Info
	globalDict = MoiraGlobalDict()

	unitConv   = {  "L" : 10,	# length cm in mm
			"D" : 10*10,	# length^2 cm^2 in mm^2
			"E" : 1000,	# energy GeV in MeV
			#"T" : 1e9,	# time s in ns
			#"A" : 1
		 }

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		super().setTag(tag)
		if tag == "PLACE" or tag.startswith("PLACE."):
			self._geo = True
		elif tag in CardInfo.bodies.get(MoiraCard.MODE,[]):
			self._type = Card.BODY
			self._geo  = True
		elif tag == "CELL" or tag in CardInfo.volumes.get(MoiraCard.MODE,[]):
			self._type = Card.CELL
			self._geo  = True
		elif tag == "LATTICE":
			self._geo  = True
		elif tag in CardInfo.typedef[MoiraCard.MODE]["solid"]:
			self._type = Card.SOLID
			self._geo  = True

	# ----------------------------------------------------------------------
	setTitle = Card.setName

	# ----------------------------------------------------------------------
	# test card against a filter-function
	# ----------------------------------------------------------------------
	def test(self, func, var, value, keepcase=False):
		# Solid, Volume or Cell
		if var in ("solid", "volume"):
			if not self._geo: return
			varlist = self.findType(var)
			if varlist:
				if not keepcase: varlist = [str(x).upper() for x in varlist]
				return value in varlist

		# particle
		#elif var=="particle":

		# Body
		elif var=="body":
			if not self._geo: return
			if self.tag == "CELL":
				opt = re.MULTILINE|re.DOTALL
				if not keepcase: opt |= re.I
				return re.match(r".*\b%s\b.*" % (value), self._text,opt) is not None
			if self.tag not in CardInfo.bodies[self.MODE]: return False
			return (str(self.name()) if keepcase else str(self.name()).upper()) == value

		else:
			return super().test(func, var, value, keepcase)

	# ----------------------------------------------------------------------
	# String functions
	# ----------------------------------------------------------------------
	# Return a formatted comment string
	# ----------------------------------------------------------------------
	def commentStr(self):
		tag = self.tag if self.info.otag is None else self.info.otag
		cs = (self.COMMENTMAC if tag[0]=="/" else self.COMMENT) + " "
		lines = [(cs+x).rstrip() for x in self._comment.split('\n')]
		return "\n".join(lines)+"\n"

	# ----------------------------------------------------------------------
	def evalWhatOrName(self, n, zero=False):
		w = super().evalWhatOrName(n, zero)
		if isinstance(w,str) and " " in w:
			return f'"{w}"'
		return w

	# ---------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def _toStr(self, prefix=""):
		case = self.case()
		try:	vtype = self.info.vtype[case]
		except: vtype = None

		# replace empty whats with default values "!"
		try:    default = self.info.default[case]
		except: default = None

		# FIXME optimize the following loops both the whats
		# and the correction for the default, length

		# Card with parameters
		whats = [self.evalWhat(i) for i in range(self.nwhats())]

		# special cases
		if self.tag == "CELL":	# FIXME maybe add an option in ini
			if len(whats)<self.info.maxwhats:
				whats.extend(["!"]*(self.info.maxwhats-len(whats)))
		n    = len(whats)-1
		last = not self.text()

		# correct whats for units and for empty/!
		for i,w in enumerate(reversed(whats)):
			j = n-i
			if w in ("","!"):
				try:
					whats[j] = default[j]
				except:
					if not last: whats[j] = "!"
			else:
				last = False
				# check for default values and special conversion
				# for length/angle
				try:	# Convert units
					whats[j] = str(float(w) * MoiraCard.unitConv[vtype[j]])
				except (KeyError, ValueError, IndexError, TypeError):
					whats[j] = str(w)

		tag = self.tag if self.info.otag is None else self.info.otag
		if tag[0] == "/":
			s = prefix + tag
		else:
			try:
				# Special case for SOLID.xxx BODY.xxx VOLU.xxx
				# that has special syntax
				pdot = tag.index(".")
				tag  = tag[:pdot]
				if self.type() == Card.SOLID:
					whats.insert(1, self.tag[pdot+1:])
				elif tag == "VOLU":
					# Move material to the end
					if len(whats)>1:
						whats.append(whats.pop(1))
						whats.insert(1, self.tag[pdot+1:])
				s = f"{prefix}:{tag}"
			except ValueError:
				if self._type == Card.BODY:
					whats.insert(1, tag)
					tag = "BODY"
				s = f"{prefix}:{tag}"

		if whats:
			# special treatment
			if tag.startswith("MIXT"):
				s += " " + " ".join(whats[:3])
				if len(whats)>3:
					maxlen = max([len(x) for x in whats[3::2]])
					for i in range(3,len(whats),2):
						try:
							s += "\n\t\t%-*s %s"%(maxlen, whats[i], whats[i+1])
						except IndexError:
							s += "\n\t\t%-*s"%(maxlen, whats[i])
			else:
				s += " " + " ".join(whats)

		if self.text():
			for i,line in enumerate(self.text().splitlines()):
				if i==0:
					s += " "
				else:
					s += "\n\t"
				s += line

		# Organize properties
		used = set()
		# Maybe go with the order that was defined
		for name,var in self.info.property.items():
			v = self.propertyStr(name, used, " ", "!", True)
			if v is not None:
				if s: s += "\n"
				if var.otag is not None:
					s += var.otag + "/" + v
				else:
					s += os.path.dirname(tag) + "/" + v

		# For anything else
		for n,v in self.property.items():
			if n in used: continue
			if n[0]=="@": continue
			if v in ("",None): continue
			if s: s += "\n"
			s += tag + f"/{n}={v}"
		return s
