# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import tempfile
from collections import namedtuple, OrderedDict

import bmath
import Global

from deck    import Input, Card, CardInfo
from .card   import MoiraCard
from .config import MoiraConfig

# FIXME should be in a database or better get dynamically from G4
G4Mat = namedtuple("G4Mat", ["density", "Z", "I"])

#===============================================================================
# Moira Input class
#===============================================================================
class Moira(Input):
	MODE    = MoiraCard.MODE
	EXT     = ".moira"		# default extension	#FIXME move to MoiraConfig
	EXTMAC  = ".macro"		# macro file extension
	EXTMAC2 = ".mac"

	config  = MoiraConfig()		# configuration information

	defaultMaterials       = OrderedDict()
	FLUKA_MAT_2_MOIRA      = {}
	FLUKA_PARTICLE_2_MOIRA = {}

	# ----------------------------------------------------------------------
	def __init__(self, filename=None):
		super().__init__()
		self.mode       = Moira.MODE
		self.globalDict = MoiraCard.globalDict
		if filename: self.load(filename)

	#-----------------------------------------------------------------------
	# Return input filename
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = super().filenames(filename)
		files.append(self.macrofile(filename))
		return files

	#-----------------------------------------------------------------------
	def macrofile(self, filename):
		fn,ext = os.path.splitext(filename)
		if ext != Moira.EXT:
			fn = filename
		return fn + Moira.EXTMAC

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @which	look the Input.cardsSorted()
	# Warning on REGION maybe the "&" you want to get rid off
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		if tag == Card.BODY:
			# include also all CELL parents
			cards = super().cardsSorted(Card.BODY, which)
			cellparents = {x.what(1) : x
					for x in self.cardsCache("CELL").values()}
			cards.update(cellparents)
			return cards
		else:
			return super().cardsSorted(tag, which)

	# ----------------------------------------------------------------------
	# Return an extend list with all cards including the ones from the voxel
	# ----------------------------------------------------------------------
	#def allcards(self):
	#	cl = super().allcards()
	#	for voxel in self["VOXELS"]:
	#		if voxel.ignore() or voxel[MoiraCard.VOXEL] is None: continue
	#		cl.extend(voxel[MoiraCard.VOXEL].cardlist)
	#	return cl

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0):
		lst = super().materialList(which)
		# Add user defined materials
		cards = []
		for tag in CardInfo.materials[self.MODE]:
			cards.extend(self[tag])
		Card.sort(cards)
		lst.extend([x.name() for x in cards])
		return lst

	#-----------------------------------------------------------------------
	def materialDict(self):
		matDict = self.defaultMaterials.copy()
		matDict.update(self.cardsSorted(Card.MATERIALS))
		return matDict

	#-----------------------------------------------------------------------
	# Scan transformations
	# @return rotation matrix from a rot-defi cards or a dictionary with all
	#         transformations
	# FIXME add the other matrix definitions (Euler, etc)
	#-----------------------------------------------------------------------
	def getTransformation(self, idx=None, rotdefi=None, inv=False):
		neg = False
		if isinstance(idx,str):
			if idx and idx[0]=='-':
				neg = True
				idx = idx[1:]
		elif isinstance(idx,int):
			if idx<0:
				neg = True
				idx = -idx
		if inv: neg = not neg

		# return value if cached
		if idx is not None:
			if rotdefi is not None:
				matrix = rotdefi.get(idx)
				if matrix is None:
					return bmath.Matrix(4, type=1)
				elif neg:
					matrix = matrix.inverse()
				return matrix
			matrix = bmath.Matrix(4, type=1)
			assert False	# FIXME temporary for debugging to check if needed

		# ------------------------------------------------------
		# Scan transformations cards
		# ------------------------------------------------------
		# ROTM cards
		# ------------------------------------------------------
		rotdefi = {}
		for card in self.cardsSorted("ROTM",-1).values():
			if not card.nwhats(): continue
			matrix = bmath.Matrix(4, type=1)
			# Set only rotation component
			matrix[0][0] = card.numWhat(1)
			matrix[0][1] = card.numWhat(2)
			matrix[0][2] = card.numWhat(3)
			matrix[1][0] = card.numWhat(4)
			matrix[1][1] = card.numWhat(5)
			matrix[1][2] = card.numWhat(6)
			matrix[2][0] = card.numWhat(7)
			matrix[2][1] = card.numWhat(8)
			matrix[2][2] = card.numWhat(9)
			if idx is None:
				rotdefi[card.name()] = matrix
			#else:
			#	pass

		# ------------------------------------------------------
		# Add translation from LATTICE card
		# FIXME wrong!!!!
		# we can have multiple LATTICE with the same rotdefi
		# but with different translation
		# A unique identifier should be found rot/lattice
		# ------------------------------------------------------
		#for card in self.cardsSorted("LATTICE",-1).values():
		#	name = card.what(1)
		#	matrix = rotdefi.get(name)
		#	if matrix is None:
		#		matrix = bmath.Matrix(4, type=1)
		#	matrix[0][3] = card.numWhat(2)
		#	matrix[1][3] = card.numWhat(3)
		#	matrix[2][3] = card.numWhat(4)
		#	rotdefi[name] = matrix

		# ------------------------------------------------------
		# MATRIX cards
		# ------------------------------------------------------
		#for card in self.cardsSorted("MATRIX",-1).values():
		#	if not card.whats(): continue
		#	matrix = bmath.Matrix(4, type=1)
		#	matrix[0][0] = card.numWhat(1)
		#	matrix[0][1] = card.numWhat(2)
		#	matrix[0][2] = card.numWhat(3)
		#	matrix[0][3] = card.numWhat(4)
		#	matrix[1][0] = card.numWhat(5)
		#	matrix[1][1] = card.numWhat(6)
		#	matrix[1][2] = card.numWhat(7)
		#	matrix[1][3] = card.numWhat(8)
		#	matrix[2][0] = card.numWhat(9)
		#	matrix[2][1] = card.numWhat(10)
		#	matrix[2][2] = card.numWhat(11)
		#	matrix[2][3] = card.numWhat(12)
		#	if idx is None:
		#		rotdefi[card.name()] = matrix
		#	#else:
		#	#	pass

		#if idx is None:
		return rotdefi
		#return matrix

	# ----------------------------------------------------------------------
	# Rotation matrix list
	# ----------------------------------------------------------------------
	# @return a list of linked ROT-DEFI cards with the same idx
	# ----------------------------------------------------------------------
	# FIXME that doens't hold for Moira
	def rotdefiCards(self, idx):
		tags = ["MATRIX","ROTM"]
		return [x for x in self.cardsSorted(tags,-1).values()
				if x.name()==idx]

	#-----------------------------------------------------------------------
	# Rotation matrix list
	# ----------------------------------------------------------------------
	def rotationMatrixList(self, negative=False):
		rot = set()
		try:
			for card in self.cards["MATRIX"]:
				rot.add(card.name())
		except:
			pass
		try:
			for card in self.cards["ROTM"]:
				rot.add(card.name())
		except:
			pass

		l = sorted(rot)
		if negative:
			lst = ["",MoiraCard.DEF_ROTM]
			for i in l:
				lst.append(i)
		else:
			lst = l
			lst.insert(0,MoiraCard.DEF_ROTM)
			lst.insert(0,"")

		del rot
		return lst

	#-----------------------------------------------------------------------
	# Return the body properties as a dictionary of the active/enabled
	# bodies
	#-----------------------------------------------------------------------
	def bodyProperties(self, rotdefi):
		matrix  = None	# current transformation matrix
		#matname = None
		for card in self.allcards():
			if card.ignore(): continue
			if card.tag == "MATRIX":
				if not card.whats():
					#matname = None
					matrix  = None
				else:
					# current transformation matrix
					#matname = card.name()
					matrix  = bmath.Matrix(4,type=1)
					matrix[0][0] = card.numWhat(1)
					matrix[0][1] = card.numWhat(2)
					matrix[0][2] = card.numWhat(3)
					matrix[0][3] = card.numWhat(4)
					matrix[1][0] = card.numWhat(5)
					matrix[1][1] = card.numWhat(6)
					matrix[1][2] = card.numWhat(7)
					matrix[1][3] = card.numWhat(8)
					matrix[2][0] = card.numWhat(9)
					matrix[2][1] = card.numWhat(10)
					matrix[2][2] = card.numWhat(11)
					matrix[2][3] = card.numWhat(12)
			elif card.type() == Card.BODY:
				if matrix:
					card[Card.MATRIX]  = matrix
					#card["@transform"] = matname
				else:
					del card[Card.MATRIX]

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		matDict = self.materialDict()

		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		regionDict = self.cardsSorted(Card.CELL)
		regionList = []
		for region in regionDict.values():
			if region.name() == "&": continue
			# Assign default values to region
			region[Card.TYPE]      = Card.REGION_NORMAL
			region[Card.ASSIGNMAT] = region	# self
			if region.tag == "CELL":
				matname = region.evalWhat(2)
			else:
				matname = region.evalWhat(1)
			region[Card.MATERIAL]= matDict.get(matname)
			region["@biasingA"]  = None	# All particles biasing
			region["@biasingH"]  = None	# Hadrons
			region["@biasingE"]  = None	# emf
			region["@biasingN"]  = None	# low-neutrons
			region[Card.LATTICE] = None
			region["@rotdefi"]   = None
			region["@rotdefi2"]  = None
			region[Card.CONT]    = False	# has continuation cards
			regionList.append(region)
		for card in self.cardsSorted(("LATTICE"),-1).values():
			reg = regionDict.get(card.evalWhat(0))
			reg[Card.TYPE]     = Card.REGION_LATTICE
			reg[Card.MATERIAL] = matDict.get("G4_Galactic")
			reg[Card.LATTICE]  = card
			reg["@rotdefi"]    = card.evalWhat(1)
			reg["@dx"]  = card.numWhat(2)
			reg["@dy"]  = card.numWhat(3)
			reg["@dz"]  = card.numWhat(4)
			#reg["@rotdefi2"] = rotdefi2

	#-----------------------------------------------------------------------
	@staticmethod
	def addFlukaMat2Moira(flukaname, g4name):
		Moira.FLUKA_MAT_2_MOIRA[flukaname] = g4name

	#-----------------------------------------------------------------------
	@staticmethod
	def materialFromFluka(material):
		# FIXME very crude
		# check on predefined ones
		try:
			return Moira.FLUKA_MAT_2_MOIRA[material]
		except KeyError:
			pass
		# check if by appending G4_ exists
		g4mat = f"G4_{material}"
		if g4mat in Moira.defaultMaterials:
			return g4mat
		raise KeyError(f"Material {material} not implemented yet!")

	#-----------------------------------------------------------------------
	#@staticmethod
	#def materialDefaultInFluka(material):
	#	return material in Moira.FLUKA_MAT_2_MOIRA

	#-----------------------------------------------------------------------
	@staticmethod
	def addFlukaParticle2Moira(flukaname, g4name):
		Moira.FLUKA_PARTICLE_2_MOIRA[flukaname] = g4name

	#-----------------------------------------------------------------------
	@staticmethod
	def particleFromFluka(particle):
		return Moira.FLUKA_PARTICLE_2_MOIRA.get(particle,'')

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Export method
	#-----------------------------------------------------------------------
	def _save(self):
		self._saveMoira()
		self._saveMacro()

	#-----------------------------------------------------------------------
	def _saveMoira(self):
		# unique name for identity rotation matrix
		self.writeline("// Default identity rotation matrix")
		self.writeline(":ROTM %s 0 0 0"%(MoiraCard.DEF_ROTM))
		self.writeline()

		# Loop all cards (write the geometry ones)
		for card in self.cardlist:
			if not isinstance(card, MoiraCard): continue
			if card.ignore(): continue
			tag = card.tag if card.info.otag is None else card.info.otag
			if tag[0]=="/": continue
			if card.comment(): self.write(card.commentStr())
			self.writeline(card.toStr())

	#-----------------------------------------------------------------------
	def _saveMacro(self):
		# Second pass write all macro cards
		# Take the filename from the first of the list
		# it could potentially be different than what the input has
		self.open(self.macrofile(self._files[0].name), "w")

		# keep track of all MESH.xxx and dump them at the end
		meshes = []
		for card in self.cardlist:
			if not isinstance(card, MoiraCard): continue
			if card.ignore(): continue
			tag = card.tag if card.info.otag is None else card.info.otag

			# FIXME to be revised for the last version
			if card.tag in ("MESH.BOX", "MESH.CYLINDER"):
				if meshes: # automatically close
					self.writeline("/score/close")
				meshes.append(card.name())
			elif card.tag == "/run/beamOn" and meshes:
				self.writeline("/score/close")

			if tag[0]!="/": continue
			if card.comment(): self.write(card.commentStr())
			self.writeline(card.toStr())
		for mesh in meshes:
			self.writeline(f"/score/dumpAllQuantitiesToFile {mesh} {mesh}.bnn")
		self.close()	# close macro

	#-----------------------------------------------------------------------
	# Import/load a macro and the related moira input files
	#-----------------------------------------------------------------------
	def _load(self):
		macrofile = self.macrofile(self._files[0].name)
		self._loadMoira()
		self._loadMacro(macrofile)

	#-----------------------------------------------------------------------
	# Import/load a moira input file
	#-----------------------------------------------------------------------
	def _loadMoira(self):
		card = None
		self._comment = []		# current comment
		self._expr = ""
		tag = ""
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line, "//")
			if line is None: continue
			words = line.split()
			if words[0][0]==":":
				tag = words[0][1:]
				del words[0]
				what = words
				if tag == "CELL":
					what = words[:3]
					self._expr = " ".join(words[3:])
				else:
					what = words
				card = self._addCard(tag, what)
				if not card: continue

				# Correct units
				case = card.case()
				if case is None: continue

				vtype = card.info.vtype[case]
				for i,w in enumerate(card.whats()):
					try:	# Convert units
						w = str(float(w) / MoiraCard.unitConv[vtype[i]])
						card.setWhat(i,w)
					except (KeyError, ValueError, IndexError, TypeError):
						pass
			else:
				if self._comment:
					card.appendComment(self._comment)
					self._comment = []
				if tag=="CELL":
					card.appendExpr(line)
				elif card:
					card.appendWhats(words)

	#-----------------------------------------------------------------------
	# Load a G4 macro file
	#-----------------------------------------------------------------------
	def _loadMacro(self, filename):
		# Second pass write all macro cards
		# Take the filename from the first of the list
		# it could potentially be different than what the input has
		self.open(filename, "r", showerror=False)
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line, "#")
			if line is None: continue
			tag,*words = line.split()
			self.add(MoiraCard(tag, words, self._comment))
			self._comment = []
		self.close()	# close macro

	#-----------------------------------------------------------------------
	def _addCard(self, tag, what):
		if tag == "BODY":
			tag = what[1]
			del what[1]

		elif tag=="SOLID":
			tag = f"{tag}.{what[1]}"
			del what[1]

		elif tag=="VOLU":
			if len(what)>3:
				tag = f"{tag}.{what[1]}"
				del what[1]
				# Move material
				what.insert(1, what.pop())

		elif tag == "ROTM":
			if what[0]==MoiraCard.DEF_ROTM:
				self._comment = []
				return None

		elif tag == "PLACE":
			if what[3]==MoiraCard.DEF_ROTM:
				what[3] = ""

		card = MoiraCard(tag, what, self._comment)
		self.add(card)
		self._comment = []
		if tag=="CELL":
			card.setExpr(self._expr)
			self._expr = ""

		return card

	#-----------------------------------------------------------------------
	def _parseComment(self, line, prefix):
		line = line.rstrip()
		if not line: return None
		if line.startswith(prefix):
			self._comment.append(line[len(prefix):])
			return None
		return line

	# ----------------------------------------------------------------------
	# Write an input file to be used for plotting
	# ----------------------------------------------------------------------
	def writePlotgeomInput(self, plotgeom, directory=".", title=""):
		ftmp = tempfile.NamedTemporaryFile(mode="wt",
					suffix=Moira.EXT,
					prefix=Global.tmpPrefix,
					dir=directory,
					delete=False)
		self.open(ftmp, "w", backup=False)
		self.preprocess()

		# Write a minimized file with only the geometry and
		# what it is necessary
		# ------------------------------------------------------
		self.write("// Temporary input file only for geometry plotting\n")
		self._saveMoira()

		# -------------------------------------------------------
		# Create a temporary macro file for plotting the geometry
		# -------------------------------------------------------
		fn = os.path.splitext(ftmp.name)[0]
		self.open(fn+Moira.EXTMAC,"w")

		O    = plotgeom.get("O", bmath.Vector.O)
		U    = plotgeom.get("U", bmath.Vector.Z)
		V    = plotgeom.get("V", bmath.Vector.X)
		extu = plotgeom.get("extu", 50.0)
		extv = plotgeom.get("extv", 50.0)
		nu   = int(plotgeom.get("nu",200))
		nv   = int(plotgeom.get("nv",200))

		prefix = ["/field/", "/control/", "FIELD", "&FIELD"]
		for card in self.cardlist:
			if not isinstance(card, MoiraCard): continue
			if card.ignore(): continue
			tag = card.tag if card.info.otag is None else card.info.otag
			if any([card.tag.startswith(x) for x in prefix]):
				Global.dump(f"Moira.writePlotgeomInput field={card}")
				self.writeline(card.toStr())
		unit="cm"
		self.write("# Temporary macro file only for geometry plotting\n")
		self.write("/run/initialize\n")
		self.write(f"/plotter/origin {O.x} {O.y} {O.z} {unit}\n")
		self.write(f"/plotter/u      {U.x} {U.y} {U.z}\n")
		self.write(f"/plotter/v      {V.x} {V.y} {V.z}\n")
		self.write(f"/plotter/extend {extu} {extv} {unit}\n")
		self.write(f"/plotter/grid   {nu} {nv}\n")
		self.write(f"/plotter/plot ! 0 {unit}\n")
		self.close()	# for the macro
		self.close()	# and for moira
		return fn
