# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
from collections import OrderedDict

import log
import Global

from deck    import Input, Card, CardInfo
from .card   import PenelopeCard, PenelopeCardInfo
from .config import PenelopeConfig

#===============================================================================
# Penelope Input class
#===============================================================================
class Penelope(Input):
	MODE    = PenelopeCard.MODE
	EXT     = ".in"				# default extension
	config  = PenelopeConfig()		# configuration information
	defaultMaterials  = OrderedDict()

	PAT_SURFACE = re.compile(r"^SURFACE \((.{4})\)(.*)$")
	PAT_INDICES = re.compile(r"^INDICES=\((.*?),(.*?),(.*?),(.*?),(.*?)\)$")
	PAT_GEOVAR  = re.compile(r"^.{,7}=\((.*?),(.*?)\)$")

	PAT_INLINE1 = re.compile(r"^(.*)(\[.*?])\s*")

	# ----------------------------------------------------------------------
	def __init__(self, filename=None):
		super().__init__()
		self.mode       = Penelope.MODE
		self.globalDict = PenelopeCard.globalDict
		if filename: self.load(filename)

	#-----------------------------------------------------------------------
	# Return input filename
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = super().filenames(filename)
		# FIXME append other files here...
		return files

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @which	look the Input.cardsSorted()
	# Warning on REGION maybe the "&" you want to get rid off
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		return super().cardsSorted(tag, which)

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0):
		return super().materialList(which)

	#-----------------------------------------------------------------------
	def materialDict(self):
		matDict = self.defaultMaterials.copy()
		matDict.update(self.cardsSorted(Card.MATERIALS))
		return matDict

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		pass
		#matDict = self.materialDict()
		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		#---------------------------------------------------------------
		#regionDict = self.cardsSorted(Card.CELL)
		#for region in regionDict.values():
		#	if region.name() == "&": continue
		#	# Assign default values to region
		#	region[Card.TYPE]      = Card.REGION_NORMAL
		#	region[Card.ASSIGNMAT] = region	# self
		#	if region.tag == "CELL":
		#		matname = region.evalWhat(2)
		#	else:
		#		matname = region.evalWhat(1)
		#	region[Card.MATERIAL]= matDict.get(matname)
		#	region["@biasingA"]  = None	# All particles biasing
		#	region["@biasingH"]  = None	# Hadrons
		#	region["@biasingE"]  = None	# emf
		#	region["@biasingN"]  = None	# low-neutrons
		#	region[Card.LATTICE] = None
		#	region["@rotdefi"]   = None
		#	region["@rotdefi2"]  = None
		#	region[Card.CONT]    = False	# has continuation cards

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Export/save method
	#-----------------------------------------------------------------------
	def _save(self):
		Global.dump("\n*-* Writing penelope file")
		# Loop all cards (write the geometry ones)
		for card in self.cardlist:
			if not isinstance(card, PenelopeCard): continue
			if not card.active: continue
			if card.comment(): self.write(card.commentStr())
			self.writeline(card.toStr())

	#-----------------------------------------------------------------------
	# Import/load a macro and the related moira input files
	#-----------------------------------------------------------------------
	def _load(self):
		Global.dump("\n*-* Reading penelope file")
		card = None
		self._comment = []		# current comment
		self._cardEnable = True
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line, "      ")	# 6 spaces
			if line is None: continue

			# Get inline comments
			m = Penelope.PAT_INLINE1.match(line)
			if m:
				self._comment.append(m.group(2))
				line = m.group(1)

			tag, what = self._parseLine(line)
			card = PenelopeCard(tag, what, self._comment)
			if not self._cardEnable: card.disable()

			self.addCard(card)
			self._comment = []
			self._cardEnable = True
			#print(f"line={line}")
			#print(f"tag={tag}, taginfo={nwhats}")

	#-----------------------------------------------------------------------
	# Parse a comment line
	#-----------------------------------------------------------------------
	def _parseComment(self, line, prefix):
		line = line.rstrip()
		if not line: return None
		if isinstance(prefix,str):
			if line.startswith(prefix):
				# Try for a commented card
				lline = line[7:]
				tag, what = self._parseLine(lline)
				ci = CardInfo.get(self.MODE, tag)
				if ci.tag == tag:
					self._cardEnable     = False
					return lline
				self._comment.append(lline)
				return None
		for p in prefix:
			if line.startswith(p):
				self._comment.append(line[len(p):])
				return None
		return line

	#-----------------------------------------------------------------------
	# Parse a line and split into tag and what
	#-----------------------------------------------------------------------
	def _parseLine(self, line):
		tag = line[:6].strip()
		if tag == "TITLE":
			what = [line[7:].rstrip()]
		else:
			ci = PenelopeCardInfo.get(self.MODE, tag)
			nwhats = ci.maxwhats
			what = line[7:].split()[:nwhats]
		return tag, what

	#-----------------------------------------------------------------------
	# Load Constructive Quadratic geometry of Penelope
	#-----------------------------------------------------------------------
	def _loadGeometry(self): #, filename):
		#self.open(filename, "r")
		self._comment = []		# current comment
		# Read prefix comments
		while True:
			line = self.readline()
			if line is None: break
			self._comment.append(line)
			if line[:8] == "00000000": break

		# Read cards separated by 0000 zero lines
		card = None
		self._expr = ""
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line,("#","c ","C "))
			if line is None: continue
			tag = line[:8].strip()

			if tag == "00000000":
				#print(card.toStr())
				card = None
				self._comment.append(line)

			elif card is None:
				if tag == "SURFACE":
					pat = Penelope.PAT_SURFACE.match(line)
					if pat:
						what = pat.group(1).strip()
						inline_comment = pat.group(2).strip()
						if inline_comment:
							self._comment.append(inline_comment)
						card = PenelopeCard(tag, [what], self._comment)
						self.add(card)
					else:
						log.error(f"Line {self._lineNo[-1]} "
							"Invalid syntax of SURFACE")
						log.errorCont(line)
				else:
					log.error("Unknown tag found", self._lineNo[-1])
					log.errorCont(line)

			elif card.tag == "SURFACE":
				if tag=="INDICES=":
					pat = Penelope.PAT_INDICES.match(line)
					if pat:
						whats = [int(x) for x in pat.groups()]
						card.appendWhats(whats)
					else:
						log.error(f"Line {self._lineNo[-1]} "
							"Invalid syntax of INDICES")
						log.errorCont(line)
				elif tag in ("X-SCALE=","Y-SCALE=","Z-SCALE=",
						"X-SHIFT=","Y-SHIFT=","Z-SHIFT="):
					pat = Penelope.PAT_GEOVAR.match(line)
					if pat:
						card.setProperty(tag[:-1], " ".join(pat.groups()))
					else:
						log.error(f"Line {self._lineNo[-1]} "
							f"Invalid syntax of {tag}")
						log.errorCont(line)
			self._comment = []
		#self.close()

	##-----------------------------------------------------------------------
	#def _addCard(self, tag, what):
	#	if tag == "BODY":
	#		tag = what[1]
	#		del what[1]

	#	elif tag=="SOLID":
	#		tag = "%s.%s"%(tag, what[1])
	#		del what[1]

	#	elif tag=="VOLU":
	#		if len(what)>3:
	#			tag = "%s.%s"%(tag, what[1])
	#			del what[1]
	#			# Move material
	#			what.insert(1, what.pop())

	#	elif tag == "ROTM":
	#		if what[0]==PenelopeCard.DEF_ROTM:
	#			self._comment = ""
	#			return None

	#	elif tag == "PLACE":
	#		if what[3]==PenelopeCard.DEF_ROTM:
	#			what[3] = ""

	#	card = PenelopeCard(tag, what, self._comment)
	#	self.add(card)
	#	self._comment = ""
	#	if tag=="CELL":
	#		card.setExpr(self._expr)
	#		self._expr = ""

	#	return card

	# ----------------------------------------------------------------------
	# Write an input file to be used for plotting
	# ----------------------------------------------------------------------
	#def writePlotgeomInput(self, plotgeom, directory=".", title=""):
	#	pass
