#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   21-Jun-2020

import os
import sys
import getopt
import random
import datetime
try:
	import pydicom
except ImportError:
	pydicom = None
try:
	import numpy as np
except ImportError:
	np = None

sys.path.insert(0,os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

import log
import tool
import Data
from Evaluation import Qe

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"
__date__   = "2020-06-21"

#-------------------------------------------------------------------------------
def usrbin2rtdose(usrbin_file, ndet, rtdose_file, rtplan_file, output_file,
		instance=None, sop=None):
	msg = ""
	log.say("\nUsrbin to RTDOSE conversion:")
	log.say(f"\tUsrbin: {usrbin_file}")
	# Load usrbin and converted it to words
	try:
		usrbin = Data.Usrbin(usrbin_file)
	except Exception as ex:
		log.error(f"Cannot open USRBIN {usrbin_file!r}")
		log.errorCont(ex)
		return False, f"Cannot open USRBIN {usrbin_file!r}"

	array  = np.transpose(np.array(usrbin.readArray(ndet)),(2,1,0))
	array[array<0.0] = 0.0		# Filter possibly negative values
	scale  = 0xFFFF / array.max()	# 16 bit word max value 0xffff
	array *= scale			# Values are in range 0..0xffff

	# Load dicom files
	log.say(f"\tRTDOSE: {rtdose_file}")
	try:
		rtdose = pydicom.read_file(rtdose_file, force=True)
	except Exception as ex:
		log.error(f"Cannot open RTDOSE {rtdose_file!r}")
		log.errorCont(ex)
		return False, f"Cannot open RTDOSE {rtdose_file!r}"
	try:
		instanceNumber = rtdose.InstanceNumber
		log.say(f"\tRTDOSE Instance Number = {instanceNumber}")
	except:
		instanceNumber = 0
		log.warning("RTDOSE Instance Number not found assumed 0")

	log.say(f"\tRTPLAN: {rtplan_file}")
	if rtplan_file:
		try:
			rtplan = pydicom.read_file(rtplan_file, force=True)
		except Exception as ex:
			log.error(f"Cannot open RTPLAN {rtdose_file!r}")
			log.errorCont(ex)
			return False, f"Cannot open RTPLAN {rtdose_file!r}"

		# Check if rtplan is the correct one
		refplan_uid = rtdose.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID
		if refplan_uid != rtplan.SOPInstanceUID:
			log.warning("The RTPLAN loaded is not the one referenced from RTDOSE")
			log.warningCont(f"RTDOSE reference: {refplan_uid}")
			log.warningCont(f"RTPLAN SOP UID:   {rtplan.SOPInstanceUID}")
			msg = "The RTPLAN loaded is not the one referenced from RTDOSE"
	else:
		rtplan = None
		log.warning("No RTPLAN was found assuming 1.0 as normalization")
		msg = "No RTPLAN was found assuming 1.0 as normalization"

	if sop is not None:
		rtdose.SOPInstanceUID = sop
	rtdose.SeriesDescription = "Monte Carlo recalculation"
	rtdose.Manufacturer = "http://flair.cern"
	rtdose.ManufacturerModelName = "fluka.4"
	if instance is None:
		rtdose.InstanceNumber = instanceNumber + 100
	else:
		rtdose.InstanceNumber = instance
	log.say(f"\tUSRBIN InstanceNumber = {rtdose.InstanceNumber}")
	sop = rtdose.SOPInstanceUID.split(".")
	# make a random new last part with the same number of digits
	sop[-1] = "".join([str(random.randint(0,9)) for x in range(len(sop[-1]))])
	rtdose.SOPInstanceUID = ".".join(sop)
	now = datetime.datetime.today()
	rtdose.StudyDate = now.strftime("%Y%m%d")
	rtdose.StudyTime = now.strftime("%H%M%S")

	# delete private elements (odd groups)
	for element in rtdose:
		if element.tag.group&1:
			del rtdose[element.tag]

	datatype = rtdose.pixel_array.dtype

	# UID prefixes
	#"1.2.300"	# Greece
	#"1.2.756"	# Switzerland

	# Defined Terms:
	#PLAN
	#	dose calculated for entire delivery of all fraction groups of RT Plan
	#MULTI_PLAN
	#	dose calculated for entire delivery of 2 or more RT Plans
	#FRACTION
	#BEAM
	#BRACHY
	#	dose calculated for entire delivery of one or more Brachy Application Setups
	#	within RT Plan
	#FRACTION_SESSION
	#	dose calculated for a single session ("fraction") of a single Fraction Group
	#	within RT Plan
	#BEAM_SESSION
	#	dose calculated for a single session ("fraction") of one or more Beams
	#	within RT Plan
	#BRACHY_SESSION
	#	dose calculated for a single session ("fraction") of one or more
	#	Brachy Application Setups within RT Plan
	#CONTROL_POINT
	#	dose calculated for one or more Control Points within a Beam for a single fraction
	#RECORD
	#	dose calculated for RT Beams Treatment Record

	# dose calculated for entire delivery of a single Fraction Group within RT Plan
	# BNV one fraction contains multiple beams
	log.say(f"\tDose Summation Type {rtdose.DoseSummationType}")
	beamMet = 0.0
	if rtplan is None:
		pass
	elif rtdose.DoseSummationType == "PLAN":
		for fracSeq in rtplan.FractionGroupSequence:
			for seq in fracSeq.ReferencedBeamSequence:
				beamMet += float(seq.BeamMeterset)
	elif rtdose.DoseSummationType == "FRACTION":
		fracId = int(rtdose.ReferencedRTPlanSequence[0].
				ReferencedFractionGroupSequence[0].
				ReferencedFractionGroupNumber)
		log.say(f"\tFractionId = {fracId}")
		assert fracId == int(rtplan.FractionGroupSequence[fracId-1].FractionGroupNumber)
		for seq in rtplan.FractionGroupSequence[fracId-1].ReferencedBeamSequence:
			beamMet += float(seq.BeamMeterset)

	# dose calculated for entire delivery of one or more Beams within RT Plan
	elif rtdose.DoseSummationType == "BEAM":
		fracId = int(rtdose.ReferencedRTPlanSequence[0]
				.ReferencedFractionGroupSequence[0]
				.ReferencedFractionGroupNumber)
		log.say(f"\tFractionId = {fracId}")
		beamId = int(rtdose.ReferencedRTPlanSequence[0]
				.ReferencedFractionGroupSequence[fracId-1]
				.ReferencedBeamSequence[0]
				.ReferencedBeamNumber)
		log.say(f"\tBeamId = {beamId}")
		assert fracId == int(rtplan.FractionGroupSequence[fracId-1].FractionGroupNumber)
		assert beamId == int(rtplan.FractionGroupSequence[fracId-1]
				.ReferencedBeamSequence[beamId-1]
				.ReferencedBeamNumber)
		beamMet = float(rtplan.FractionGroupSequence[fracId-1]
				.ReferencedBeamSequence[beamId-1]
				.BeamMeterset)

	else:
		log.error(f"Cannot handle DoseSummationType={rtdose.DoseSummationType}")
		return False, f"Cannot handle DoseSummationType={rtdose.DoseSummationType}"

	log.say(f"\tTotal Plan/Fraction/Beam Meterset = {beamMet}")

	# Dose type
	if rtdose.DoseType == "EFFECTIVE":
		rbe = 1.1
	elif rtdose.DoseType == "PHYSICAL":
		rbe = 1.0
	else:
		raise Exception(f"Unknown Dose Type {rtdose.DoseType}")
	log.say(f"\tRBE = {rbe}")

	# Monitor units or number of particles
	if rtplan:
		ionplan = "IonBeamSequence" in rtplan
		beamSequence = rtplan.IonBeamSequence if ionplan \
					else rtplan.BeamSequence
		if beamSequence[0].PrimaryDosimeterUnit == "MU":	# Monitor Units
			log.error("MU (Monitor Units) is not yet implemented.")
			log.errorCont("Assuming 1.0 as normalization")
			msg = "MU (Monitor Units) is not yet implemented."
			norm = 1.0
		elif beamSequence[0].PrimaryDosimeterUnit == "NP":	# Number of Particles
			# Number of fractions
			#no_of_fractions = int(rtplan.FractionGroupSequence[0].NumberOfFractionsPlanned)
			# GeV/g -> eV/kg = 1e12
			norm = beamMet * rbe * Qe * 1e12
		else:
			raise ValueError(f"Unknown Primary Dosimeter Unit {beamSequence[0].PrimaryDosimeterUnit}")
	else:
		norm = 1.0
	log.say(f"\tNormalization = {norm}")

	# Add normalization to Gy or RBE
	detector = usrbin.detector[0]
	rtdose.Columns        = detector.nx
	rtdose.Rows           = detector.ny
	rtdose.NumberOfFrames = detector.nz
	rtdose.PixelSpacing   = [detector.dx*10, detector.dy*10]
	rtdose.ImagePositionPatient = [ (detector.xlow + detector.dx/2)*10,
					(detector.ylow + detector.dy/2)*10,
					(detector.zlow + detector.dz/2)*10]
	rtdose.GridFrameOffsetVector = [f"{round(z*detector.dz*10,8):g}" for z in range(detector.nz)]
	rtdose.DoseGridScaling = norm/scale
	rtdose.PixelData = np.array(array, dtype=datatype).tostring()

	tool.backup(output_file)
	log.say(f"\tOutput RTDOSE file: {output_file}")
	try:
		rtdose.save_as(output_file)
	except Exception as ex:
		log.error(f"Cannot save RTDOSE {output_file!r}")
		log.errorCont(ex)
		return False, f"Cannot save RTDOSE {output_file!r}"
	return True,msg

#-------------------------------------------------------------------------------
# Display usage notice and exit
#-------------------------------------------------------------------------------
def usage(rc=0):
	print()
	print("syntax usrbin2rtdose.py -u usrbin -d rtdoseIN -p plan -o rtdoseOUT [args...]")
	print()
	print(" Options:")
	print("\t-h|-?|--help\tprint this help page")
	print("\t-u #|--usrbin #\tUSRBIN file to load (no-default)")
	print("\t-n #|--det #\tUSRBIN detector to load (Default: 1)")
	print("\t-d #|--rtdose #\tRTDOSE file to load as template (no-default)")
	print("\t-p #|--rtplan #\tRTPLAN file to load (no-default)")
	print("\t-o #|--output #\tOutput RTDOSE file to save (no-default)")
	#print("\t\t\tIf only the filename is specified it will be saved")
	#print("\t\t\tin the same directory of rtdose")
	print("\t--instance #\tSpecify instance number to be used (default 100+current)")
	print("\t--sop #\t\tSpecify SOP id to be used (default current+random last digits)")
	print()
	print(f" Author:\t{__author__}")
	print(f" Email:\t\t{__email__}")
	print(f" Date:\t\t{__date__}")
	sys.exit(rc)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	if len(sys.argv) == 1:
		usage()
	try:
		optlist, args = getopt.getopt(sys.argv[1:],
					"?hu:n:d:p:o:",
					["help",
					"usrbin=",
					"det=",
					"rtdose=",
					"rtplan=",
					"output=",
					"instance=",
					])
	except getopt.GetoptError:
		usage(1)

	usrbin_file = None
	usrbin_det  = 1
	rtdose_file = None
	rtplan_file = None
	output_file = None
	instance    = None
	sop         = None

	for opt, val in optlist:
		if opt in ("-h", "-?", "--help"):
			usage(0)
		elif opt in ("-u", "--usrbin"):
			usrbin_file = val
		elif opt in ("-n", "--det"):
			try:
				usrbin_det = int(val)
			except ValueError:
				log.error(f"Invalid detector number {val!r} specified")
		elif opt in ("-d", "--rtdose"):
			rtdose_file = val
		elif opt in ("-p", "--rtplan"):
			rtplan_file = val
		elif opt in ("-o", "--output"):
			output_file = val
		elif opt == "--instance":
			instance = int(val)
		elif opt == "--sop":
			sop = val

	if usrbin_file is None:
		log.error("Usrbin file not specified")
		sys.exit(1)
	if rtdose_file is None:
		log.error("RTDOSE file not specified")
		sys.exit(1)
	if rtplan_file is None:
		log.error("RTPLAN file not specified")
		sys.exit(1)
	if output_file is None:
		log.error("Output RTDOSE file not specified")
		sys.exit(1)

	#if not os.path.dirname(output_file):
	#	output_file = os.path.join(os.path.dirname(rtdose_file), output_file)

	ok, msg = usrbin2rtdose(usrbin_file, usrbin_det-1, rtdose_file, rtplan_file, output_file,
		instance=instance, sop=sop)
